/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"
#include "idm.h"
#include <sys/stat.h>

#define GNUGPL     "GNUGPL.WRI"


BOOL CALLBACK SimpleProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG:
            CenterDialogBox(hwnd);
            return (TRUE);

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_KEYDOWN:
        case WM_CHAR:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK:
                case IDCANCEL:
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            return (FALSE);
    }

    return (FALSE);
}



BOOL CALLBACK AboutProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG:
            CenterDialogBox(hwnd);
            return (TRUE);

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4201:      /* Disclaimers */
                    DialogBox (hInstance, "Disclaimers", hwnd, SimpleProc);
                    return (TRUE);

                case 4202:      /* Read Copyright Notices */
                    WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_COPYRIGHTS);
                    return (TRUE);

                case 4203: {    /* Read GNU GPL */
                    int len;
                    OFSTRUCT of;
                    char buffer[MAXLINELEN];

                    WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_GNUGPL);
                    return (TRUE);
                }

                case 4204:      /* See Acknowledgements */
                    DialogBox (hInstance, "Acknowledgements", hwnd, SimpleProc);
                    return (TRUE);

                case IDOK:
                case IDCANCEL:
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            return (TRUE);

        case WM_KEYDOWN:
        case WM_CHAR:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;

    }

    return (FALSE);
}



BOOL CALLBACK SignOnProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            RECT wrect, crect;

            GetWindowRect(hwnd, &wrect);
            GetClientRect(hwnd, &crect);

            /*
            MoveWindow(hwnd, 0, 0, 556 + (wrect.right - wrect.left - crect.right),
                                   332 + (wrect.bottom - wrect.top - crect.bottom),
                                   FALSE);
            */
            MoveWindow(hwnd, 0, 0, 440 + (wrect.right - wrect.left - crect.right),
                                   315 + (wrect.bottom - wrect.top - crect.bottom),
                                   FALSE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_PAINT: {
            PAINTSTRUCT ps;
            HDC hdc, hdcmem;
            HBITMAP hbitmap;

            hdc = BeginPaint(hwnd, &ps);
            hdcmem = CreateCompatibleDC(hdc);

            if (hdcmem != NULL) {
                if (GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE) {
                    if (GetDeviceCaps(hdc, SIZEPALETTE) >= 256 &&
                        GetDeviceCaps(hdc, COLORRES) >= 8) {
                            hbitmap = LoadBitmap(hInstance, "TITLE256");
                    } else {
                            hbitmap = LoadBitmap(hInstance, "TITLE16");
                    }
                } else {
                    hbitmap = LoadBitmap(hInstance, "TITLE256");
                }

                if (hbitmap != NULL) {
                    SelectObject(hdcmem, hbitmap);
                    //BitBlt(hdc, 0, 0, 556, 332, hdcmem, 0, 0, SRCCOPY);
                    BitBlt(hdc, 0, 0, 440, 315, hdcmem, 0, 0, SRCCOPY);
                    DeleteObject(hbitmap);
                }

                DeleteDC(hdcmem);
            }

            EndPaint(hwnd, &ps);
            return (TRUE);
        }
    }

    return (FALSE);
}



