/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"

#include "idm.h"

#include <commdlg.h>

#define HIGH_BIT_ESCAPE     0

#define MAXUNDOLEVELS       32000

#define STATUSMARKER        40L     /* Number of markers on status bar */

#ifdef USEWIN32
  #define FILEOPENLIMIT     (256L * 1024L)
#else
  #define FILEOPENLIMIT     (50L * 1024L)
#endif USEWIN32


static struct {
    char *name;
    BOOL automatic;
} SupportedVersions[] = {
    { "B1",   FALSE },      /* Version 1.01 */
    { "B2",   TRUE },       /* Version 1.1 */
    { NULL,   FALSE }
};

extern BOOL FAR PASCAL SummaryProc (HWND, UINT, WPARAM, LPARAM);

static char filename[MAXFILENAMELEN];
static OFSTRUCT of;
static FILEOPTIONS *filep;
static BOOL ConvertLineBreaks, AlignKanjis, OutputLineBreaks;
static int LineLength;
static INT32 FileLength, StatusSteps, CharRead;
static HWND StatusWindowHwnd, StatusWindowHwnd_X;

static HFILE IEfd;
static FILEFORMAT OriginalFormat;
static FILEFORMAT CurrentFormat;
static FILEFORMAT FormatChoice;
static char *OverwriteMessage;

static BOOL HasText, InWord;
static int LineNumber;
static int BlankSpaces, CharPosition;
static int LastIndent, CurrentIndent, FirstIndent;
static int LastChar;
static BOOL LastTextIsKanji;
static BOOL SkipHighBitWarnings;
static BOOL ConfirmFormatOnly;

static POSITION p;


char *FileExtensions[] = {
    ".*",
    ".jwp",
    ".tpl",
    ".euc",
    ".sjs",
    ".jis",
    ".old",
    ".nec",
    NULL
};

char *FileFormatNames[] = {
    "Auto-Detect",
    "Normal",
    "Template",
    "EUC",
    "Shift JIS",
    "New JIS",
    "Old JIS",
    "NEC JIS",
    NULL
};

BOOL FAR PASCAL FileOpenDlgProc (HWND, UINT, WPARAM, LPARAM);
BOOL FAR PASCAL FileOpenConfirmDlgProc (HWND, UINT, WPARAM, LPARAM);
BOOL FAR PASCAL FileSaveAsDlgProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL FileFindProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL SimpleProc (HWND, UINT, WPARAM, LPARAM);



static void MoveStatusBar (INT32 percent)
{
    INT32           len;
    char            buffer[10];
    HWND            hwnd;
    RECT            rect;
    HDC             hdc;
    static INT32    LastPosition = 0L;

    if (StatusWindowHwnd == NULL) return;
    if (percent < 0L) return;

    if (percent > 100L) percent = 100L;

    if (percent == LastPosition) return;

    sprintf(buffer, "%ld%%", percent);
    SetDlgItemText(StatusWindowHwnd, 4204, buffer);

    hwnd = GetDlgItem(StatusWindowHwnd, 4205);
    GetClientRect(hwnd, &rect);
    len = (rect.right - 4) * percent / 100L;

    hdc = GetDC(hwnd);
    SelectObject(hdc, GetStockObject(GRAY_BRUSH));
    Rectangle(hdc, 2, 2, len + 2, rect.bottom - 2);
    ReleaseDC(hwnd, hdc);

    LastPosition = percent;
}



void NewParagraph (FILEOPTIONS *f, POSITION *pos, unsigned int n)
{
    POSITION p = *pos;
    PARAGRAPH far *pp;
    ONELINE far *lp;


    if (PARAOF(p) == NULL) {
        PARAOF(p) = f->paragraph = f->eof = pp = StructAlloc(PARAGRAPH);
        pp->prev = pp->next = NULL;
    } else {
        PARAOF(p)->next = f->eof = pp = StructAlloc(PARAGRAPH);
        pp->prev = PARAOF(p);
        pp->next = NULL;
        PARAOF(p) = pp;
    }

    pp->leftindent = pp->firstindent = pp->rightindent = 0;

    if (pp->prev != NULL) {
        pp->spacing = pp->prev->spacing;
        pp->spacemulti = pp->prev->spacemulti;
    } else {
        pp->spacing = f->basefont->spacing;
        pp->spacemulti = 100;
    }
    n *= TEXTBLOCKSIZE;
    pp->text = (KUNIT far *) BlockAlloc(n * sizeof(KUNIT));
    pp->textsize = n;
    TEXTCHAROF(pp, 0) = '\0';

    LINEOF(p) = lp = pp->lines = pp->lastline = StructAlloc(ONELINE);
    lp->next = lp->prev = NULL;
    lp->position = 0;
    lp->height = BASEFONT->height;
    POSOF(p) = lp->length = 0;

    f->nr_lines++;

    LineNumber = 0;
    *pos = p;
}



static void WriteNormalFile (HFILE fd, FILEOPTIONS *f, BOOL saveundo)
{
    int i, j, k;
    KUNIT far *cp;
    PARAGRAPH far *pp;
    FILEHEADER header;
    FILEPARAHEADER pheader;
    UNDOBUF far *up;

    memset(&header, 0, sizeof(FILEHEADER));
    memset(&pheader, 0, sizeof(FILEPARAHEADER));

    header.magic = MAGIC;
    strcpy(header.version, JWPFILEVERSION);
    header.landscape = f->pagesetup.landscape ? 1 : 0;
    header.paragraphs = 0;

    header.lrheader = f->lrheader;
    header.nofirstpage = f->nofirstpage;
    header.vertical = f->vertical;

    for (i = 0; i < 4; i++) header.margins[i] = (float) f->pagesetup.margins[i];

    for (i = 0; i < NRSUMMARIES && f->summary[i] == NULL; i++);
    if (i < NRSUMMARIES) header.summary = TRUE;

    for (j = 0; j < 4; j++) {
        for (i = 0; i < NRHEADERS && f->header[j][i] == NULL; i++);
        if (i < NRHEADERS) {
            header.headers = TRUE;
            break;
        }
    }


    /* Undo's */

    if (saveundo != 0) {
        if (saveundo < 0) saveundo = f->undolevels;

        for (i = 0, up = f->undo; up != NULL; up = up->next, i++);
        f->undolevels = i;

        if (saveundo > f->undolevels) saveundo = f->undolevels;
        header.undo = saveundo;
    } else {
        header.undo = 0;
    }


    /* How many paragraphs? */

    for (pp = f->paragraph; pp != NULL; pp = pp->next, header.paragraphs++);

    _lwrite (fd, &header, sizeof(FILEHEADER));


    /* Summaries */

    if (header.summary) {
        for (i = 0; i < NRSUMMARIES; i++) {
            if (f->summary[i] == NULL) {
                WriteShortInt(fd, 0);
                continue;
            }

            j = kanjilen(f->summary[i]) + 1;
            if (j >= MAXLINELEN) j = MAXLINELEN - 1;

            WriteShortInt(fd, j);
            _lwrite(fd, f->summary[i], j * sizeof(KANJI));
        }
    }

    /* Headers */

    if (header.headers) {
        for (k = 0; k < 4; k++) {
            for (i = 0; i < NRHEADERS; i++) {
                if (f->header[k][i] == NULL) {
                    WriteShortInt(fd, 0);
                    continue;
                }

                j = kanjilen(f->header[k][i]) + 1;
                if (j >= MAXLINELEN) j = MAXLINELEN - 1;

                WriteShortInt(fd, j);
                _lwrite(fd, f->header[k][i], j * sizeof(KANJI));
            }
        }
    }

    /* Undo's */

    if (saveundo > 0) {
        unsigned int len;
        PARAGRAPH far *pp;

        for (i = 0, up = f->undo; i < saveundo && up != NULL; i++, up = up->next) {
            _lwrite(fd, up, sizeof(UNDOBUF));

            switch (up->action) {
                case U_INSERT: break;
                case U_DELETE:
                case U_REPLACE:
                case U_PARAFORMAT:
                    for (pp = (PARAGRAPH far *) up->data; pp != NULL; pp = pp->next) {
                        _lwrite(fd, pp, sizeof(PARAGRAPH));
                        if (pp->text == NULL) {
                            WriteShortInt(fd, 0);
                        } else {
                            len = unitlen(pp->text) + 1;
                            WriteShortInt(fd, len);
                            _lwrite(fd, pp->text, len * sizeof(KUNIT));
                        }
                    }
                    break;
            }
        }
    }

    /* Output the file */

    for (CharRead = 0L, pp = f->paragraph; pp != NULL; pp = pp->next) {
        pheader.textsize = unitlen(pp->text) + 1;
        pheader.spacemulti = pp->spacemulti;
        pheader.firstindent = pp->firstindent;
        pheader.leftindent = pp->leftindent;
        pheader.rightindent = pp->rightindent;
        pheader.pagebreak = pp->pagebreak;
        _lwrite(fd, &pheader, sizeof(FILEPARAHEADER));

        for (cp = pp->text; cp->kanji; cp++, CharRead++) {
            if (CharRead % StatusSteps == 0)
                MoveStatusBar((CharRead / StatusSteps) * 100L / STATUSMARKER);

            if (ISKANJI(cp->kanji)) {
                WriteChar(fd, HIBYTE(cp->kanji) | 0x0080);
                WriteChar(fd, LOBYTE(cp->kanji) | 0x0080);
            } else if (cp->kanji & 0x0080) {
                WriteChar(fd, HIGH_BIT_ESCAPE);
                WriteChar(fd, cp->kanji & 0x007f);
            } else {
                WriteChar(fd, cp->kanji & 0x007f);
            }
        }

        WriteChar(fd, '\n');
    }

    MoveStatusBar(100L);
}



static BOOL ReadNormalFile (HFILE fd, FILEOPTIONS *f, char * filename, HWND hwnd, BOOL template)
{
    unsigned int i, j, n;
    int ch;
    int len;
    int version;
    FILEHEADER header;
    FILEPARAHEADER pheader;
    POSITION p;

    _lread (fd, &header, sizeof(FILEHEADER));

	if (header.magic != MAGIC) {
        ErrorMessage(hwnd, "%s is not a normal " PROGNAME " %s!",
                            filename, template ? "template" : "file");
        return (FALSE);
    }

    MoveStatusBar((_llseek(fd, 0, 1) / StatusSteps) * 100L / STATUSMARKER);


    /* Verify the version */

    if (!strcmp(header.version, JWPFILEVERSION)) {
        version = -1;
    } else {
		for (i = 0; SupportedVersions[i].name != NULL; i++)
            if (!strcmp(header.version, SupportedVersions[i].name)) break;

        if (SupportedVersions[i].name == NULL) {
            ErrorMessage(hwnd, "%s is a %s created by an unknown version of "
                                PROGNAME " (%s).  Unfortunately, it cannot be read by "
                                "the current version.  We are so very sorry...",
                                filename, template ? "template" : "file",
                                header.version);
            return (FALSE);
        } else if (!SupportedVersions[i].automatic) {
            if (YesNo(hwnd, "%s is a %s created by an earlier version of " PROGNAME " (%s).  %s",
                            filename, template ? "template" : "file",
                            header.version,
							template ? "Do you still want to use it?" :
                                       "It can be read but will be overwritten "
                                       "by the new format when you save it.  "
                                       "Do you still want to open it?") != IDYES)
                return (FALSE);
        }
        version = i;
    }

    for (i = 0; i < 4; i++) f->pagesetup.margins[i] = (double) header.margins[i];
    f->pagesetup.landscape = header.landscape;
    f->lrheader = header.lrheader;
    f->nofirstpage = header.nofirstpage;
    f->vertical = header.vertical;
	f->linelen = CalcLineLength(f);


    /* Any summary? */

    for (i = 0; i < NRSUMMARIES; i++) f->summary[i] = NULL;

    if (header.summary) {
        for (i = 0; i < NRSUMMARIES; i++) {
            ReadShortInt(fd, &n);

            if (n > 0) {
                f->summary[i] = (KANJI far *) BlockAlloc((n + 2) * sizeof(KANJI));
                _lread(fd, f->summary[i], n * sizeof(KANJI));
			}

            MoveStatusBar((_llseek(fd, 0, 1) / StatusSteps) * 100L / STATUSMARKER);
        }
    }



    /* Any headers? */

    for (j = 0; j < 4; j++) for (i = 0; i < NRHEADERS; i++) f->header[j][i] = NULL;

    if (header.headers) {
        for (j = 0; j < 4; j++) {
			for (i = 0; i < NRHEADERS; i++) {
                ReadShortInt(fd, &n);

                if (n > 0) {
                    f->header[j][i] = (KANJI far *) BlockAlloc((n + 2) * sizeof(KANJI));
                    _lread(fd, f->header[j][i], n * sizeof(KANJI));
                }

                MoveStatusBar((_llseek(fd, 0, 1) / StatusSteps) * 100L / STATUSMARKER);
            }
        }
    }


	/* Any undo chains? */

    f->undo = f->undotail = NULL;
    f->undolevels = 0;

    if (!template && header.undo) {
        unsigned int len;
        UNDOBUF far *up = NULL, far *up1;
        PARAGRAPH far *pp, far *pp1;

        for (i = 0; i < header.undo; i++) {
            MoveStatusBar((_llseek(fd, 0, 1) / StatusSteps) * 100L / STATUSMARKER);

			if (up == NULL) {
                up1 = up = StructAlloc(UNDOBUF);
                _lread(fd, up1, sizeof(UNDOBUF));
                up1->next = up1->prev = NULL;
            } else {
                up1->next = StructAlloc(UNDOBUF);
                _lread(fd, up1->next, sizeof(UNDOBUF));
                up1->next->prev = up1;
                up1 = up1->next;
                up1->next = NULL;
            }

            f->undolevels++;

            switch (up1->action) {
                case U_INSERT: up1->data = NULL; break;

                case U_REPLACE: {
                    INT32 here, max, curplace;
                    PARAGRAPH pp;
                    BOOL bug;

                    up1->data = NULL;

                    if (version != -1) {
                        /* Old version - has a bug in Replace */
                        break;
                    }

                    /* Attempts to salvage the situation */

                    curplace = here = _llseek(fd, 0L, 1);
                    max = _llseek(fd, 0L, 2);
                    _llseek(fd, here, 0);

                    do {
                        _lread(fd, &pp, sizeof(PARAGRAPH));
                        curplace += sizeof(PARAGRAPH);
                        if (pp.reserved != 0) break;
                        if (pp.spacing <= 0 || pp.spacemulti <= 0) break;

                        ReadShortInt(fd, &len);
                        curplace += sizeof(INT16);
                        if (len <= 0L) break;
                        if (curplace + (len * sizeof(KUNIT)) >= max) break;

                        _llseek(fd, len * sizeof(KUNIT), 1);
                        curplace += (len * sizeof(KUNIT));
                    } while (pp.next != NULL);


                    bug = FALSE;

                    if (i + 1 < header.undo) {
                        UNDOBUF ub;

                        /* still more undo's */
                        _lread(fd, &ub, sizeof(UNDOBUF));
                        switch(ub.action) {
                            case U_INSERT:
                            case U_DELETE:
                            case U_REPLACE:
                            case U_PARAFORMAT: bug = FALSE; break;
                            default: bug = TRUE; break;
                        }
                    } else {
                        int x;
                        FILEPARAHEADER pheader;

                        /* no more undo's */
                        _lread(fd, &pheader, sizeof(FILEPARAHEADER));
                        if (pheader.textsize <= 0) bug = TRUE;
                        if (here + (len * sizeof(KUNIT)) + pheader.textsize >= max) bug = TRUE;
                        if (pheader.spacemulti < 0) bug = TRUE;
                        if (pheader.reserved != 0) bug = TRUE;
                        for (x = 0; x < 7; x++) if (pheader.extra[x] != 0) bug = TRUE;
                    }

                    _llseek(fd, here, 0);

                    if (bug) break;
                }

                /* ... Falls into next case */

                case U_DELETE:
                case U_PARAFORMAT:
                    pp = NULL;
                    for (;;) {
                        if (pp == NULL) {
                            pp1 = pp = StructAlloc(PARAGRAPH);
                            _lread(fd, pp1, sizeof(PARAGRAPH));
                            pp1->prev = NULL;
                        } else {
                            pp1->next = StructAlloc(PARAGRAPH);
                            _lread(fd, pp1->next, sizeof(PARAGRAPH));
                            pp1->next->prev = pp1;
                            pp1 = pp1->next;
                        }
                        // pp1->next = NULL;   /* Needs pp1->next to end loop */

                        ReadShortInt(fd, &len);

                        if (len > 0) {
                            pp1->text = (KUNIT far *) BlockAlloc((len + 5) * sizeof(KUNIT));
                            _lread(fd, pp1->text, len * sizeof(KUNIT));
                            TEXTCHAROF(pp1, len) = 0;
                        } else {
                            pp1->text = NULL;
                        }

                        if (pp1->next == NULL) break;
                    }

                    up1->data = (void far *) pp;
                    break;
            }
        }

        f->undo = up;
        f->undotail = up1;

        TrimUndoChain(f);
    }

    /* Read the file */

    PARAOF(p) = NULL;
    LINEOF(p) = NULL;
    POSOF(p) = 0;

    CharRead = 0L;

    for (i = 0; i < header.paragraphs; i++) {
        OLDFILEPARAHEADER100 old1;

        switch (version) {
            default:
            case -1: _lread(fd, &pheader, sizeof(FILEPARAHEADER)); break;
            case 0:  _lread(fd, &old1, sizeof(OLDFILEPARAHEADER100));
                     pheader.textsize = old1.textsize;
                     pheader.firstindent = old1.firstindent;
                     pheader.leftindent = old1.leftindent;
                     pheader.rightindent = old1.rightindent;
                     pheader.spacemulti = 100;
                     pheader.pagebreak = FALSE;
                     break;
        }

        MoveStatusBar((_llseek(fd, 0, 1) / StatusSteps) * 100L / STATUSMARKER);

        len = pheader.textsize;
        n = (len / TEXTBLOCKSIZE) + 1;

        NewParagraph(f, &p, n);

        PARAOF(p)->spacemulti = pheader.spacemulti;
        PARAOF(p)->spacing = (f->basefont->height + f->basefont->spacing) * pheader.spacemulti / 100;
        PARAOF(p)->spacing -= f->basefont->height;
        PARAOF(p)->firstindent = pheader.firstindent;
        PARAOF(p)->leftindent = pheader.leftindent;
        PARAOF(p)->rightindent = pheader.rightindent;
        PARAOF(p)->pagebreak = pheader.pagebreak;

        for (j = 0; j < len; j++) {
            ch = ReadChar(fd);

            if (ch & 0x0080) {      /* Kanji */
                ABSCHAROF(p, j) = ((ch & 0x007f) << 8) | (ReadChar(fd) & 0x007f);
                CharRead += 2L;
            } else if (ch == HIGH_BIT_ESCAPE) {
                ABSCHAROF(p, j) = ReadChar(fd) | 0x0080;
                CharRead += 2L;
            } else {
                ABSCHAROF(p, j) = ch;
                CharRead++;
            }
            f->nr_bytes++;
            if (StatusSteps > 0 && CharRead % StatusSteps == 0)
                MoveStatusBar((_llseek(fd, 0, 1) / StatusSteps) * 100L / STATUSMARKER);
        }
        ABSCHAROF(p, len-1) = 0;
        LINEOF(p)->length = len-1;
    }

    MoveStatusBar(100L);
    return (TRUE);
}



static int OpenCharIn (void)
{
    INT32 percent;
    int ch;

    ch = ReadChar(IEfd);

    if (StatusWindowHwnd == NULL) return (ch);

    CharRead++;

    if (CharRead % StatusSteps == 0 || ch == EOF) {
        percent = CharRead / StatusSteps;
        if (ch == EOF || CharRead >= FileLength) percent = 100L;

        MoveStatusBar(percent * 100L / STATUSMARKER);
    }

    return (ch);
}


static void OpenCharOut (int ch)
{
    int i, j;
    KANJI k;
    KANJI buf[10];


    if (ch == '\r') return;         /* Ignore */

    /* Leading spaces */

    if (ch <= ' ' && ch != '\n') {
        if (!HasText) {
            CurrentIndent++;
        } else {
            BlankSpaces++;
        }
        return;
    }

    if (ch == '\n') {
        CurrentIndent = 0;
        LineNumber++;

        if (!ConvertLineBreaks) {
            NewParagraph(filep, &p, 1);
            FirstIndent = LastIndent = 0;
            LineNumber = 0;
        } else if (!HasText) {             /* Blank line */
            NewParagraph(filep, &p, 1);
            FirstIndent = LastIndent = -1;
            LineNumber = 2;
        }

        BlankSpaces = 0;

        HasText = FALSE;
        return;
    }

    /* Check indents */

    if (!HasText) {
        switch (LineNumber) {
            case 0:
                /* No need to check the first line */
                FirstIndent = LastIndent = CurrentIndent;
                PARAOF(p)->firstindent = PARAOF(p)->leftindent = FirstIndent / 2;
                break;

            case 1:
                /* The second line */
                LastIndent = CurrentIndent;
                PARAOF(p)->leftindent = LastIndent / 2;
                break;

            default:
                /* The rest */
                if (LastIndent != CurrentIndent) {
                    NewParagraph(filep, &p, 1);

                    FirstIndent = LastIndent = CurrentIndent;
                    PARAOF(p)->firstindent = PARAOF(p)->leftindent = FirstIndent / 2;
                }
                break;
        }
    }


    /* Take care of the blank spaces */

    if (BlankSpaces > 0) {
        j = (BlankSpaces + 1) / 2;

        if (!HasText) {
            k = 0x2121;
        } else if (!LastTextIsKanji && (ch & 0x80)) {
            if (AlignKanjis) {
                j = 1;
                k = '\t';
            } else {
                j = BlankSpaces;
                k = ' ';
            }
        } else if (LastTextIsKanji && !(ch & 0x80)) {
            j = BlankSpaces;
            k = ' ';
        } else if (!LastTextIsKanji && !(ch & 0x80)) {
            j = BlankSpaces;
            k = ' ';
        } else {
            k = 0x2121;
        }

        buf[0] = k;
        buf[1] = 0;
        for (i = 0; i < j; i++, POSOF(p)++) InsertString (filep, p, buf, 0);

        BlankSpaces = 0;
    }


    /* Insert the character */

    i = 0;

    if (LastChar != 0) {
        if (!HasText && LineNumber > 0 && !LastTextIsKanji) buf[i++] = ' ';
        if (ch & 0x80) {
            buf[i++] = (LastChar << 8) | (ch & 0x7f);
            LastChar = 0;
            LastTextIsKanji = TRUE;
        } else {
            buf[i++] = LastChar;
            buf[i++] = ch;
            LastChar = 0;
            LastTextIsKanji = FALSE;
        }
    } else if (ch & 0x80) {
        LastChar = (ch & 0x7f);
        return;
    } else {
        if (!HasText && LineNumber > 0) buf[i++] = ' ';
        buf[i++] = ch;
        LastTextIsKanji = FALSE;
    }

    buf[i] = 0;

    InsertString (filep, p, buf, 0);
    HasText = TRUE;
    BlankSpaces = 0;

    POSOF(p) += i;
}



BOOL DoFileOpen (char *pathname)
{
	int         i, FileNameOffset;
	HFILE       fd;
	FILEFORMAT  incode;
	HCURSOR     hCursor;
    FILEOPTIONS *ToClose = NULL;
    char        buffer[MAXLINELEN];


    /* Is this the first file opened? */

    if (fileoptions != NULL && fileoptions->next == NULL && !fileoptions->changed) {
        GetWindowText(fileoptions->parent, buffer, 256);
        if (!strcmp(buffer, "Untitled #1")) ToClose = fileoptions;
    }


    if (pathname == NULL) {
		int j;
        OPENFILENAME ofn;

        memset(&ofn, 0, sizeof(OPENFILENAME));

        /* Process the filter string for common dialog boxes*/

        for (i = j = 0; FileExtensions[i] != NULL; i++) {
            strcpy(buffer+j, FileFormatNames[i]);
            j += strlen(FileFormatNames[i]) + 1;
            buffer[j++] = '*';
            strcpy(buffer+j, FileExtensions[i]);
            j += strlen(FileExtensions[i]) + 1;
        }
        buffer[j++] = '\0';
        buffer[j++] = '\0';

        /* Fill in the common dialog box structure */

        ofn.lStructSize = sizeof(OPENFILENAME);
        ofn.hwndOwner = global.hwnd;
        ofn.hInstance = hInstance;
        ofn.lpstrFilter = buffer;
        ofn.lpstrCustomFilter = NULL;
        ofn.nMaxCustFilter = 0L;
        ofn.nFilterIndex = (LONG) (global.openformat + 1);

        filename[0] = '\0';
        ofn.lpstrFile= filename;
        ofn.nMaxFile = MAXFILENAMELEN;
        ofn.lpstrFileTitle = NULL;
        ofn.nMaxFileTitle = 0;
        ofn.lpstrInitialDir = NULL;
        ofn.lpstrTitle = "Open File";
        ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
        ofn.nFileOffset = 0;
        ofn.nFileExtension = 0;
        ofn.lpstrDefExt = NULL;
        ofn.lpfnHook = NULL;

        ofn.lpTemplateName = "FileOpen";

        /* Open the file using a common dialog box */

        if (!GetOpenFileName(&ofn)) return (FALSE);

        FileNameOffset = ofn.nFileOffset;


        /* Check out the file format */

        FormatChoice = ofn.nFilterIndex - 1;
        if (FormatChoice < FF_UNKNOWN || FormatChoice > FF_NEC)
            FormatChoice = FF_UNKNOWN;
    } else {
        strcpy(filename, pathname);
        FormatChoice = FF_UNKNOWN;

        for (i = strlen(pathname) - 1; i >= 0; i--) {
            if (pathname[i] == '\\') break;
        }
        FileNameOffset = i >= 0 ? i+1 : 0;
    }

#ifndef USEWIN32
    AnsiUpper(filename);
#endif USEWIN32

    hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);

    incode = (FILEFORMAT) FormatChoice;

    LastIndent = CurrentIndent = FirstIndent = LastChar = 0;
    LineNumber = BlankSpaces = 0;
    HasText = FALSE;
    LastTextIsKanji = TRUE;

    fd = OpenFile(filename, &of, OF_READ);
    if (fd == HFILE_ERROR) return (FALSE);

    /* Normal format? */

    if (incode == FF_UNKNOWN) {
        unsigned INT32 magic;

        _llseek(fd, 0L, 0);
        _lread(fd, &magic, sizeof(unsigned INT32));
        _llseek(fd, 0L, 0);

        if (magic == MAGIC) incode = FF_NORMAL;
    }


    SetupIO(OpenCharIn, OpenCharOut);
    IEfd = fd;

    FileLength = _llseek(fd, 0L, 2);

    if (FileLength > FILEOPENLIMIT) {
        if (IDYES != YesNo (global.hwnd, "The file %s is %dK.  " PROGNAME " " JWPVERSION " may not be able to "
                                         "handle a file this large.  Do you STILL want "
                                         "to open it?", filename, FileLength / 1024)) {
            _lclose(fd);
            DestroyWindow(StatusWindowHwnd_X);
            StatusWindowHwnd_X = NULL;
            return (FALSE);
        }
    }

    StatusSteps = (FileLength / STATUSMARKER) + 1L;
    CharRead = 0L;

    _llseek(fd, 0L, 0);


    /* Open the status dialog box */

    StatusWindowHwnd_X = CreateDialog(hInstance, "FileStatus", 0, SimpleProc);
    SetDlgItemText(StatusWindowHwnd_X, 4201, "Reading File:");
    sprintf(buffer, "%s  (%0.2fK)", filename, (double) FileLength / 1024.0);
    SetDlgItemText(StatusWindowHwnd_X, 4202, buffer);
    SetDlgItemText(StatusWindowHwnd_X, 4203, "Determining File Format...");
    SetDlgItemText(StatusWindowHwnd_X, 4204, "0%");

    StatusWindowHwnd = NULL;
    if (incode == FF_UNKNOWN) {
        incode = DetectCodeType();
        _llseek(fd, 0L, 0);
    }

    if (incode == FF_UNKNOWN) {
        if (YesNo(global.hwnd, "The file %s is not in any of the "
                               "supported Japanese text formats, or "
                               "it may not contain any Japanese text!\n\n"
                               "Do you still want to open it?", filename) != IDYES) {
            _lclose(fd);
            DestroyWindow(StatusWindowHwnd_X);
            StatusWindowHwnd_X = NULL;
            return (FALSE);
        }
        incode = FF_EUC;
    } else if (incode >= FF_EUC) {
        CurrentFormat = incode;
        ConfirmFormatOnly = FALSE;
        DialogBox (hInstance, "FileFormatConfirm", StatusWindowHwnd_X, FileOpenConfirmDlgProc);
        if (CurrentFormat == FF_UNKNOWN) {  /* User cancelled */
            _lclose(fd);
            DestroyWindow(StatusWindowHwnd_X);
            StatusWindowHwnd_X = NULL;
            return (FALSE);
        }
        incode = CurrentFormat;
    }

    sprintf(buffer, "File Format: %s", (incode == FF_UNKNOWN) ? "Unknown" : FileFormatNames[incode]);
    StatusWindowHwnd = StatusWindowHwnd_X;
    SetDlgItemText(StatusWindowHwnd, 4203, buffer);


    filep = NewFile(FN_NORMAL, (incode != FF_NORMAL && incode != FF_TEMPLATE));
    if (filep == NULL) {
        _lclose(fd);
        ErrorMessage(global.hwnd, "Cannot open any more file window!");
        DestroyWindow(StatusWindowHwnd);
        StatusWindowHwnd = NULL;
        return (FALSE);
    }

    if (incode == FF_NORMAL || incode == FF_TEMPLATE) {
        strcpy(filep->filename, of.szPathName);

        if (!ReadNormalFile(fd, filep, filep->filename, StatusWindowHwnd, FALSE)) {
            FILEOPTIONS *f1;

            _lclose(fd);
            DestroyWindow(StatusWindowHwnd);
            StatusWindowHwnd = NULL;

            ShowCursor(FALSE);
            SetCursor(hCursor);

            CloseFile(filep);

            /* Unlink it */

            if (fileoptions == filep) {
                fileoptions = filep->next;
            } else {
                for (f1 = fileoptions; f1 != NULL && f1->next != filep; f1 = f1->next);
                if (f1 == NULL) {
                    ErrorMessage(global.hwnd, "Funny: cannot find FILEOPTIONS record to deallocate");
                } else {
                    f1->next = filep->next;
                }
            }
            return (FALSE);
        }
    } else {
        filep->filename[0] = '\0';

        PARAOF(p) = filep->paragraph;
        LINEOF(p) = filep->paragraph->lines;
        POSOF(p) = 0;

        FileImport(incode);
    }

    _lclose(fd);



    /* Reformat */

    SetDlgItemText(StatusWindowHwnd_X, 4201, "Reformatting...");

    TOPPARA(filep) = CURPARA(filep) = filep->paragraph;
    TOPLINE(filep) = CURLINE(filep) = filep->paragraph->lines;
    CURCHAR(filep) = 0;

    ChangePrinterOrFonts(filep, TRUE, TRUE);

    if (filep->repage) RepagenateFile(filep);

    filep->repage = FALSE;
    filep->changed = FALSE;

    DestroyWindow(StatusWindowHwnd);
    StatusWindowHwnd = NULL;

    strcpy(filep->filename, filename);

    if (incode != FF_UNKNOWN && incode != FF_NORMAL && incode != FF_TEMPLATE) {
        filep->fileformat = incode;
        sprintf(buffer, "%s (%s)", filename + FileNameOffset, FileFormatNames[incode]);
    } else {
        strcpy(buffer, filename + FileNameOffset);
    }
    ShowFileWindow(buffer, filep);


    if (ToClose != NULL) {
        ToClose->changed = FALSE;
        SendMessage(global.clienthwnd, WM_MDIDESTROY, (WPARAM) ToClose->parent, 0L);
    }

    ShowCursor(FALSE);
    SetCursor(hCursor);

    PutQuickFile(of.szPathName);
    UpdateQuickFiles();

    return (TRUE);
}



FILEOPTIONS *CreateUsingTemplate (char far *tpl, int Untitled)
{
    HFILE fd;
    OFSTRUCT of;
    FILEOPTIONS *f, *ToClose = NULL;
    char buffer[MAXLINELEN];

    /* Is this the first file opened? */

    if (fileoptions != NULL && fileoptions->next == NULL && !fileoptions->changed) {
        GetWindowText(fileoptions->parent, buffer, 256);
        if (!strcmp(buffer, "Untitled #1")) ToClose = fileoptions;
    }


    fd = OpenFile(tpl, &of, OF_READ);
    if (fd == HFILE_ERROR) return (NULL);

    _llseek(fd, 0L, 0);

    f = NewFile(FN_NORMAL, FALSE);
    if (f == NULL) {
        _lclose(fd);
        ErrorMessage(global.hwnd, "Cannot open any more file window!");
        return (NULL);
    }

    StatusWindowHwnd = NULL;
    StatusSteps = 1L;           /* To prevent divide by zero */

    if (!ReadNormalFile(fd, f, "This", global.hwnd, TRUE)) {
        FILEOPTIONS *f1;

        _lclose(fd);
        CloseFile(f);

        /* Unlink it */

        if (fileoptions == f) {
            fileoptions = f->next;
        } else {
            for (f1 = fileoptions; f1 != NULL && f1->next != f; f1 = f1->next);
            if (f1 == NULL) {
                ErrorMessage(global.hwnd, "Funny: cannot find FILEOPTIONS record to deallocate");
            } else {
                f1->next = f->next;
            }
        }
    }

    _lclose(fd);

    TOPPARA(f) = CURPARA(f) = f->paragraph;
    TOPLINE(f) = CURLINE(f) = f->paragraph->lines;
    CURCHAR(f) = 0;

    ChangePrinterOrFonts(f, TRUE, TRUE);

    f->fileformat = FF_NORMAL;
    sprintf(buffer, "Untitled #%d", Untitled);
    ShowFileWindow(buffer, f);

    RepagenateFile(f);

    f->changed = FALSE;
    f->repage = FALSE;

    /* Null the title */

    if (f->summary[0] != NULL) {        /* Title */
        FreeBlock(f->summary[0]);
        f->summary[0] = NULL;
    }

    if (ToClose != NULL) {
        ToClose->changed = FALSE;
        SendMessage(global.clienthwnd, WM_MDIDESTROY, (WPARAM) ToClose->parent, 0L);
    }
    return (f);
}



static int SaveCharIn (void)
{
    int i;
    KANJI ch, temp;
    INT32 percent;
    HWND hwnd;

    if (LastChar != 0) {
        ch = LastChar;
        LastChar = 0;
        CharPosition++;
        return (ch);
    } else if (PARAOF(p) == NULL || LINEOF(p) == NULL) {
        goto EndOfFile;
    } if (OutputLineBreaks && !HasText && CharPosition <
            (LineNumber ? PARAOF(p)->leftindent : PARAOF(p)->firstindent) * 2) {
        CharPosition++;
        return (' ');
    } else if (OutputLineBreaks && CharPosition >= LineLength - PARAOF(p)->rightindent) {
        HasText = InWord = FALSE;
        CharPosition = 0;
        LineNumber++;
        return ('\n');
    }


    for (;;) {
        ch = (LINECHAROF(p, POSOF(p)) & 0x7fff);
        if (ISKANJI(ch)) ch &= 0x7f7f;

        if (!ch) {
            POSOF(p) = 0;
            PARAOF(p) = PARAOF(p)->next;
            if (PARAOF(p) != NULL) LINEOF(p) = PARAOF(p)->lines;
            else LINEOF(p) = NULL;

            HasText = InWord = FALSE;
            CharPosition = LineNumber = 0;
            return ('\n');
        }

        if (!HasText && (ch <= ' ' || ch == 0x2121)) {
            POSOF(p)++;
            CharRead++;
            continue;
        }

        if (ISKANJI(ch)) {
            if (ch == 0x2121) {
                ch = LastChar = ' ';
            } else {
                LastChar = (LOBYTE(ch) | 0x80);
                ch = (HIBYTE(ch) | 0x80);
            }
            InWord = FALSE;
        } else if (ch <= ' ') {
            ch = ' ';
            InWord = FALSE;
        } else {
            if (!SkipHighBitWarnings && ch & 0x80) {
                if (StatusWindowHwnd == NULL) hwnd = global.hwnd; else hwnd = StatusWindowHwnd;

                if (YesNo(hwnd, "This file contains "
                                "the extended ASCII character '%c'.  "
                                "Unfortunately, the file format you have "
                                "chosen does not support extended ASCII "
                                "characters.  "
                                "The high-bit of this character will "
                                "be lost in the saved file, and it will "
                                "become '%c'.\n\n"
                                "Do you want to continue reporting "
                                "other extended ASCII characters?",
                                ch, ch & 0x7f) == IDNO)

                    SkipHighBitWarnings = TRUE;
            }

            ch &= 0x7f7f;

            if (strchr("-", ch) != NULL) {
                InWord = FALSE;
            } else {
                if (!InWord) {          /* Search for word break */
                    for (i = 1; ; i++) {
                        temp = LINECHAROF(p, POSOF(p)+i);
                        if (!temp) break;
                        if (IsSmallKana(temp)) break;
                        if (temp <= ' ' || strchr("-", temp) != NULL) break;

                        if (OutputLineBreaks && CharPosition + i >= LineLength - PARAOF(p)->rightindent) {
                            HasText = InWord = FALSE;
                            CharPosition = 0;
                            LineNumber++;
                            return ('\n');
                        }
                    }
                }
                InWord = TRUE;
            }
        }

        break;
    }

    POSOF(p)++;
    CharPosition++;
    HasText = TRUE;

    if (StatusWindowHwnd == NULL) return (ch);

    CharRead++;

    if (CharRead % StatusSteps == 0 || ch == EOF) {
        percent = CharRead / StatusSteps;
        if (ch == EOF || CharRead >= FileLength) percent = 100L;

        MoveStatusBar(percent * 100L / STATUSMARKER);
    }
    return (ch);

EndOfFile:

    percent = 100L;
    MoveStatusBar(100L);
    return (EOF);
}



static void SaveCharOut (int ch)
{
    /* Convert LF to CR-LF pairs */
    if (ch == '\r') return;
    if (ch == '\n') WriteChar(IEfd, '\r');
    WriteChar(IEfd, ch);
}



BOOL FAR PASCAL FileSaveFormatProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            char buffer[100];

#ifndef USEWIN32
            AnsiUpper(filename);
#endif USEWIN32
            sprintf(buffer, OverwriteMessage, filename, FileFormatNames[OriginalFormat]);

            SetDlgItemText(hwnd, 4211, buffer);
            SetDlgItemText(hwnd, 4201, FileFormatNames[OriginalFormat]);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32
        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK:      OriginalFormat = FF_NORMAL;
                                EndDialog(hwnd, TRUE);
                                return (TRUE);

                case IDCANCEL:  OriginalFormat = FF_UNKNOWN;
                                EndDialog(hwnd, TRUE);
                                return (TRUE);

                case 4201:      EndDialog(hwnd, TRUE);
                                return (TRUE);

                default:        break;
            }
            break;
    }

    return (FALSE);
}



static BOOL SwitchToTemplatesDirectory(char *olddir)
{
    int  i;
    char buffer[MAXLINELEN];

    GetCurrentDirectory(MAXFILENAMELEN, olddir);
    _fstrcpy(buffer, global.templatepath);
    i = strlen(buffer);
    if (i > 3) buffer[i-1] = '\0';

    if (!SetCurrentDirectory(buffer)) {
        if (!MakeDirectory(buffer)) {
            ErrorMessage(global.hwnd, "Cannot switch to / create the JWP templates directory!");
            SetCurrentDirectory(olddir);
            return (FALSE);
        }
        if (!SetCurrentDirectory(buffer)) {
            ErrorMessage(global.hwnd, "Cannot switch to the JWP templates directory!");
            SetCurrentDirectory(olddir);
            return (FALSE);
        }
    }

    return (TRUE);
}



static void ResetOldDirectory(char *olddir)
{
    SetCurrentDirectory(olddir);
}



BOOL DoFileSaveAs (BOOL Dialog)
{
    HFILE       fd;
    int         i, FileNameOffset;
    FILEFORMAT  outcode;
    HCURSOR     hCursor;
    char        olddir[MAXFILENAMELEN];
    char        buffer[MAXLINELEN];

    if (global.active == NULL) return (FALSE);

    filep = global.active;

    if (Dialog || !filep->filename[0]) {
        int i, j;
        OPENFILENAME ofn;

        strcpy(filename, filep->filename);

        memset(&ofn, 0, sizeof(OPENFILENAME));

        /* Process the filter string for common dialog boxes*/

        for (i = FF_NORMAL, j = 0; FileExtensions[i] != NULL; i++) {
            if (i == FF_UNKNOWN) continue;
            strcpy(buffer+j, FileFormatNames[i]);
            j += strlen(FileFormatNames[i]) + 1;
            buffer[j++] = '*';
            strcpy(buffer+j, FileExtensions[i]);
            j += strlen(FileExtensions[i]) + 1;
        }
        buffer[j++] = '\0';
        buffer[j++] = '\0';

        /* Gets rid of the extension */
        for (i = strlen(filename) - 1; i >= 0; i--) {
            if (filename[i] == '.') break;
            if (filename[i] == '\\') break;
        }
        if (i > 0 && filename[i] == '.') filename[i] = '\0';

        /* Initialize the common dialog box */

        ofn.lStructSize = sizeof(OPENFILENAME);
        ofn.hwndOwner = global.hwnd;
        ofn.lpstrFilter = buffer;
        ofn.nFilterIndex = (filep->fileformat == FF_UNKNOWN) ? 1 : filep->fileformat - FF_NORMAL + 1;
        ofn.lpstrFile= filename;
        ofn.nMaxFile = MAXLINELEN;
        ofn.lpstrFileTitle = NULL;
        ofn.nMaxFileTitle = 0;
        ofn.lpstrInitialDir = NULL;
        ofn.Flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;

        if (!GetSaveFileName(&ofn)) return (FALSE);

        FileNameOffset = ofn.nFileOffset;


        /* Check out the file format */

        FormatChoice = (ofn.nFilterIndex - 1) + FF_NORMAL;
        if (FormatChoice < FF_NORMAL || FormatChoice > FF_NEC)
            FormatChoice = FF_UNKNOWN;

        if (FormatChoice == FF_UNKNOWN) {
            CurrentFormat = FormatChoice;
            ConfirmFormatOnly = TRUE;
            DialogBox (hInstance, "FileFormatConfirm", global.hwnd, FileOpenConfirmDlgProc);
            if (CurrentFormat == FF_UNKNOWN) {  /* User cancelled */
                return (FALSE);
            }
            FormatChoice = CurrentFormat;
        }

        outcode = FormatChoice;

        if (ofn.nFileExtension > 0) {
            if (!filename[ofn.nFileExtension]) {
                strcat(filename, FileExtensions[FormatChoice]);
            }
        }

        if (outcode == FF_NORMAL || outcode == FF_TEMPLATE) {
            i = DialogBox (hInstance, "Summary", global.hwnd, SummaryProc);
            if (!i) return (FALSE);
        }
        if (outcode == FF_TEMPLATE) {
            /* Modify to template directory */
            sprintf(buffer, "%s%s", global.templatepath, filename + FileNameOffset);
            strcpy(filename, buffer);
            FileNameOffset = strlen(global.templatepath);
        }
    } else if (filep->fileformat != FF_NORMAL && filep->fileformat != FF_TEMPLATE) {
        BOOL same;

        strcpy(filename, filep->filename);

        /* Change the extension to .JWP */

        for (i = strlen(filename) - 1; i >= 0; i--) {
            if (filename[i] == '\\' || filename[i] == '/' ||
                filename[i] == '.' || filename[i] == ':') break;
        }

        if (i < 0 || filename[i] != '.') {
            strcat(filename, FileExtensions[1]);    /* Normal extension */
            same = FALSE;
        } else if (!strncpy(filename + i, FileExtensions[1], 4)) {
            same = TRUE;
        } else {
            strcpy(filename + i, FileExtensions[1]);
            same = FALSE;
        }


        /* Overwrite with normal format? */

        OriginalFormat = filep->fileformat;

        if (!same) OverwriteMessage = "Save file as %s (Normal Format)?\nOr keep it %s?";
        else OverwriteMessage = "Overwrite %s with Normal Format?\nOr keep it %s?";

        MessageBeep(MB_ICONQUESTION);

        DialogBox(hInstance, "FileSaveFormat", global.hwnd, FileSaveFormatProc);

        if (OriginalFormat == FF_UNKNOWN) return (FALSE);

        if (OriginalFormat != FF_NORMAL) strcpy(filename, filep->filename);

        outcode = OriginalFormat;
        FormatChoice = outcode - FF_NORMAL;

        if (outcode == FF_NORMAL)
            SendMessage(global.hwnd, WM_COMMAND, IDM_FILESUMMARY, 0L);

        for (i = strlen(filename) - 1; i >= 0; i--) {
            if (filename[i] == '\\') break;
        }
        FileNameOffset = i >= 0 ? i+1 : 0;
    } else {
        strcpy(filename, filep->filename);
        outcode = filep->fileformat;
        FormatChoice = outcode - FF_NORMAL;

        for (i = strlen(filename) - 1; i >= 0; i--) {
            if (filename[i] == '\\') break;
        }
        FileNameOffset = i >= 0 ? i+1 : 0;
    }

    if (outcode == FF_TEMPLATE) {
        /* Template? */

        if (filep->summary[0] == NULL) {    /* No title */
            ErrorMessage(global.hwnd, "You must provide a Title for this template.");
            return (FALSE);
        }

        if (!SwitchToTemplatesDirectory(olddir)) return (FALSE);

    } else if (outcode >= FF_EUC) {
        /* Non-JWP format */

        CurrentFormat = outcode;
        DialogBox (hInstance, "SaveOptions", global.hwnd, FileSaveAsDlgProc);
        if (CurrentFormat == FF_UNKNOWN) {  /* User cancelled */
            return (FALSE);
        }
    }


#ifndef USEWIN32
    AnsiUpper(filename);
#endif USEWIN32

    /* Now save the file */

    hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);

    LineNumber = LastChar = CharPosition = 0;
    HasText = InWord = FALSE;

    fd = OpenFile(filename, &of, OF_WRITE | OF_CREATE);
    if (fd == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Cannot create file %s!", filename);
        return (FALSE);
    }

    _llseek(fd, 0L, 0);

    SetupIO(SaveCharIn, SaveCharOut);
    IEfd = fd;


    /* Start saving */

    FileLength = 0;
    StatusSteps = (FileLength / STATUSMARKER) + 1L;
    CharRead = 0L;

    /* Open the status dialog box */

    StatusWindowHwnd = CreateDialog(hInstance, "FileStatus", 0, SimpleProc);
    SetDlgItemText(StatusWindowHwnd, 4201, "Writing File:");
    SetDlgItemText(StatusWindowHwnd, 4202, filename);
    sprintf(buffer, "File Format: %s", FileFormatNames[outcode]);
    SetDlgItemText(StatusWindowHwnd, 4203, buffer);
    SetDlgItemText(StatusWindowHwnd, 4204, "0%");


    CharRead = 0L;
    FileLength = filep->nr_bytes;

    StatusSteps = (FileLength / STATUSMARKER) + 1L;


    if (outcode == FF_NORMAL) {
        WriteNormalFile(fd, filep, global.saveundolevels);
    } else if (outcode == FF_TEMPLATE) {
        WriteNormalFile(fd, filep, 0);
    } else {
        PARAOF(p) = filep->paragraph;
        LINEOF(p) = filep->paragraph->lines;
        POSOF(p) = 0;

        SkipHighBitWarnings = FALSE;
        FileExport(outcode);
    }

    _lclose(fd);

    strcpy(filep->filename, filename);
    filep->fileformat = outcode;

    if (outcode != FF_NORMAL && outcode != FF_TEMPLATE) {
        sprintf(buffer, "%s (%s)", filep->filename + FileNameOffset, FileFormatNames[outcode]);
    } else {
        strcpy(buffer, filep->filename + FileNameOffset);
    }
    SetWindowText(filep->parent, buffer);

    DestroyWindow(StatusWindowHwnd);
    StatusWindowHwnd = NULL;

    if (outcode == FF_TEMPLATE) ResetOldDirectory(olddir);

    ShowCursor(FALSE);
    SetCursor(hCursor);

    filep->changed = FALSE;

    PutQuickFile(of.szPathName);
    UpdateQuickFiles();

    return (TRUE);
}



BOOL FAR PASCAL FileOpenConfirmDlgProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int i;


    switch (message) {

    case WM_INITDIALOG:
        if (ConfirmFormatOnly) {
            for (i = FF_EUC; FileFormatNames[i] != NULL; i++)
                SendDlgItemMessage(hwnd, 4211, LB_INSERTSTRING, -1, (LONG) ((LPSTR) FileFormatNames[i]));
            SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, 0, 0L);

            ShowWindow(GetDlgItem(hwnd, 4221), FALSE);
            ShowWindow(GetDlgItem(hwnd, 4222), FALSE);

            SetWindowText(GetDlgItem(hwnd, IDOK), "Save!");
        } else {
            for (i = FF_EUC; FileFormatNames[i] != NULL; i++)
                SendDlgItemMessage(hwnd, 4211, LB_INSERTSTRING, -1, (LONG) ((LPSTR) FileFormatNames[i]));
            SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, CurrentFormat - FF_EUC, 0L);

            CheckDlgButton (hwnd, 4221, FALSE);
            CheckDlgButton (hwnd, 4222, TRUE);

            SetWindowText(GetDlgItem(hwnd, IDOK), "Open!");
        }

        CenterDialogBox(hwnd);
        return (TRUE);

    case WM_CHAR:
        switch (wParam) {
            case '\x1b': SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0L);
                         return (TRUE);
        }
        break;

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

    case WM_COMMAND:
        switch (WINCMDID) {
            case IDOK:
                CurrentFormat = (WORD) SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                CurrentFormat += FF_EUC;

                if (!ConfirmFormatOnly) {
                    ConvertLineBreaks = IsDlgButtonChecked(hwnd, 4221);
                    AlignKanjis = IsDlgButtonChecked(hwnd, 4222);
                }

                EndDialog(hwnd, TRUE);
                return (TRUE);

            case IDCANCEL:
                CurrentFormat = FF_UNKNOWN;
                EndDialog(hwnd, FALSE);
                return (TRUE);
        }
    }

    return (FALSE);
}



BOOL FAR PASCAL FileSaveAsDlgProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    char buffer[10];


    switch (message) {

    case WM_INITDIALOG:
        CheckDlgButton (hwnd, 4211, 0);
        SetDlgItemText(hwnd, 4212, "");
        EnableWindow(GetDlgItem(hwnd, 4223), FALSE);
        return (TRUE);

    case WM_CHAR:
        switch (wParam) {
            case '\x1b': SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0L);
                         return (TRUE);
        }
        break;

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

    case WM_COMMAND:
        switch (WINCMDID) {
            case IDOK:
                OutputLineBreaks = IsDlgButtonChecked(hwnd, 4211);
                if (OutputLineBreaks) {
                    GetDlgItemText(hwnd, 4212, buffer, 3);
                    LineLength = atoi(buffer);
                    if (LineLength <= 0) {
                        ErrorMessage(global.hwnd, "Bad Line Length: %s", buffer);
                        sprintf(buffer, "%d", global.active->linelen);
                        SetDlgItemText(hwnd, 4212, buffer);
                        return (TRUE);
                    }
                    LineLength *= 2;
                }
                EndDialog(hwnd, TRUE);
                return (TRUE);

            case IDCANCEL:
                CurrentFormat = FF_UNKNOWN;
                EndDialog(hwnd, FALSE);
                return (TRUE);

            case 4211:
                if (IsDlgButtonChecked(hwnd, 4211)) {
                    sprintf(buffer, "%d", global.active->linelen);
                    EnableWindow(GetDlgItem(hwnd, 4212), TRUE);
                    SetDlgItemText(hwnd, 4212, buffer);
                } else {
                    EnableWindow(GetDlgItem(hwnd, 4212), FALSE);
                    SetDlgItemText(hwnd, 4212, "");
                }
                return (TRUE);
        }
    }

    return (FALSE);
}
