/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"
#include "idm.h"

#include <shellapi.h>
#include <math.h>


#define CTRLCHAR            '^'
#define SHIFTCHAR           '+'

#define RESERVEDCTRLCHAR    "HIM"

#define NRCTRLKEYS          ('~' - ' ')
#define NRFUNCKEYS          12

#define ALPHASTART          4333        /* Decimal 32 is space */
#define FUNCSTART           4501

#define PSBLOCKSIZE         20

#define MAXUNDOLEVELS       32000
#define DEFUNDOLEVELS       50


typedef struct {
    BOOL ctrl;
    BOOL shift;
    char alpha;
    WORD scancode;
} HOTKEY;


typedef struct {
    char *desc;
    int menuid;
    HOTKEY hotkey;
} JWPFUNC;


static FARPROC OldStaticProc;
static int Selected;
static int CurrentFunction = 1;

static HOTKEY OldKey, CurrentKey;

static JWPFUNC jwpfunc[] = {
    { "Create New File",                IDM_FILENEW,            { 0, 0, 0, 0 }},
    { "Open File",                      IDM_FILEOPEN,           { 0, 0, 0, 0 }},
    { "Close File",                     IDM_FILECLOSE,          { 0, 0, 0, 0 }},
    { "Save File",                      IDM_FILESAVE,           { 0, 0, 0, 0 }},
    { "Save as Something Else",         IDM_FILESAVEAS,         { 0, 0, 0, 0 }},
    { "Find File",                      IDM_FILEFIND,           { 0, 0, 0, 0 }},
    { "Print File",                     IDM_FILEPRINT,          { 0, 0, 0, 0 }},
    { "Printer Setup",                  IDM_FILEPRINTER,        { 0, 0, 0, 0 }},
    { "Undo Last Change",               IDM_EDITUNDO,           { 0, 0, 0, 0 }},
    { "Redo Last Change",               IDM_EDITREDO,           { 0, 0, 0, 0 }},
    { "Cut Text",                       IDM_EDITCUT,            { 0, 0, 0, 0 }},
    { "Copy Text",                      IDM_EDITCOPY,           { 0, 0, 0, 0 }},
    { "Paste Text",                     IDM_EDITPASTE,          { 0, 0, 0, 0 }},
    { "Select All",                     IDM_EDITSELECTALL,      { 0, 0, 0, 0 }},
    { "Goto",                           IDM_EDITGOTO,           { 0, 0, 0, 0 }},
    { "Insert Page Break",              IDM_EDITPAGEBREAK,      { 0, 0, 0, 0 }},
    { "Insert Soft Return",             IDM_EDITSOFTRETURN,     { 0, 0, 0, 0 }},
    { "Search for Text",                IDM_EDITSEARCH,         { 0, 0, 0, 0 }},
    { "Continue Search/Replace",        IDM_CONTINUESEARCH,     { 0, 0, 0, 0 }},
    { "Replace Text",                   IDM_EDITREPLACE,        { 0, 0, 0, 0 }},
    { "Edit File Summary",              IDM_FILESUMMARY,        { 0, 0, 0, 0 }},
    { "Edit Header/Footer",             IDM_FILEHEADERS,        { 0, 0, 0, 0 }},
    { "Toggle Kana/Text mode",          IDM_TOGGLEMODE,         { 0, 0, 0, 0 }},
    { "Toggle ASCII / JASCII mode",     IDM_TOGGLEJASCII,       { 0, 0, 0, 0 }},
    { "Convert Kana to Kanji (L->R)",   IDM_KANJICONVERT,       { 0, 0, 0, 0 }},
    { "Convert Kana to Kanji (R->L)",   IDM_KANJILCONVERT,      { 0, 0, 0, 0 }},
    { "Toggle Input Line Editor",       IDM_VIEWIME,            { 0, 0, 0, 0 }},
    { "Toggle Active Input Window",     IDM_TOGGLEEDITWIN,      { 0, 0, 0, 0 }},
    { "Toggle Draft View",              IDM_VIEWDRAFT,          { 0, 0, 0, 0 }},
    { "Toggle Iconbar",                 IDM_VIEWRIBBON,         { 0, 0, 0, 0 }},
    { "Toggle Status Bar",              IDM_VIEWSPECIAL,        { 0, 0, 0, 0 }},
    { "Toggle File Ruler",              IDM_VIEWSPECIAL,        { 0, 0, 0, 0 }},
    { "Toggle Special Symbols",         IDM_VIEWSPECIAL,        { 0, 0, 0, 0 }},
    { "Toggle Dynamic Glossary",        IDM_TOGGLEGLOSSARY,     { 0, 0, 0, 0 }},
    { "Toggle Glossary List",           IDM_VIEWGLOSSARY,       { 0, 0, 0, 0 }},
    { "Kanji Lookup",                   IDM_KANJILOOKUP,        { 0, 0, 0, 0 }},
    { "Radical Lookup",                 IDM_RLOOKUP,            { 0, 0, 0, 0 }},
    { "Kanji Info",                     IDM_KANJIINFO,          { 0, 0, 0, 0 }},
    { "JIS Input",                      IDM_KANJIINPUT,         { 0, 0, 0, 0 }},
    { "JIS Table",                      IDM_KANJITABLE,         { 0, 0, 0, 0 }},
    { "Format Paragraph",               IDM_FORMATPARAGRAPH,    { 0, 0, 0, 0 }},
    { "Page Setup",                     IDM_FORMATPAGE,         { 0, 0, 0, 0 }},
    { "Repagenate File",                IDM_FORMATREPAGE,       { 0, 0, 0, 0 }},
    { "Dictionary",                     IDM_UTILITIESDICT,      { 0, 0, 0, 0 }},
    { "Edit Glossary",                  IDM_UTILITIESGLOSSARY,  { 0, 0, 0, 0 }},
    { "Edit User Conversions",          IDM_UTILITIESCONVERT,   { 0, 0, 0, 0 }},
    { "Change Hot-Keys",                IDM_UTILITIESHOTKEYS,   { 0, 0, 0, 0 }},
    { "Set Options",                    IDM_UTILITIESOPTIONS,   { 0, 0, 0, 0 }},
    { "Maintain Kanji Fonts",           IDM_UTILITIESFONTS,     { 0, 0, 0, 0 }},
    { "Show Statistics",                IDM_UTILITIESSTAT,      { 0, 0, 0, 0 }},
    { "Cascade File Windows",           IDM_WINDOWCASCADE,      { 0, 0, 0, 0 }},
    { "Tile File Windows",              IDM_WINDOWTILE,         { 0, 0, 0, 0 }},
    { "Close all File Windows",         IDM_WINDOWCLOSE,        { 0, 0, 0, 0 }},
    { "Help",                           IDM_HELPINDEX,          { 0, 0, 0, 0 }}
};

#define NRJWPFUNC   (sizeof(jwpfunc) / sizeof(JWPFUNC))

static HOTKEY HotKeysBackup[NRJWPFUNC];
BOOL OptionsChanged = FALSE;

static double far *PointSizes = NULL;

/* Option backup's */

static RECT glistsize, convbar;
static FILEFORMAT clipboard;
static BOOL DraftView, ShowSpecial, ShowRuler, NNConvert;
static BOOL ProgramMaximized;
static RECT ProgramPosition;
static int SysFontNum, BaseFontNum, PrintFontNum;
static char *AsciiFontName = NULL;
static double AsciiFontSize;
static MEASUREMENT Measure;
static PAGESETUP PageSetup;
static JUSTIFICATION LongJust, ShortJust;
static int UndoLevels, SaveUndoLevels;


WORD DecodeKeyStroke (UINT message, WPARAM wParam, BOOL shift, BOOL ctrl)
{
    int i;

    if (message == WM_KEYDOWN) {
        for (i = 0; i < NRJWPFUNC; i++) {
            if (jwpfunc[i].hotkey.alpha == 0 &&
                jwpfunc[i].hotkey.scancode == wParam &&
                jwpfunc[i].hotkey.shift == shift &&
                jwpfunc[i].hotkey.ctrl == ctrl)
                    return (jwpfunc[i].menuid);
        }
    } else if (message == WM_CHAR) {
        for (i = 0; i < NRJWPFUNC; i++) {
            if (jwpfunc[i].hotkey.alpha == wParam &&
                jwpfunc[i].hotkey.shift == shift)
                    return (jwpfunc[i].menuid);
        }
    }

    return (0);
}



static int ConvertKeyCode (HOTKEY key, char *buffer)
{
    if (key.alpha == 0) {
        buffer[0] = '\0';
        if (key.shift) strcat(buffer, "Shift+");
        if (key.ctrl) strcat(buffer, "Ctrl+");

        switch (key.scancode) {
            case VK_F1:  strcat(buffer, "F1"); return (FUNCSTART + 0);
            case VK_F2:  strcat(buffer, "F2"); return (FUNCSTART + 1);
            case VK_F3:  strcat(buffer, "F3"); return (FUNCSTART + 2);
            case VK_F4:  strcat(buffer, "F4"); return (FUNCSTART + 3);
            case VK_F5:  strcat(buffer, "F5"); return (FUNCSTART + 4);
            case VK_F6:  strcat(buffer, "F6"); return (FUNCSTART + 5);
            case VK_F7:  strcat(buffer, "F7"); return (FUNCSTART + 6);
            case VK_F8:  strcat(buffer, "F8"); return (FUNCSTART + 7);
            case VK_F9:  strcat(buffer, "F9"); return (FUNCSTART + 8);
            case VK_F10: strcat(buffer, "F10"); return (FUNCSTART + 9);
            case VK_F11: strcat(buffer, "F11"); return (FUNCSTART + 10);
            case VK_F12: strcat(buffer, "F12"); return (FUNCSTART + 11);
        }

        return (-1);
    }

    sprintf(buffer, "%sCtrl+%c", key.shift ? "Shift+" : "", key.alpha);

    return (ALPHASTART + (key.alpha - ' ' - 1));
}



static void IDToKeyCode (int id, HOTKEY *key)
{
    if (id >= FUNCSTART) {
        key->alpha = 0;
        switch (id - FUNCSTART) {
            case 0:  key->scancode = VK_F1; break;
            case 1:  key->scancode = VK_F2; break;
            case 2:  key->scancode = VK_F3; break;
            case 3:  key->scancode = VK_F4; break;
            case 4:  key->scancode = VK_F5; break;
            case 5:  key->scancode = VK_F6; break;
            case 6:  key->scancode = VK_F7; break;
            case 7:  key->scancode = VK_F8; break;
            case 8:  key->scancode = VK_F9; break;
            case 9:  key->scancode = VK_F10; break;
            case 10: key->scancode = VK_F11; break;
            case 11: key->scancode = VK_F12; break;
            default: key->scancode = VK_F1; break;
        }
    } else {
        key->alpha = ' ' + (id - ALPHASTART) + 1;
        key->scancode = 0;
    }
}


static BOOL HotKeyEqual (HOTKEY k1, HOTKEY k2)
{
    if (k1.alpha == 0) {
        return (k1.shift == k2.shift &&
                k1.ctrl == k2.ctrl &&
                k1.scancode == k2.scancode);
    } else {
        return (k1.alpha == k2.alpha &&
                k1.shift == k2.shift);
    }
}



void ProcessMenuTexts (void)
{
    int i, j;
    WORD menuid;
    char buffer[MAXLINELEN], keytext[50];

    for (i = 0; i < NRJWPFUNC; i++) {
        menuid = jwpfunc[i].menuid;
        GetMenuString(hmenu, menuid, buffer, MAXLINELEN, MF_BYCOMMAND);

        for (j = 0; buffer[j] && buffer[j] != '\t'; j++);
        buffer[j] = '\0';

        if (ConvertKeyCode(jwpfunc[i].hotkey, keytext) >= 0) {
            strcat(buffer, "\t");
            strcat(buffer, keytext);
        } else if (jwpfunc[i].menuid == IDM_EDITUNDO) {
            strcat(buffer, "\tAlt+Backspace");
        } else if (jwpfunc[i].menuid == IDM_EDITREDO) {
            strcat(buffer, "\tShift+Alt+Backspace");
        } else if (jwpfunc[i].menuid == IDM_EDITCUT) {
            strcat(buffer, "\tShift+Del");
        } else if (jwpfunc[i].menuid == IDM_EDITCOPY) {
            strcat(buffer, "\tCtrl+Ins");
        } else if (jwpfunc[i].menuid == IDM_EDITPASTE) {
            strcat(buffer, "\tShift+Ins");
        } else if (jwpfunc[i].menuid == IDM_EDITPAGEBREAK) {
            strcat(buffer, "\tCtrl+Enter");
        } else if (jwpfunc[i].menuid == IDM_EDITSOFTRETURN) {
            strcat(buffer, "\tShift+Enter");
        }

        j = GetMenuState(hmenu, menuid, MF_BYCOMMAND);
        j |= (MF_BYCOMMAND | MF_STRING);

        ModifyMenu(hmenu, menuid, j, menuid, buffer);
    }

    /* Now check the menu items */

    if (global.draftview)
        CheckMenuItem(hmenu, IDM_VIEWDRAFT, MF_CHECKED | MF_BYCOMMAND);
    if (global.showruler)
        CheckMenuItem(hmenu, IDM_VIEWRULER, MF_CHECKED | MF_BYCOMMAND);
    if (global.tbhwnd != NULL)
        CheckMenuItem(hmenu, IDM_VIEWRIBBON, MF_CHECKED | MF_BYCOMMAND);
    if (global.statushwnd != NULL)
        CheckMenuItem(hmenu, IDM_VIEWSTATUS, MF_CHECKED | MF_BYCOMMAND);
    if (global.glisthwnd != NULL)
        CheckMenuItem(hmenu, IDM_VIEWGLOSSARY, MF_CHECKED | MF_BYCOMMAND);
    if (global.showspecial)
        CheckMenuItem(hmenu, IDM_VIEWSPECIAL, MF_CHECKED | MF_BYCOMMAND);
    if (global.imehwnd != NULL)
        CheckMenuItem(hmenu, IDM_VIEWIME, MF_CHECKED | MF_BYCOMMAND);
}


static void PaintKeyButton(HWND hwnd, HDC hdc, int id, BOOL ThreeD)
{
    int width, height;
    DWORD dimensions;
    RECT rect;
    char buf[5];

    GetClientRect(hwnd, &rect);

    Create3DEffect (hdc, &rect, ThreeD ? 2 : 0, 0);
    SetBkMode(hdc, TRANSPARENT);

#ifdef USEWIN32
    {
        SIZE tsize;

        if (id >= FUNCSTART) {
            sprintf(buf, "F%d", id - FUNCSTART + 1);
            GetTextExtentPoint32(hdc, buf, strlen(buf), &tsize);
        } else {
            sprintf(buf, "%c", ' ' + (id - ALPHASTART) + 1);
            GetTextExtentPoint32(hdc, buf, 1, &tsize);
        }
        dimensions = MAKELONG((WORD) tsize.cx, (WORD) tsize.cy);
    }
#else
    if (id >= FUNCSTART) {
        sprintf(buf, "F%d", id - FUNCSTART + 1);
        dimensions = GetTextExtent(hdc, buf, strlen(buf));
    } else {
        sprintf(buf, "%c", ' ' + (id - ALPHASTART) + 1);
        dimensions = GetTextExtent(hdc, buf, 1);
    }
#endif USEWIN32

    width = LOWORD(dimensions);
    height = HIWORD(dimensions);

    TextOut(hdc, (rect.right - width) / 2, (rect.bottom - height) / 2,
            buf, strlen(buf));

    if (id == Selected)
        PatBlt(hdc, 1, 1, rect.right - 2, rect.bottom - 2, DSTINVERT);
}



LONG CALLBACK HotKeyChoiceProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_LBUTTONDOWN: {
            int i;
            HWND CurSelectedHwnd;

            i = GetDlgCtrlID(hwnd);
            if (i == Selected) return (0);

            CurSelectedHwnd = GetDlgItem(GetParent(hwnd), Selected);
            SendMessage(hwnd, WM_USER, 0, 0L);
            if (CurSelectedHwnd != NULL) SendMessage(CurSelectedHwnd, WM_USER, 0, 0L);

            Selected = i;
            IDToKeyCode(i, &CurrentKey);
            EnableWindow(GetDlgItem(GetParent(hwnd), 4211), (CurrentKey.alpha == 0));
            //EnableWindow(GetDlgItem(GetParent(hwnd), 4212), (CurrentKey.alpha == 0));

            SendMessage(GetParent(hwnd), WM_USER, 0, 0L);
            return (0);
        }

        case WM_PAINT: {
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);
            PaintKeyButton(hwnd, hdc, GetDlgCtrlID(hwnd), TRUE);
            EndPaint(hwnd, &ps);
            return (0);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_USER: {
            RECT rect;
            HDC hdc;

            GetClientRect(hwnd, &rect);
            hdc = GetDC(hwnd);
            PatBlt(hdc, 1, 1, rect.right - 2, rect.bottom - 2, DSTINVERT);
            ReleaseDC(hwnd, hdc);
            return (TRUE);
        }


        default: break;
    }

    return (CallWindowProc(OldStaticProc, hwnd, message, wParam, lParam));
}



BOOL CALLBACK ChangeHotKeyProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int i;
    POINT point;
    HWND child;
    char buffer[MAXLINELEN], keytext[50];

    switch (message) {
        case WM_INITDIALOG:
            /* Find a non-empty button */

            for (i = 0; i < NRCTRLKEYS; i++) {
                child = GetDlgItem(hwnd, ALPHASTART + i);
                if (child != NULL) break;
            }
            if (child == NULL) {
                for (i = 0; i < NRFUNCKEYS; i++) {
                    child = GetDlgItem(hwnd, FUNCSTART + i);
                    if (child != NULL) break;
                }
            }
            if (child == NULL) {    /* None! */
                EndDialog(hwnd, FALSE);
                return (TRUE);
            }

            OldStaticProc = (FARPROC) GetWindowLong(child, GWL_WNDPROC);

            for (i = 0; i < NRCTRLKEYS; i++) {
                if (strchr(RESERVEDCTRLCHAR, ' ' + i + 1) != NULL) continue;
                child = GetDlgItem(hwnd, ALPHASTART + i);
                if (child == NULL) continue;
                SetWindowLong(child, GWL_WNDPROC, (LONG) HotKeyChoiceProc);
            }

            for (i = 0; i < NRFUNCKEYS; i++) {
                child = GetDlgItem(hwnd, FUNCSTART + i);
                if (child == NULL) continue;
                SetWindowLong(child, GWL_WNDPROC, (LONG) HotKeyChoiceProc);
            }

            Selected = ConvertKeyCode(OldKey, keytext);
            if (GetDlgItem(hwnd, Selected) == NULL) Selected = -1;
            CurrentKey = OldKey;

            sprintf(buffer, "Hot-Key for \"%s\"", jwpfunc[CurrentFunction].desc);
            SetDlgItemText(hwnd, 4201, buffer);

            if (Selected >= 0) {
                sprintf(buffer, "Old Hot-Key:\t%s", keytext);
                SetDlgItemText(hwnd, 4202, buffer);
                sprintf(buffer, "New Hot-Key:\t%s", keytext);
                SetDlgItemText(hwnd, 4203, buffer);
            } else {
                SetDlgItemText(hwnd, 4202, "Old Hot-Key:\tNone");
                SetDlgItemText(hwnd, 4203, "New Hot-Key:\tNone");
            }

            CheckDlgButton(hwnd, 4211, OldKey.ctrl);
            CheckDlgButton(hwnd, 4212, OldKey.shift);

            EnableWindow(GetDlgItem(hwnd, 4211), (CurrentKey.alpha == 0));
            //EnableWindow(GetDlgItem(hwnd, 4212), (CurrentKey.alpha == 0));

            TrapHelpInDialogBoxes(hwnd, "HotKeysEditDlg", 1, TRUE);
            CenterDialogBox(hwnd);

            return (TRUE);

        case WM_USER:
            if (ConvertKeyCode(CurrentKey, keytext) >= 0) {
                sprintf(buffer, "New Hot-Key:\t%s", keytext);
                SetDlgItemText(hwnd, 4203, buffer);
            } else {
                SetDlgItemText(hwnd, 4203, "New Hot-Key:\tNone");
            }
            return (TRUE);

        case WM_LBUTTONDOWN:

            point.x = LOWORD(lParam);
            point.y = HIWORD(lParam);

            child = ChildWindowFromPoint(hwnd, point);

            if (child == hwnd || child == NULL) return (TRUE);

            SendMessage(child, message, wParam, lParam);
            return (TRUE);

        case WM_PAINT: {
            HDC hdc;
            PAINTSTRUCT ps;

			(void) BeginPaint(hwnd, &ps);

            for (i = 0; i < NRCTRLKEYS; i++) {
				child = GetDlgItem(hwnd, ALPHASTART + i);
				if (child == NULL) continue;

				hdc = GetDC(child);
                PaintKeyButton(child, hdc, ALPHASTART + i,
                                (strchr(RESERVEDCTRLCHAR, ' ' + i + 1) == NULL));
                ReleaseDC(child, hdc);
            }

            for (i = 0; i < NRFUNCKEYS; i++) {
                child = GetDlgItem(hwnd, FUNCSTART + i);
				if (child == NULL) continue;

				hdc = GetDC(child);
				PaintKeyButton(child, hdc, FUNCSTART + i, TRUE);
				ReleaseDC(child, hdc);
			}

			EndPaint(hwnd, &ps);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4211:      /* Check boxes */
                case 4212:
                    CurrentKey.ctrl = IsDlgButtonChecked(hwnd, 4211);
                    CurrentKey.shift = IsDlgButtonChecked(hwnd, 4212);

                    if (ConvertKeyCode(CurrentKey, keytext) >= 0) {
                        sprintf(buffer, "New Hot-Key:\t%s", keytext);
                        SetDlgItemText(hwnd, 4203, buffer);
                    }
                    return (TRUE);

                case IDOK:
                    if (CurrentKey.alpha != 0 && strchr(RESERVEDCTRLCHAR, CurrentKey.alpha) != NULL) {
                        ConvertKeyCode(CurrentKey, keytext);
                        ErrorMessage(hwnd, "You cannot assign %s to anything", keytext);
                        return (TRUE);
                    }

                    for (i = 0; i < NRJWPFUNC; i++) {
                        if (i == CurrentFunction) continue;
                        if (HotKeyEqual(CurrentKey, jwpfunc[i].hotkey)) break;
                    }

                    if (i < NRJWPFUNC) {
                        if (ConvertKeyCode(CurrentKey, keytext) >= 0) {
                            ErrorMessage(hwnd, "The key %s is already assigned to\n"
                                                  "\"%s\"", keytext, jwpfunc[i].desc);
                            return (TRUE);
                        }
                    }

                    if (CurrentKey.alpha != 0) CurrentKey.ctrl = TRUE;

                    jwpfunc[CurrentFunction].hotkey = CurrentKey;

                    OptionsChanged = TRUE;

                    ProcessMenuTexts();

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
    }

    return (FALSE);
}



BOOL CALLBACK HotKeysProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int i;

    switch (message) {
        case WM_INITDIALOG: {
            char buffer[MAXLINELEN];

            for (i = 0; i < NRJWPFUNC; i++) {
                SendDlgItemMessage(hwnd, 4201, LB_ADDSTRING, 0, (LONG) jwpfunc[i].desc);
                HotKeysBackup[i] = jwpfunc[i].hotkey;
            }

            SendDlgItemMessage(hwnd, 4201, LB_SETCURSEL, 0, 0L);

            if (ConvertKeyCode(jwpfunc[0].hotkey, buffer) >= 0) {
                SetDlgItemText(hwnd, 4202, buffer);
            } else {
                SetDlgItemText(hwnd, 4202, "None");
            }

            TrapHelpInDialogBoxes(hwnd, "HotKeysDlg", 1, TRUE);
            CenterDialogBox(hwnd);

            return (TRUE);
        }

        case WM_KEYDOWN:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDCANCEL, 0L);
                                return (TRUE);
            }
            break;

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4201:      /* List Box */
                    switch (NOTIFYCODE) {
                        case LBN_SELCHANGE: {
                            char buffer[MAXLINELEN];

                            i = SendDlgItemMessage(hwnd, 4201, LB_GETCURSEL, 0, 0L);
                            if (i == LB_ERR) {
                                MessageBeep(MB_ICONEXCLAMATION);
                                return (TRUE);
                            }
                            if (ConvertKeyCode(jwpfunc[i].hotkey, buffer) >= 0) {
                                SetDlgItemText(hwnd, 4202, buffer);
                            } else {
                                SetDlgItemText(hwnd, 4202, "None");
                            }
                            return (TRUE);
                        }

                        case LBN_DBLCLK:
                            i = SendDlgItemMessage(hwnd, 4201, LB_GETCURSEL, 0, 0L);
                            if (i == LB_ERR) {
                                MessageBeep(MB_ICONEXCLAMATION);
                                return (TRUE);
                            }
                            SendMessage(hwnd, WM_COMMAND, 4211, 0L);
                            return (TRUE);
                    }
                    return (TRUE);

                case 4211: {    /* Change */
                    char buffer[MAXLINELEN];

                    i = SendDlgItemMessage(hwnd, 4201, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) {
                        MessageBeep(MB_ICONEXCLAMATION);
                        return (TRUE);
                    }
                    CurrentFunction = i;
                    OldKey = CurrentKey = jwpfunc[i].hotkey;

                    DialogBox (hInstance, "ChangeHotKey", hwnd, ChangeHotKeyProc);

                    if (ConvertKeyCode(jwpfunc[i].hotkey, buffer) >= 0) {
                        SetDlgItemText(hwnd, 4202, buffer);
                    } else {
                        SetDlgItemText(hwnd, 4202, "None");
                    }

                    return (TRUE);
                }

                case 4212:      /* Delete */
                    i = SendDlgItemMessage(hwnd, 4201, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) {
                        MessageBeep(MB_ICONEXCLAMATION);
                        return (TRUE);
                    }
                    CurrentKey.shift = CurrentKey.ctrl = FALSE;
                    CurrentKey.alpha = CurrentKey.scancode = 0;
                    jwpfunc[i].hotkey = CurrentKey;
                    SetDlgItemText(hwnd, 4202, "None");
                    OptionsChanged = TRUE;
                    return (TRUE);

                case IDCANCEL:
                    for (i = 0; i < NRJWPFUNC; i++) jwpfunc[i].hotkey = HotKeysBackup[i];
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                case IDOK:
                    for (i = 0; i < NRJWPFUNC; i++) {
                        if (HotKeysBackup[i].alpha != jwpfunc[i].hotkey.alpha)
                            OptionsChanged = TRUE;

                        if (jwpfunc[i].hotkey.alpha == 0) {
                            if (jwpfunc[i].hotkey.shift != HotKeysBackup[i].shift ||
                                jwpfunc[i].hotkey.ctrl != HotKeysBackup[i].ctrl ||
                                jwpfunc[i].hotkey.scancode != HotKeysBackup[i].scancode)
                                    OptionsChanged = TRUE;
                        } else {
                            if (jwpfunc[i].hotkey.shift != HotKeysBackup[i].shift)
                                OptionsChanged = TRUE;
                        }
                    }

                    ProcessMenuTexts();

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
    }

    return (FALSE);
}



char *DupString (char *s)
{
    char *cp;

    cp = MemAlloc(strlen(s) + 5);

    if (cp != NULL) strcpy(cp, s);

    return (cp);
}



#ifdef USEWIN32

static BOOL ReadOptionString(char *section, char *option, char *dfvalue, char *buffer, int max, char *filename)
{
    DWORD size;
    HKEY hsubkey;
    char subkey[MAXLINELEN];

    strcpy(buffer, dfvalue);
    sprintf(subkey, "Software\\" PROGNAME "\\Options\\%s", section);
    if (RegOpenKeyEx(HKEY_CURRENT_USER, subkey, 0, KEY_QUERY_VALUE, &hsubkey) != ERROR_SUCCESS)
        goto UseProfile;

    size = MAXLINELEN;

    if (RegQueryValueEx(hsubkey, option, NULL, NULL, buffer, &size) != ERROR_SUCCESS) {
        RegCloseKey(hsubkey);
        goto UseProfile;
    } else {
        RegCloseKey(hsubkey);
        return (TRUE);
    }

UseProfile:
    if (filename[0]) {
        GetPrivateProfileString(section, option, dfvalue, buffer, max, filename);
    } else {
        strcpy(buffer, dfvalue);
    }
    return (strcmp(buffer, dfvalue) == 0 ? FALSE : TRUE);
}


static int ReadOptionInt(char *section, char *option, int dfvalue, char *filename)
{
    char buffer[MAXLINELEN];

    if (ReadOptionString(section, option, "", buffer, MAXLINELEN, filename)) {
        return (atoi(buffer));
    } else {
        return (dfvalue);
    }
}

#else

    #define ReadOptionString    GetPrivateProfileString
    #define ReadOptionInt       GetPrivateProfileInt

#endif USEWIN32



BOOL ReadProgramSizeOptions (int *x, int *y, int *width, int *length)
{
    char sizebuf[MAXLINELEN];
    char posbuf[MAXLINELEN];

    ReadOptionString("Others", "Program Size", "500 400", sizebuf, MAXLINELEN, PROGNAME ".INI");
    ReadOptionString("Others", "Program Position", "Maximized", posbuf, MAXLINELEN, PROGNAME ".INI");

    sscanf(sizebuf, "%d %d", width, length);
    ProgramPosition.right = *width;
    ProgramPosition.bottom = *length;

    if (!stricmp(posbuf, "Maximized")) {
        ProgramMaximized = TRUE;
        *x = *y = 0;
        return (TRUE);
    } else {
        ProgramMaximized = TRUE;
        sscanf(posbuf, "%d %d", x, y);
        ProgramPosition.left = *x;
        ProgramPosition.top = *y;
        return (FALSE);
    }
}



BOOL Read3DDialogOption (void)
{
    char buffer[MAXLINELEN];

    ReadOptionString("Others", "3D Dialog Boxes", "Yes", buffer, MAXLINELEN, PROGNAME ".INI");
    return (buffer[0] == 'Y' || buffer[0] == 'y');
}



BOOL ReadOptionsFile (void)
{
    int i, j, x;
    HFILE fd;
    unsigned char ch;
    OFSTRUCT of;
    HOTKEY hotkey;
    char buffer[MAXLINELEN];
    char keyname[50];
    char OptionsFileName[MAXFILENAMELEN];


    if (!CreateOptionsFile(OptionsFileName)) return (FALSE);


    /* Path names */

    ReadOptionString("File Names", PROGNAME " Directory", "C:\\" PROGNAME "\\", buffer, MAXLINELEN, OptionsFileName);
    i = strlen(buffer);
    if (i > 0 && buffer[i-1] != '\\') strcat(buffer, "\\");
    global.jwppath = DupString(buffer);

    {
        char buffer1[MAXLINELEN];

        sprintf(buffer1, "%stemplate\\", global.jwppath);
        ReadOptionString("File Names", "Templates", buffer1, buffer, MAXLINELEN, OptionsFileName);
        i = strlen(buffer);
        if (i > 0 && buffer[i-1] != '\\') strcat(buffer, "\\");
        global.templatepath = DupString(buffer);
    }

    ReadOptionString("File Names", "Help File", PROGNAME ".HLP", buffer, MAXLINELEN, OptionsFileName);
    global.help = DupString(buffer);
    ReadOptionString("File Names", "Conversion Dictionary", "KANJIDCT.DCT", buffer, MAXLINELEN, OptionsFileName);
    global.convdict = DupString(buffer);
    ReadOptionString("File Names", "Conversion Dictionary Index", "KANJIDCT.IDX", buffer, MAXLINELEN, OptionsFileName);
    global.convidx = DupString(buffer);
    ReadOptionString("File Names", "User Conversion Dictionary", "USERDICT.DCT", buffer, MAXLINELEN, OptionsFileName);
    global.userdict = DupString(buffer);
    ReadOptionString("File Names", "Conversion Cache", "CNVCACHE.SAV", buffer, MAXLINELEN, OptionsFileName);
    global.convcache = DupString(buffer);
    ReadOptionString("File Names", "Kanji Info", "KINFO.DAT", buffer, MAXLINELEN, OptionsFileName);
    global.kinfo = DupString(buffer);
    ReadOptionString("File Names", "Kanji Info Index (Bushu)", "KINFOBSH.IDX", buffer, MAXLINELEN, OptionsFileName);
    global.kinfobshidx = DupString(buffer);
    ReadOptionString("File Names", "Kanji Info Index (Stroke)", "KINFOSTR.IDX", buffer, MAXLINELEN, OptionsFileName);
    global.kinfostridx = DupString(buffer);

    ReadOptionString("File Names", "Radical Lookup Database (Radical)", "RADKANJI.DAT", buffer, MAXLINELEN, OptionsFileName);
    global.radkanjidat = DupString(buffer);
    ReadOptionString("File Names", "Radical Lookup Database Index (Radical)", "RADKANJI.IDX", buffer, MAXLINELEN, OptionsFileName);
    global.radkanjiidx = DupString(buffer);
    ReadOptionString("File Names", "Radical Lookup Database (Stroke)", "STROKNJI.DAT", buffer, MAXLINELEN, OptionsFileName);
    global.stroknjidat = DupString(buffer);
    ReadOptionString("File Names", "Radical Lookup Database Index (Stroke)", "STROKNJI.IDX", buffer, MAXLINELEN, OptionsFileName);
    global.stroknjiidx = DupString(buffer);

    ReadOptionString("File Names", "Global Glossary", "GLOSSARY.DAT", buffer, MAXLINELEN, OptionsFileName);
    global.glossary = DupString(buffer);

    /* Page Setup */

    ReadOptionString("Page Setup", "Measurement Unit", "inches", buffer, MAXLINELEN, OptionsFileName);
    global.measure = Measure = FindMeasurement(buffer, -1);
    ReadOptionString("Page Setup", "Margins", "1.0  1.0  1.0  1.0", buffer, MAXLINELEN, OptionsFileName);
    sscanf(buffer, "%lf %lf %lf %lf", &(global.pagesetup.margins[0]), &(global.pagesetup.margins[1]),
                &(global.pagesetup.margins[2]), &(global.pagesetup.margins[3]));

    ReadOptionString("Page Setup", "Long Text Justification", "Justify", buffer, MAXLINELEN, OptionsFileName);
    if (!stricmp(buffer, "Justify")) global.longjust = J_JUSTIFY;
    else if (!strcmp(buffer, "Left")) global.longjust = J_LEFT;
    else global.longjust = J_JUSTIFY;

    LongJust = global.longjust;

    ReadOptionString("Page Setup", "Short Text Justification", "Center", buffer, MAXLINELEN, OptionsFileName);
    if (!stricmp(buffer, "Justify")) global.shortjust = J_JUSTIFY;
    else if (!stricmp(buffer, "Left")) global.shortjust = J_LEFT;
    else if (!stricmp(buffer, "Right")) global.shortjust = J_RIGHT;
    else if (!stricmp(buffer, "Center")) global.shortjust = J_CENTER;
    else global.shortjust = J_CENTER;

    ShortJust = global.shortjust;

    PageSetup = global.pagesetup;


    /* ASCII Fonts */

    ReadOptionString("Base ASCII Font", "Face Name", "Times New Roman", buffer, MAXLINELEN, OptionsFileName);
    DefAsciiFont.facename = DupString(buffer);
    AsciiFontName = DupString(buffer);

    ReadOptionString("Base ASCII Font", "Size", "12", buffer, MAXLINELEN, OptionsFileName);
    AsciiFontSize = DefAsciiFont.size = atof(buffer);

    /* Installed Kanji Fonts */

    global.nr_fonts = ReadOptionInt("Kanji Fonts", "Number of Kanji Fonts", 0, OptionsFileName);

    SysFontNum = ReadOptionInt("Kanji Fonts", "System Font", 1, OptionsFileName) - 1;
    BaseFontNum = ReadOptionInt("Kanji Fonts", "Default Base Font", 1, OptionsFileName) - 1;
    PrintFontNum = ReadOptionInt("Kanji Fonts", "Default Print Font", 1, OptionsFileName) - 1;

    InstalledFonts = MemAlloc(global.nr_fonts * sizeof(KANJIFONT));

    PRINTFONT = SYSFONT = BASEFONT = NULL;

    for (i = j = 0; i < global.nr_fonts; i++) {
        sprintf(keyname, "Font #%d", i+1);
        ReadOptionString("Kanji Fonts", keyname, "", buffer, MAXLINELEN, OptionsFileName);

        if (buffer[0]) {
            fd = OpenFont(buffer, &(InstalledFonts[j]));
            if (fd != HFILE_ERROR) {
                _lclose(fd);
                if (PrintFontNum == i) {
                    PRINTFONT = &(InstalledFonts[j]);
                    PrintFontNum = j;
                }
                if (SysFontNum == i) {
                    SYSFONT = &(InstalledFonts[j]);
                    SysFontNum = j;
                }
                if (BaseFontNum == i) {
                    BASEFONT = &(InstalledFonts[j]);
                    BaseFontNum = j;
                }
                j++;
            }
        }
    }


    global.nr_fonts = j;

    if (global.nr_fonts <= 0) {
        ErrorMessage(global.hwnd, "No Installed fonts!");
        return (FALSE);                               /* Cannot go on */
    }

    if (PRINTFONT == NULL) {
        PRINTFONT = &(InstalledFonts[j - 1]);
        PrintFontNum = j - 1;
    }
    if (SYSFONT == NULL) {
        SYSFONT = &(InstalledFonts[0]);
        SysFontNum = 0;
    }
    if (BASEFONT == NULL) {
        BASEFONT = &(InstalledFonts[0]);
        BaseFontNum = 0;
    }

    if (!ReadOptionString("Kanji Fonts", "Default Print Size", "12.0", buffer, MAXLINELEN, OptionsFileName)) {
        double printscale;

        ReadOptionString("Kanji Fonts", "Default Print Scale", "1.0000", buffer, MAXLINELEN, OptionsFileName);
        sscanf(buffer, "%lf", &printscale);
        global.printscale = ((printscale * (double) PRINTFONT->height) * 72.0)/((double) global.resolution.y);
    } else {
        sscanf(buffer, "%lf", &(global.printscale));
    }

    /* Force even number spacing */

    if (SYSFONT->leading % 2 != 0) SYSFONT->leading++;
    if (SYSFONT->spacing % 2 != 0) SYSFONT->spacing++;

    /* Glossary Options */

    ReadOptionString("Glossary", "Dynamic Glossary", "Yes", buffer, MAXLINELEN, OptionsFileName);
    global.dynamicglossary = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Glossary", "Glossary List Disappears", "No", buffer, MAXLINELEN, OptionsFileName);
    global.glistbehaviour = (buffer[0] == 'Y' || buffer[0] == 'y') ? 1 : 0;
    ReadOptionString("Glossary", "Glossary List Visible", "No", buffer, MAXLINELEN, OptionsFileName);
    global.glistvisible = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Glossary", "Glossary List Position", "-1 -1", buffer, MAXLINELEN, OptionsFileName);
    sscanf(buffer, "%d %d", &global.glistsize.left, &global.glistsize.top);
    ReadOptionString("Glossary", "Glossary List Size", "-1 -1", buffer, MAXLINELEN, OptionsFileName);
    sscanf(buffer, "%d %d", &global.glistsize.right, &global.glistsize.bottom);

    glistsize = global.glistsize;


    /* Undo Options */

    global.undolevels = ReadOptionInt("Undo", "Number of Undo Levels", DEFUNDOLEVELS, OptionsFileName);
    UndoLevels = global.undolevels;
    ReadOptionString("Undo", "Undo Levels Saved", "Yes", buffer, MAXLINELEN, OptionsFileName);
    if (!stricmp(buffer, "All")) global.saveundolevels = -1;
    else {
        sscanf(buffer, "%d", &global.saveundolevels);
    }
    SaveUndoLevels = global.undolevels;


    /* Other Options */

    ReadOptionString("Others", "Show Input Line Editor", "No", buffer, MAXLINELEN, OptionsFileName);
    global.showime = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Relax Margin for Punctuation", "Yes", buffer, MAXLINELEN, OptionsFileName);
    global.relaxmargin = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Dynamic Conversion", "No", buffer, MAXLINELEN, OptionsFileName);
    global.dynamicconvert = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Conversion Bar Position", "Bottom", buffer, MAXLINELEN, OptionsFileName);

    if (!stricmp(buffer, "Top")) global.listposition = 1;
    else if (!stricmp(buffer, "Bottom")) global.listposition = 2;
    else {
        global.listposition = 0;
        sscanf(buffer, "%d %d", &global.convbar.left, &global.convbar.top);
    }

    ReadOptionString("Others", "Conversion Bar Size", "-1 -1", buffer, MAXLINELEN, OptionsFileName);
    sscanf(buffer, "%d %d", &global.convbar.right, &global.convbar.bottom);

    convbar = global.convbar;

    ReadOptionString("Others", "Clipboard Format", "EUC", buffer, MAXLINELEN, OptionsFileName);
    for (i = FF_EUC; FileFormatNames[i] != NULL; i++) {
        if (!strcmp(FileFormatNames[i], buffer)) break;
    }
    if (FileFormatNames[i] != NULL) global.clipboard = i; else global.clipboard = FF_EUC;
    clipboard = global.clipboard;

    ReadOptionString("Others", "Default Open Format", "Normal", buffer, MAXLINELEN, OptionsFileName);
    for (i = FF_UNKNOWN; FileFormatNames[i] != NULL; i++) {
        if (!strcmp(FileFormatNames[i], buffer)) break;
    }
    if (FileFormatNames[i] != NULL) global.openformat = i; else global.openformat = FF_NORMAL;

    ReadOptionString("Others", "Draft View", "No", buffer, MAXLINELEN, OptionsFileName);
    DraftView = global.draftview = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Show Special Symbols", "No", buffer, MAXLINELEN, OptionsFileName);
    ShowSpecial = global.showspecial = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Show Iconbar", "Yes", buffer, MAXLINELEN, OptionsFileName);
    global.showribbon = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Show Status Bar", "Yes", buffer, MAXLINELEN, OptionsFileName);
    global.showstatus = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "Show File Ruler", "Yes", buffer, MAXLINELEN, OptionsFileName);
    ShowRuler = global.showruler = (buffer[0] == 'Y' || buffer[0] == 'y');
    ReadOptionString("Others", "3D Dialog Boxes", "Yes", buffer, MAXLINELEN, OptionsFileName);
    Dialogs3D = global.ctl3d = (buffer[0] == 'Y' || buffer[0] == 'y');
    NNConvert = global.nnconvert = TRUE;

    ReadOptionString("Others", "Automatic Radical Lookup", "No", buffer, MAXLINELEN, OptionsFileName);
    global.autorlookup = (buffer[0] == 'Y' || buffer[0] == 'y');
    global.rlookupsort = ReadOptionInt("Others", "Radical Lookup Sort Method", 1, OptionsFileName);
    if (global.rlookupsort < 0 || global.rlookupsort > 3) global.rlookupsort = 1;

    ReadOptionString("Others", "Automatic Repagenation", "No", buffer, MAXLINELEN, OptionsFileName);
    global.autorepage = (buffer[0] == 'Y' || buffer[0] == 'y');

    /* Quick files */

    j = ReadOptionInt("Quick Files", "Number of Quick Files", 0, OptionsFileName);

    PutQuickFile(NULL);     /* Clear all the quick files first */

    for (i = j; i > 0; i--) {
        sprintf(keyname, "File%d", i);
        ReadOptionString("Quick Files", keyname, "", buffer, MAXLINELEN, OptionsFileName);
        if (buffer[0]) PutQuickFile(buffer);
    }

    /* Dictionary */

    j = ReadOptionInt("Dictionary", "Number of Dictionaries", -1, OptionsFileName);

    if (j == 0) {
        global.jdict = (char * *) MemAlloc(sizeof(char *) + 5);
        global.jdictidx = (char * *) MemAlloc(sizeof(char *) + 5);
        global.jdictdesc = (BYTE *) MemAlloc(sizeof(BYTE *) + 5);
        global.jdictactive = (BOOL *) MemAlloc(sizeof(BOOL) + 5);
        global.nr_jdict = 0;
    } else if (j > 0) {
        global.jdict = (char * *) MemAlloc(j * sizeof(char *) + 5);
        global.jdictidx = (char * *) MemAlloc(j * sizeof(char *) + 5);
        global.jdictdesc = (BYTE *) MemAlloc(j * sizeof(BYTE *) + 5);
        global.jdictactive = (BOOL *) MemAlloc(j * sizeof(BOOL) + 5);

        for (i = x = 0; i < j; i++) {
            char *p;

            sprintf(keyname, "Dictionary #%d", i+1);
            ReadOptionString("Dictionary", keyname, "", buffer, MAXLINELEN, OptionsFileName);
            if (!buffer[0]) continue;

            if ((p = strtok(buffer, "|")) == NULL) continue;
            global.jdict[x] = DupString(p);

            if ((p = strtok(NULL, "|")) == NULL) continue;
            global.jdictidx[x] = DupString(p);

            if ((p = strtok(NULL, "|")) == NULL) {
                global.jdictactive[x] = FALSE;
            } else {
                global.jdictactive[x] = (p[0] == 'Y' || p[0] == 'y');
            }

            if ((p = strtok(NULL, "|")) == NULL) {
                global.jdictdesc[x] = DupString(global.jdict[x]);
            } else {
                global.jdictdesc[x] = DupString(p);
            }

            x++;
        }
        global.nr_jdict = x;
    } else {
        global.jdict = (char * *) MemAlloc(sizeof(char *) + 5);
        global.jdictidx = (char * *) MemAlloc(sizeof(char *) + 5);
        global.jdictdesc = (BYTE *) MemAlloc(sizeof(BYTE *) + 5);
        global.jdictactive = (BOOL *) MemAlloc(sizeof(BOOL) + 5);

        ReadOptionString("File Names", "Japanese Dictionary", "", buffer, MAXLINELEN, OptionsFileName);
        global.jdict[0] = DupString(buffer);
        ReadOptionString("File Names", "Japanese Dictionary Index", "EDICT.JDX", buffer, MAXLINELEN, OptionsFileName);
        global.jdictidx[0] = DupString(buffer);
        global.jdictdesc[0] = DupString("EDICT: Jim Breen's Japanese-English Dictionary");
        global.jdictactive[0] = TRUE;
        global.nr_jdict = 1;
    }

    j = ReadOptionInt("Dictionary", "Number of Keywords", 0, OptionsFileName);

    if (j > 0) {
        global.dictkeywords = (BYTE * *) MemAlloc(j * sizeof(BYTE *) + 5);
        global.dictkeywordsdesc = (BYTE * *) MemAlloc(j * sizeof(BYTE *) + 5);
        global.dictkeywordsshow = (BOOL *) MemAlloc(j * sizeof(BOOL) + 5);

        for (i = x = 0; i < j; i++) {
            char *p;

            sprintf(keyname, "Key #%d", i+1);
            ReadOptionString("Dictionary", keyname, "", buffer, MAXLINELEN, OptionsFileName);
            if (!buffer[0]) continue;

            if ((p = strtok(buffer, "|")) == NULL) continue;
            global.dictkeywords[x] = DupString(p);

            if ((p = strtok(NULL, "|")) == NULL) {
                global.dictkeywordsshow[x] = FALSE;
            } else {
                global.dictkeywordsshow[x] = (p[0] == 'Y' || p[0] == 'y');
            }

            if ((p = strtok(NULL, "|")) == NULL) {
                global.dictkeywordsdesc[x] = DupString(global.dictkeywords[x]);
            } else {
                global.dictkeywordsdesc[x] = DupString(p);
            }

            x++;
        }
        global.nr_dictkeywords = x;
    } else {
        for (j = 0; FixedDictKeywords[j][0] != NULL; j++);
        global.nr_dictkeywords = j;
        global.dictkeywords = (BYTE * *) MemAlloc(j * sizeof(BYTE *) + 5);
        global.dictkeywordsdesc = (BYTE * *) MemAlloc(j * sizeof(BYTE *) + 5);
        global.dictkeywordsshow = (BOOL *) MemAlloc(j * sizeof(BOOL) + 5);

        for (i = 0; i < j; i++) {
            global.dictkeywords[i] = FixedDictKeywords[i][0];
            global.dictkeywordsdesc[i] = FixedDictKeywords[i][1];
            global.dictkeywordsshow[i] = TRUE;
        }
    }

    /* Hot-keys */

    for (i = 0; i < NRJWPFUNC; i++) {
        ReadOptionString("Hot Keys", jwpfunc[i].desc, "None", buffer, MAXLINELEN, OptionsFileName);
        hotkey.ctrl = hotkey.shift = FALSE;
        hotkey.alpha = hotkey.scancode = 0;

        if (!buffer[0] || !stricmp(buffer, "None")) continue;

        for (j = 0; buffer[j]; j++) {
            if (buffer[j] == SHIFTCHAR) {
                hotkey.shift = TRUE;
            } else if (buffer[j] == CTRLCHAR) {
                ch = buffer[j+1];
                if ('a' <= ch && ch <= 'z') ch -= 0x20;
                if (ch > ' ' && ch < 0x7f && (ch != 'F' || !buffer[j+2])) {
                    hotkey.scancode = 0;
                    hotkey.alpha = ch;
                    hotkey.ctrl = TRUE;
                    break;
                }
                hotkey.ctrl = TRUE;
            } else if (buffer[j] == 'F' || buffer[j] == 'f') {
                hotkey.alpha = 0;
                switch(atoi(buffer + j + 1)) {
                    case 1: hotkey.scancode = VK_F1; break;
                    case 2: hotkey.scancode = VK_F2; break;
                    case 3: hotkey.scancode = VK_F3; break;
                    case 4: hotkey.scancode = VK_F4; break;
                    case 5: hotkey.scancode = VK_F5; break;
                    case 6: hotkey.scancode = VK_F6; break;
                    case 7: hotkey.scancode = VK_F7; break;
                    case 8: hotkey.scancode = VK_F8; break;
                    case 9: hotkey.scancode = VK_F9; break;
                    case 10: hotkey.scancode = VK_F10; break;
                    case 11: hotkey.scancode = VK_F11; break;
                    case 12: hotkey.scancode = VK_F12; break;
                }
                break;
            }
        }

        if (hotkey.alpha == 0 && hotkey.scancode == 0) continue;

        for (j = 0; j < i; j++) {
            if (!memcmp(&(jwpfunc[j].hotkey), &hotkey, sizeof(HOTKEY))) {
                ErrorMessage(global.hwnd, "Hot-key assigned to \"%s\" conflicts "
                                          "with the hot-key assigned to \"%s\"!",
                                          jwpfunc[i].desc, jwpfunc[j].desc);
                break;
            }
        }

        if (j != i) {
            hotkey.ctrl = hotkey.shift = FALSE;
            hotkey.alpha = hotkey.scancode = 0;
        }

        jwpfunc[i].hotkey = hotkey;
    }

    ProcessMenuTexts();

    OptionsChanged = FALSE;

    return (TRUE);
}



void WriteOptionsFile (BOOL always)
{
    int i;
    HFILE fd;
    HKEY hsubkey;
    FILE *fp;
    OFSTRUCT of;
    char buffer[MAXLINELEN];
    BOOL DoIt = FALSE;
    RECT progrect;
    time_t t;


    GetWindowRect(global.hwnd, &progrect);

    if (always || OptionsChanged) {
        DoIt = TRUE;
    } else {
        /* Things that may change but do not affect OptionsChanged */

        if (global.glistvisible && global.glisthwnd == NULL) DoIt = TRUE;
        if (!global.glistvisible && global.glisthwnd != NULL) DoIt = TRUE;

        if (global.statushwnd == NULL && global.showstatus) DoIt = TRUE;
        if (global.statushwnd != NULL && !global.showstatus) DoIt = TRUE;

        if (global.tbhwnd == NULL && global.showribbon) DoIt = TRUE;
        if (global.tbhwnd != NULL && !global.showribbon) DoIt = TRUE;

        if (fabs(AsciiFontSize - DefAsciiFont.size) > 0.01) DoIt = TRUE;
        if (strcmp(AsciiFontName, DefAsciiFont.facename)) DoIt = TRUE;

        if (memcmp(&glistsize, &global.glistsize, sizeof(RECT)) != 0) DoIt = TRUE;
        if (memcmp(&convbar, &global.convbar, sizeof(RECT)) != 0) DoIt = TRUE;

        if (clipboard != global.clipboard) DoIt = TRUE;
        if (global.draftview != DraftView) DoIt = TRUE;
        if (global.showspecial != ShowSpecial) DoIt = TRUE;
        if (global.showruler != ShowRuler) DoIt = TRUE;
        if (global.measure != Measure) DoIt = TRUE;
        if (global.longjust != LongJust) DoIt = TRUE;
        if (global.shortjust != ShortJust) DoIt = TRUE;
        if (global.ctl3d != Dialogs3D) DoIt = TRUE;
        if (global.undolevels != UndoLevels) DoIt = TRUE;
        if (global.saveundolevels != SaveUndoLevels) DoIt = TRUE;

        if (fabs(global.pagesetup.margins[0] - PageSetup.margins[0]) > 0.01) DoIt = TRUE;
        if (fabs(global.pagesetup.margins[1] - PageSetup.margins[1]) > 0.01) DoIt = TRUE;
        if (fabs(global.pagesetup.margins[2] - PageSetup.margins[2]) > 0.01) DoIt = TRUE;
        if (fabs(global.pagesetup.margins[3] - PageSetup.margins[3]) > 0.01) DoIt = TRUE;
        if (global.pagesetup.landscape != PageSetup.landscape) DoIt = TRUE;

        if (IsZoomed(global.hwnd)) {
            if (!ProgramMaximized) DoIt = TRUE;
        } else {
            if (ProgramMaximized) DoIt = TRUE;
            if (ProgramPosition.left != progrect.left) DoIt = TRUE;
            if (ProgramPosition.right != progrect.right) DoIt = TRUE;
            if (ProgramPosition.top != progrect.top) DoIt = TRUE;
            if (ProgramPosition.bottom != progrect.bottom) DoIt = TRUE;
        }
    }

    if (!DoIt) return;


#ifndef USEWIN32

    GetWindowsDirectory(buffer, MAXLINELEN);
    i = strlen(buffer);
    if (i > 0 && buffer[i-1] != '\\') {
        buffer[i] = '\\';
        buffer[i+1] = '\0';
    }
    strcat(buffer, PROGNAME ".INI");

    fd = OpenFile(buffer, &of, OF_WRITE | OF_CREATE);
    if (fd == HFILE_ERROR) {
        ErrorMessage(global.hwnd, "Cannot write initialization file %s!", buffer);
        return;
    }

    fp = fdopen(fd, "w");
    if (fp == NULL) {
        ErrorMessage(global.hwnd, "Cannot create " PROGNAME " " JWPVERSION " initialization file!");
        _lclose(fd);
        return;
    }

#else

    fp = NULL;

#endif USEWIN32


    /* Now we do things in a certain way so that the compiler can merge */
    /* similar strings (mainly the option names) with the strings in    */
    /* ReadOptionsFile() to conserve local data space (for 16-bit)!     */

    /* Information */

    time(&t);

    if (fp != NULL) {
        fprintf(fp, "; " PROGNAME ".INI - Initialization file for " PROGNAME "\n");
        fprintf(fp, ";           version " JWPVERSION "\n");
        fprintf(fp, "; Last Modified %s", ctime(&t));
    }

#ifdef USEWIN32
    strcpy(buffer, ctime(&t));
    buffer[strlen(buffer) - 1] = '\0';

    if (RegCreateKey(HKEY_CURRENT_USER, "Software\\" PROGNAME "\\Program", &hsubkey) == ERROR_SUCCESS) {
        RegSetValueEx(hsubkey, "Options Saved", 0, REG_SZ, buffer, strlen(buffer) + 1);
        RegSetValueEx(hsubkey, "Version", 0, REG_SZ, JWPVERSION, strlen(JWPVERSION) + 1);
        RegCloseKey(hsubkey);
    }

#endif USEWIN32

    if (fp != NULL) fprintf(fp, "\n[Program]\n");
    if (fp != NULL) fprintf(fp, "%s=%s\n", "Version", JWPVERSION);

    /* Path names */

    if (fp != NULL) fprintf(fp, "\n[File Names]\n");
    SaveKeyAndINI(fp, "File Names", PROGNAME " Directory",         "%s", global.jwppath);
    SaveKeyAndINI(fp, "File Names", "Templates",                   "%s", global.templatepath);
    SaveKeyAndINI(fp, "File Names", "Help File",                   "%s", global.help);
    SaveKeyAndINI(fp, "File Names", "Conversion Dictionary",       "%s", global.convdict);
    SaveKeyAndINI(fp, "File Names", "Conversion Dictionary Index", "%s", global.convidx);
    SaveKeyAndINI(fp, "File Names", "User Conversion Dictionary",  "%s", global.userdict);
    SaveKeyAndINI(fp, "File Names", "Conversion Cache",            "%s", global.convcache);
    SaveKeyAndINI(fp, "File Names", "Global Glossary",             "%s", global.glossary);
    SaveKeyAndINI(fp, "File Names", "Kanji Info",                  "%s", global.kinfo);
    SaveKeyAndINI(fp, "File Names", "Kanji Info Index (Stroke)",   "%s", global.kinfostridx);
    SaveKeyAndINI(fp, "File Names", "Kanji Info Index (Bushu)",    "%s", global.kinfobshidx);
    SaveKeyAndINI(fp, "File Names", "Radical Lookup Database (Radical)",        "%s", global.radkanjidat);
    SaveKeyAndINI(fp, "File Names", "Radical Lookup Database Index (Radical)",  "%s", global.radkanjiidx);
    SaveKeyAndINI(fp, "File Names", "Radical Lookup Database (Stroke)",         "%s", global.stroknjidat);
    SaveKeyAndINI(fp, "File Names", "Radical Lookup Database Index (Stroke)",   "%s", global.stroknjiidx);

    /* Page Setup */

    if (fp != NULL) fprintf(fp, "\n[Page Setup]\n");
    FindMeasurement(buffer, global.measure);
    SaveKeyAndINI(fp, "Page Setup", "Measurement Unit", "%s", buffer);
    SaveKeyAndINI(fp, "Page Setup", "Margins", "%4.2lf %4.2lf %4.2lf %4.2lf",
                    global.pagesetup.margins[0], global.pagesetup.margins[1],
                    global.pagesetup.margins[1], global.pagesetup.margins[3]);

    switch (global.longjust) {
        case J_LEFT:    SaveKeyAndINI(fp, "Page Setup", "Long Text Justification", "%s", "Left"); break;
        default:
        case J_JUSTIFY: SaveKeyAndINI(fp, "Page Setup", "Long Text Justification", "%s", "Justify"); break;
    }
    switch (global.shortjust) {
        case J_LEFT:    SaveKeyAndINI(fp, "Page Setup", "Short Text Justification", "%s", "Left"); break;
        case J_RIGHT:   SaveKeyAndINI(fp, "Page Setup", "Short Text Justification", "%s", "Right"); break;
        default:
        case J_CENTER:  SaveKeyAndINI(fp, "Page Setup", "Short Text Justification", "%s", "Center"); break;
        case J_JUSTIFY: SaveKeyAndINI(fp, "Page Setup", "Short Text Justification", "%s", "Justify"); break;
    }

    /* ASCII Font */

    if (fp != NULL) fprintf(fp, "\n[Base ASCII Font]\n");
    SaveKeyAndINI(fp, "Base ASCII Font", "Face Name",   "%s", DefAsciiFont.facename);
    SaveKeyAndINI(fp, "Base ASCII Font", "Size",        "%3.1lf", DefAsciiFont.size);

    /* Kanji Fonts */

    if (fp != NULL) fprintf(fp, "\n[Kanji Fonts]\n");
    SaveKeyAndINI(fp, "Kanji Fonts", "Number of Kanji Fonts",    "%d",      global.nr_fonts);
    SaveKeyAndINI(fp, "Kanji Fonts", "System Font",              "%d",      SysFontNum + 1);
    SaveKeyAndINI(fp, "Kanji Fonts", "Default Base Font",        "%d",      BaseFontNum + 1);
    SaveKeyAndINI(fp, "Kanji Fonts", "Default Print Font",       "%d",      PrintFontNum + 1);
    SaveKeyAndINI(fp, "Kanji Fonts", "Default Print Size",       "%3.1lf",  global.printscale);

    for(i = 0; i < global.nr_fonts; i++) {
        char name[MAXLINELEN];

        sprintf(name, "Font #%d", i+1);
        SaveKeyAndINI(fp, "Kanji Fonts", name, "%s", InstalledFonts[i].filename);
    }

    /* Glossary Options */

    if (fp != NULL) fprintf(fp, "\n[Glossary]\n");
    SaveKeyAndINI(fp, "Glossary", "Dynamic Glossary",         "%s",     (global.dynamicglossary ? "Yes" : "No"));
    SaveKeyAndINI(fp, "Glossary", "Glossary List Disappears", "%s",     (global.glistbehaviour ? "Yes" : "No"));
    SaveKeyAndINI(fp, "Glossary", "Glossary List Visible",    "%s",     (global.glisthwnd != NULL) ? "Yes" : "No");
    SaveKeyAndINI(fp, "Glossary", "Glossary List Position",   "%d %d",  global.glistsize.left, global.glistsize.top);
    SaveKeyAndINI(fp, "Glossary", "Glossary List Size",       "%d %d",  global.glistsize.right, global.glistsize.bottom);

    /* Undo Options */

    if (fp != NULL) fprintf(fp, "\n[Undo]\n");
    if (global.undolevels < 0) {
        SaveKeyAndINI(fp, "Undo", "Number of Undo Levels", "%d", 0);
    } else {
        SaveKeyAndINI(fp, "Undo", "Number of Undo Levels", "%d", global.undolevels);
    }
    if (global.saveundolevels < 0) {
        SaveKeyAndINI(fp, "Undo", "Undo Levels Saved", "%s", "All");
    } else {
        SaveKeyAndINI(fp, "Undo", "Undo Levels Saved", "%d", global.saveundolevels);
    }

    /* Other Options */

    if (fp != NULL) fprintf(fp, "\n[Others]\n");

    if (IsZoomed(global.hwnd)) {
        SaveKeyAndINI(fp, "Others", "Program Position", "%s", "Maximized");
    } else {
        SaveKeyAndINI(fp, "Others", "Program Position", "%d %d", progrect.left, progrect.top);
        SaveKeyAndINI(fp, "Others", "Program Size",     "%d %d",
                            progrect.right - progrect.left,
                            progrect.bottom - progrect.top);
    }

    SaveKeyAndINI(fp, "Others", "Show Input Line Editor",       "%s", (global.imehwnd != NULL) ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Relax Margin for Punctuation", "%s", (global.relaxmargin ? "Yes" : "No"));
    SaveKeyAndINI(fp, "Others", "Dynamic Conversion",           "%s", (global.dynamicconvert ? "Yes" : "No"));

    switch (global.listposition) {
        case 0: SaveKeyAndINI(fp, "Others", "Conversion Bar Position", "%d %d", global.convbar.left, global.convbar.top);
                break;
        case 1: SaveKeyAndINI(fp, "Others", "Conversion Bar Position", "%s", "Top");
                break;
        default:
        case 2: SaveKeyAndINI(fp, "Others", "Conversion Bar Position", "%s", "Bottom");
                break;
    }

    SaveKeyAndINI(fp, "Others", "Conversion Bar Size",          "%d %d", global.convbar.right, global.convbar.bottom);
    SaveKeyAndINI(fp, "Others", "Clipboard Format",             "%s",    FileFormatNames[global.clipboard]);
    SaveKeyAndINI(fp, "Others", "Default Open Format",          "%s",    FileFormatNames[global.openformat]);
    SaveKeyAndINI(fp, "Others", "Draft View",                   "%s",    global.draftview ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Show Special Symbols",         "%s",    global.showspecial ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Show Iconbar",                 "%s",    (global.tbhwnd != NULL) ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Show Status Bar",              "%s",    (global.statushwnd != NULL) ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Show File Ruler",              "%s",    global.showruler ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "3D Dialog Boxes",              "%s",    global.ctl3d ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Automatic Repagenation",       "%s",    global.autorepage ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Automatic Radical Lookup",     "%s",    global.autorlookup ? "Yes" : "No");
    SaveKeyAndINI(fp, "Others", "Radical Lookup Sort Method",   "%d  ; 0 = No sort, 1 = strokes, 2 = level/strokes",
                                    global.rlookupsort);

    /* Quick files */

    if (fp != NULL) fprintf(fp, "\n[Quick Files]\n");

    for (i = 0; ; i++) {
        char *cp;
        char name[MAXLINELEN];
        extern char *GetQuickFile(int);

        cp = GetQuickFile(i);
        if (cp == NULL) break;
        sprintf(name, "File%d", i+1);
        SaveKeyAndINI(fp, "Quick Files", name, "%s", cp);
    }

    SaveKeyAndINI(fp, "Quick Files", "Number of Quick Files", "%d", i);

    /* Dictionary */

    if (fp != NULL) fprintf(fp, "\n[Dictionary]\n");
    SaveKeyAndINI(fp, "Dictionary", "Number of Dictionaries", "%d", global.nr_jdict);

    for (i = 0; i < global.nr_jdict; i++) {
		char name[MAXLINELEN];

        sprintf(name, "Dictionary #%d", i+1);
        sprintf(buffer, "%s|%s|%s|%s", global.jdict[i], global.jdictidx[i],
                                    global.jdictactive[i] ? "Yes" : "No",
                                    global.jdictdesc[i]);
        SaveKeyAndINI(fp, "Dictionary", name, "%s", buffer);
    }

    SaveKeyAndINI(fp, "Dictionary", "Number of Keywords", "%d", global.nr_dictkeywords);

    for (i = 0; i < global.nr_dictkeywords; i++) {
		char name[MAXLINELEN];

		sprintf(name, "Key #%d", i+1);
        sprintf(buffer, "%s|%s|%s", global.dictkeywords[i],
                                    global.dictkeywordsshow[i] ? "Yes" : "No",
                                    global.dictkeywordsdesc[i]);
        SaveKeyAndINI(fp, "Dictionary", name, "%s", buffer);
    }

    /* Hot-keys */

    if (fp != NULL) fprintf(fp, "\n[Hot Keys]\n");

    for (i = 0; i < NRJWPFUNC; i++) {
        if (jwpfunc[i].hotkey.alpha == 0 && jwpfunc[i].hotkey.scancode == 0) {
            SaveKeyAndINI(fp, "Hot Keys", jwpfunc[i].desc, "%s", "None");
            continue;
        }

        if (jwpfunc[i].hotkey.alpha != 0) {
            if (jwpfunc[i].hotkey.shift)
                sprintf(buffer, "%c%c%c",SHIFTCHAR, CTRLCHAR, jwpfunc[i].hotkey.alpha);
            else
                sprintf(buffer, "%c%c", CTRLCHAR, jwpfunc[i].hotkey.alpha);
        } else {
            if (jwpfunc[i].hotkey.ctrl) {
                if (jwpfunc[i].hotkey.shift) sprintf(buffer, "%c%cF", SHIFTCHAR, CTRLCHAR);
                else sprintf(buffer, "%cF", CTRLCHAR);
            } else {
                if (jwpfunc[i].hotkey.shift) sprintf(buffer, "%cF", SHIFTCHAR);
                else strcpy(buffer, "F");
            }

            switch (jwpfunc[i].hotkey.scancode) {
                case VK_F1:  strcat(buffer, "1"); break;
                case VK_F2:  strcat(buffer, "2"); break;
                case VK_F3:  strcat(buffer, "3"); break;
                case VK_F4:  strcat(buffer, "4"); break;
                case VK_F5:  strcat(buffer, "5"); break;
                case VK_F6:  strcat(buffer, "6"); break;
                case VK_F7:  strcat(buffer, "7"); break;
                case VK_F8:  strcat(buffer, "8"); break;
                case VK_F9:  strcat(buffer, "9"); break;
                case VK_F10: strcat(buffer, "10"); break;
                case VK_F11: strcat(buffer, "11"); break;
                case VK_F12: strcat(buffer, "12"); break;
                default:     strcat(buffer, "0"); break;
            }
        }

        SaveKeyAndINI(fp, "Hot Keys", jwpfunc[i].desc, "%s", buffer);
    }

    if (fp != NULL) fclose(fp);
}



BOOL CALLBACK OptionsProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static FILEFORMAT OldCBFormat;

    switch (message) {
        case WM_INITDIALOG: {
            int i;
            static char temp[] = "\x5a\x67\x60\x3c\x30\x53\x7f\x80\x89\x82"
            "\x79\x77\x78\x84\x30\x38\x53\x39\x30\x63"
            "\x84\x75\x80\x78\x75\x7e\x30\x53\x78\x85"
            "\x7e\x77";
            char buffer[10];

            CheckDlgButton (hwnd, 4211, global.relaxmargin);
            CheckDlgButton (hwnd, 4212, global.dynamicconvert);
            buffer[0] = temp[0];

#ifdef CTL3D
            CheckDlgButton (hwnd, 4213, global.ctl3d);
#else CTL3D
            EnableWindow (GetDlgItem(hwnd, 4213), FALSE);
#endif CTL3D
            CheckDlgButton (hwnd, 4215, global.glistbehaviour);
            CheckDlgButton (hwnd, 4216, global.saveundolevels != 0);

            SendDlgItemMessage(hwnd, 4221, BM_SETCHECK, FALSE, 0L);
            SendDlgItemMessage(hwnd, 4222, BM_SETCHECK, FALSE, 0L);
            SendDlgItemMessage(hwnd, 4223, BM_SETCHECK, FALSE, 0L);

            switch (global.listposition) {
                case 0: SendDlgItemMessage(hwnd, 4223, BM_SETCHECK, TRUE, 0L);
                        break;
                case 1: SendDlgItemMessage(hwnd, 4221, BM_SETCHECK, TRUE, 0L);
                        break;
                case 2: SendDlgItemMessage(hwnd, 4222, BM_SETCHECK, TRUE, 0L);
                        break;
            }

            SendDlgItemMessage(hwnd, 4231, CB_ADDSTRING, 0, (LONG) "Left");
            SendDlgItemMessage(hwnd, 4231, CB_ADDSTRING, 0, (LONG) "Right");
            SendDlgItemMessage(hwnd, 4231, CB_ADDSTRING, 0, (LONG) "Center");
            SendDlgItemMessage(hwnd, 4231, CB_ADDSTRING, 0, (LONG) "Justify");
            switch (global.shortjust) {
                case J_LEFT:    SendDlgItemMessage(hwnd, 4231, CB_SETCURSEL, 0, 0L); break;
                case J_RIGHT:   SendDlgItemMessage(hwnd, 4231, CB_SETCURSEL, 1, 0L); break;
                default:
                case J_CENTER:  SendDlgItemMessage(hwnd, 4231, CB_SETCURSEL, 2, 0L); break;
                case J_JUSTIFY: SendDlgItemMessage(hwnd, 4231, CB_SETCURSEL, 3, 0L); break;
            }

            SendDlgItemMessage(hwnd, 4232, CB_ADDSTRING, 0, (LONG) "Left");
            SendDlgItemMessage(hwnd, 4232, CB_ADDSTRING, 0, (LONG) "Justify");
            switch (global.longjust) {
                case J_LEFT:    SendDlgItemMessage(hwnd, 4232, CB_SETCURSEL, 0, 0L); break;
                default:
                case J_JUSTIFY: SendDlgItemMessage(hwnd, 4232, CB_SETCURSEL, 1, 0L); break;
            }


            /* We start from EUC, skipping Unknown and Normal */

            for (i = FF_EUC; FileFormatNames[i] != NULL; i++)
                SendDlgItemMessage(hwnd, 4241, CB_INSERTSTRING, -1, (LONG) ((LPSTR) FileFormatNames[i]));
            SendDlgItemMessage(hwnd, 4241, CB_SETCURSEL, global.clipboard - FF_EUC, 0L);

            OldCBFormat = global.clipboard;

            /* Levels of undo, several pre-set levels */

            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "No Undo");
            SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 0, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "1");
            if (global.undolevels >= 1) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 1, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "5");
            if (global.undolevels >= 5) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 2, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "10");
            if (global.undolevels >= 10) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 3, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "50");
            if (global.undolevels >= 50) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 4, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "100");
            if (global.undolevels >= 100) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 5, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "500");
            if (global.undolevels >= 500) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 6, 0L);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) "1000");
            if (global.undolevels >= 1000) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 7, 0L);
            sprintf(buffer, "%d", MAXUNDOLEVELS);
            SendDlgItemMessage(hwnd, 4242, CB_INSERTSTRING, -1, (LONG) buffer);
            if (global.undolevels >= MAXUNDOLEVELS) SendDlgItemMessage(hwnd, 4242, CB_SETCURSEL, 8, 0L);

            TrapHelpInDialogBoxes(hwnd, "OptionsDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4211:      /* Check boxes */
                case 4212:
                case 4213:
                case 4215:
                case 4216:
                    if (NOTIFYCODE == BN_DOUBLECLICKED) SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    return (TRUE);

                case 4221:      /* Top */
                    SendDlgItemMessage(hwnd, 4221, BM_SETCHECK, TRUE, 0L);
                    SendDlgItemMessage(hwnd, 4222, BM_SETCHECK, FALSE, 0L);
                    SendDlgItemMessage(hwnd, 4223, BM_SETCHECK, FALSE, 0L);
                    if (NOTIFYCODE == BN_DOUBLECLICKED) SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    return (TRUE);

                case 4222:      /* Bottom */
                    SendDlgItemMessage(hwnd, 4222, BM_SETCHECK, TRUE, 0L);
                    SendDlgItemMessage(hwnd, 4221, BM_SETCHECK, FALSE, 0L);
                    SendDlgItemMessage(hwnd, 4223, BM_SETCHECK, FALSE, 0L);
                    if (NOTIFYCODE == BN_DOUBLECLICKED) SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    return (TRUE);

                case 4223:      /* Floating */
                    SendDlgItemMessage(hwnd, 4223, BM_SETCHECK, TRUE, 0L);
                    SendDlgItemMessage(hwnd, 4221, BM_SETCHECK, FALSE, 0L);
                    SendDlgItemMessage(hwnd, 4222, BM_SETCHECK, FALSE, 0L);
                    if (NOTIFYCODE == BN_DOUBLECLICKED) SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    return (TRUE);

                case 4241: {        /* Clipboard format */
                    int i;

                    switch (NOTIFYCODE) {
                        case LBN_SELCHANGE:
                        case LBN_DBLCLK:
                            i = (WORD) SendMessage(GetDlgItem(hwnd, 4241), CB_GETCURSEL, 0, 0L);
                            if (i == LB_ERR) {
                                MessageBeep(MB_ICONEXCLAMATION);
                                return (TRUE);
                            }
                            global.clipboard = i + FF_EUC;   /* We start from EUC */

                            if (NOTIFYCODE == LBN_DBLCLK) SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                            return (TRUE);
                    }
                    break;
                }

                case 4242:          /* Undo levels */
                    switch (NOTIFYCODE) {
                        case LBN_DBLCLK:
                            SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                            return (TRUE);
                    }
                    break;

                case IDOK: {
                    int i;
                    BOOL Redraw = FALSE;
                    RECT rect;

                    i = IsDlgButtonChecked(hwnd, 4211);
                    if (i != global.relaxmargin) OptionsChanged = TRUE;
                    global.relaxmargin = i;

                    i = IsDlgButtonChecked(hwnd, 4212);
                    if (i != global.dynamicconvert) OptionsChanged = TRUE;
                    global.dynamicconvert = i;

#ifdef CTL3D
                    i = IsDlgButtonChecked(hwnd, 4213);
                    if (i != global.ctl3d) OptionsChanged = TRUE;
                    global.ctl3d = i;
#endif CTL3D

                    i = IsDlgButtonChecked(hwnd, 4215);
                    if (i != global.glistbehaviour) OptionsChanged = TRUE;
                    global.glistbehaviour = i;

                    i = IsDlgButtonChecked(hwnd, 4216);
                    if (i) {
                        if (global.saveundolevels == 0) OptionsChanged = TRUE;
                        global.saveundolevels = -1;
                    } else {
                        if (global.saveundolevels != 0) OptionsChanged = TRUE;
                        global.saveundolevels = 0;
                    }

                    if (IsDlgButtonChecked(hwnd, 4223)) {
                        if (global.listposition != 0) OptionsChanged = Redraw = TRUE;
                        global.listposition = 0;
                    } else if (IsDlgButtonChecked(hwnd, 4221)) {
                        if (global.listposition != 1) OptionsChanged = Redraw = TRUE;
                        global.listposition = 1;
                    } else {
                        if (global.listposition != 2) OptionsChanged = Redraw = TRUE;
                        global.listposition = 2;
                    }

                    i = SendDlgItemMessage(hwnd, 4231, CB_GETCURSEL, 0, 0L);
                    if (i != CB_ERR) {
                        switch (i) {
                            case 0: global.shortjust = J_LEFT; break;
                            case 1: global.shortjust = J_RIGHT; break;
                            case 2: global.shortjust = J_CENTER; break;
                            case 3: global.shortjust = J_JUSTIFY; break;
                        }
                    }

                    i = SendDlgItemMessage(hwnd, 4232, CB_GETCURSEL, 0, 0L);
                    if (i != CB_ERR) {
                        switch (i) {
                            case 0: global.longjust = J_LEFT; break;
                            case 1: global.longjust = J_JUSTIFY; break;
                        }
                    }

                    i = (WORD) SendMessage(GetDlgItem(hwnd, 4242), CB_GETCURSEL, 0, 0L);
                    if (i != LB_ERR) {
                        switch (i) {
                            case 0: global.undolevels = 0; break;
                            case 1: global.undolevels = 1; break;
                            case 2: global.undolevels = 5; break;
                            case 3: global.undolevels = 10; break;
                            case 4: global.undolevels = 50; break;
                            case 5: global.undolevels = 100; break;
                            case 6: global.undolevels = 500; break;
                            case 7: global.undolevels = 1000; break;
                            case 8: global.undolevels = MAXUNDOLEVELS; break;
                        }
                    }

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);

                    if (Redraw) {
                        GetClientRect(global.hwnd, &rect);
                        i = rect.bottom - rect.top;
                        PostMessage(global.hwnd, WM_SIZE, 0, MAKELONG(rect.right - rect.left, i));
                    }

                    if (OldCBFormat != global.clipboard)
                        ResetClipboard();

                    return (TRUE);
                }

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
    }

    return (FALSE);
}



static KANJI far *ConvertFontName (int id, LONG lParam, KANJI *buf)
{
    int i, j;
    KANJI *kp;
    char tempbuf[50];

    kp = InstalledFonts[lParam].facename;

    for (i = 0; kp[i]; i++) buf[i] = kp[i];

    sprintf(tempbuf, " (%d x %d)", InstalledFonts[lParam].width, InstalledFonts[lParam].height);

    for (j = 0; tempbuf[j]; j++) buf[i++] = tempbuf[j];
    buf[i] = 0;

    return (buf);
}



short CALLBACK AsciiFontEnumProc (LOGFONT far *lf, TEXTMETRIC far *tm, short ftype, LPSTR data)
{
    int i, len;
    HWND hwnd;
    char buffer[MAXLINELEN];
    char temp[MAXLINELEN];

    hwnd = (HWND) data;

    if (ftype & DEVICE_FONTTYPE) {
        buffer[0] = '*';
        _fstrcpy(buffer + 1, lf->lfFaceName);
    } else {
        _fstrcpy(buffer, lf->lfFaceName);
    }

    len = SendMessage(hwnd, CB_GETCOUNT, 0, 0L);

    for (i = 0; i < len; i++) {
        SendMessage(hwnd, CB_GETLBTEXT, i, (LONG) ((char far *) temp));
        if (!stricmp(buffer, temp)) return (1);
    }

    SendMessage(hwnd, CB_ADDSTRING, 0, (LONG) buffer);

    return (1);
}



short CALLBACK AsciiFontSizeEnumProc (LOGFONT far *lf, TEXTMETRIC far *tm, short ftype, LPSTR data)
{
    int i, j, len, dpi;
    double ps;
    HDC hdc;

    if (!(ftype & RASTER_FONTTYPE) || (ftype & TRUETYPE_FONTTYPE)) {
        for (i = 8, j = 0; i <= 24; i++) PointSizes[j++] = i;
        PointSizes[j++] = 36.0;
        PointSizes[j++] = 48.0;
        PointSizes[j] = 0.0;

        return (1);
    }

    len = SegHeapGetSize(PointSizes) / sizeof(double);

    hdc = (HDC) data;

    dpi = GetDeviceCaps(hdc, LOGPIXELSY);
    ps = ((double) tm->tmHeight - (double) tm->tmInternalLeading) * 72.0 / (double) dpi;

    for (i = 0; i < len; i++) {
        if (PointSizes[i] == 0.0) break;
        if (fabs(ps - PointSizes[i]) < 0.01) return (0);
    }

    if (i >= len) PointSizes = (double far *) BlockRealloc(PointSizes, (len + PSBLOCKSIZE) * sizeof(double));

    PointSizes[i++] = ps;
    PointSizes[i] = 0.0;

    return (0);
}



static void FillAuxStuff (HWND hwnd, int group, int window, int printfont, double scale)
{
    int x, len;
    double dpp, ps;
    char buffer[MAXLINELEN];

    /* Fill in some aux stuff */

    x = SendDlgItemMessage(hwnd, printfont, CB_GETCURSEL, 0, 0L);
    if (x == CB_ERR) {
        sprintf(buffer, "Kanji Printing Size = Unknown");
        SetDlgItemText(hwnd, group, buffer);
        SetDlgItemText(hwnd, window, "");
        return;
    }

    dpp = (double) global.resolution.y / 72.0;
    ps = ((double) InstalledFonts[x].height) / dpp;

    sprintf(buffer, "Kanji Printing Size = %3.1lf point%s",
            ps, (ps > 1.10) ? "s" : "");
    SetDlgItemText(hwnd, group, buffer);

    sprintf(buffer, "%3.1lf", ps * scale);

    len = strlen(buffer);
    if (len > 2 && buffer[len-1] == '0') buffer[len-2] = '\0';
    SetDlgItemText(hwnd, window, buffer);
}



static void FillPointSizes (HWND hwnd, int window, char *facename, double ps)
{
    int i, len;
    double itemps;
    char buffer[20];
    HDC hdc;
    FARPROC enumproc;

    SendDlgItemMessage(hwnd, window, WM_SETREDRAW, FALSE, 0L);
    SendDlgItemMessage(hwnd, window, CB_RESETCONTENT, 0, 0L);

    hdc = GetPrinterDC(TRUE, FALSE, NULL);
    if (hdc == NULL) hdc = CreateIC("DISPLAY", NULL, NULL, NULL);

    if (*facename == '*') facename++;

    PointSizes = (double far *) BlockAlloc(sizeof(double) * PSBLOCKSIZE);
    PointSizes[0] = 0.0;

    enumproc = MakeProcInstance((FARPROC) AsciiFontSizeEnumProc, hInstance);
    EnumFonts (hdc, facename, enumproc, (LONG) hdc);
    FreeProcInstance(enumproc);

    for (i = 0; PointSizes[i] != 0.0; i++) {
        sprintf(buffer, "%3.1lf", PointSizes[i]);

        len = strlen(buffer);
        if (buffer[len - 1] == '0') buffer[len - 2] = '\0';

        SendDlgItemMessage(hwnd, window, CB_ADDSTRING, 0, (LONG) ((char far *) buffer));
    }
    FreeBlock(PointSizes);

    len = SendDlgItemMessage(hwnd, window, CB_GETCOUNT, 0, 0L);
    for (i = 0; i < len; i++) {
        SendDlgItemMessage(hwnd, window, CB_GETLBTEXT, i, (LONG) ((char far *) buffer));
        itemps = atof(buffer);
        if (fabs(itemps - ps) < 0.01) break;
    }

    if (i < len) SendDlgItemMessage(hwnd, window, CB_SETCURSEL, i, 0L);
    DeleteDC(hdc);

    SendDlgItemMessage(hwnd, window, WM_SETREDRAW, TRUE, 0L);
    InvalidateRect(GetDlgItem(hwnd, window), NULL, TRUE);
}



BOOL CALLBACK InstalledFontsProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i, x, len;
            double printscale;
            HDC hdc;
            FARPROC enumproc;
            char buffer[MAXLINELEN];

            SendDlgItemMessage(hwnd, 4231, EM_LIMITTEXT, 4, 0L);
            SendDlgItemMessage(hwnd, 4202, CB_LIMITTEXT, 4, 0L);

            /* Fill in the ASCII font box */

            hdc = GetPrinterDC(TRUE, FALSE, NULL);
            if (hdc == NULL) hdc = CreateIC("DISPLAY", NULL, NULL, NULL);

            enumproc = MakeProcInstance((FARPROC) AsciiFontEnumProc, hInstance);
            EnumFonts (hdc, NULL, enumproc, (LONG) GetDlgItem(hwnd, 4201));
            FreeProcInstance(enumproc);

            len = SendDlgItemMessage(hwnd, 4201, CB_GETCOUNT, 0, 0L);
            for (i = 0; i < len; i++) {
                SendDlgItemMessage(hwnd, 4201, CB_GETLBTEXT, i, (LONG) ((char far *) buffer));
                if (buffer[0] == '*') {
                    if (!strcmp(buffer + 1, DefAsciiFont.facename)) break;
                } else {
                    if (!strcmp(buffer, DefAsciiFont.facename)) break;
                }
            }

            if (i < len) {
                SendDlgItemMessage(hwnd, 4201, CB_SETCURSEL, i, 0L);
                FillPointSizes(hwnd, 4202, buffer, DefAsciiFont.size);
            }
            DeleteDC(hdc);


            /* Fill in all the kanji font boxes */

            for (i = 0; i < global.nr_fonts; i++) {
                SendDlgItemMessage(hwnd, 4211, CB_ADDSTRING, 0, (LONG) i);
                SendDlgItemMessage(hwnd, 4221, CB_ADDSTRING, 0, (LONG) i);
            }

            /* Select the right items */

            for (i = 0; i < global.nr_fonts; i++) {
                if (SendDlgItemMessage(hwnd, 4211, CB_GETITEMDATA, i, 0L) == BaseFontNum) {
                    SendDlgItemMessage(hwnd, 4211, CB_SETCURSEL, i, 0L);
                }
                if (SendDlgItemMessage(hwnd, 4221, CB_GETITEMDATA, i, 0L) == PrintFontNum) {
                    SendDlgItemMessage(hwnd, 4221, CB_SETCURSEL, i, 0L);
                }
            }


            x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
            printscale = (double) x / (double) PRINTFONT->height;

            FillAuxStuff(hwnd, 4232, 4231, 4221, printscale);

            SetFocus(GetDlgItem(hwnd, 4201));

            TrapHelpInDialogBoxes(hwnd, "KanjiFontsDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            int i, j;
            BOOL NeedReformat;
            double ratio, ps;
            char buffer[MAXLINELEN];

            switch (WINCMDID) {
                case 4211:      /* Base font */
                    switch (NOTIFYCODE) {
                        case CBN_SELCHANGE:
                            i = SendDlgItemMessage(hwnd, 4211, CB_GETCURSEL, 0, 0L);

                            if (i != CB_ERR) {
                                int x;
                                double printscale;

                                x = (int) floor(global.printscale * (double) global.resolution.y / 72.0 + 0.5);      /* dots */
                                printscale = (double) x / (double) InstalledFonts[i].height;

                                //FillAuxStuff(hwnd, 4232, 4231, 4221, printscale);
                            }
                            return (TRUE);

                        case CBN_DBLCLK:
                            SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                            return (TRUE);
                    }
                    return (TRUE);

                case 4201:      /* ASCII font */
                    switch (NOTIFYCODE) {
                        case CBN_SELCHANGE:
                            i = SendDlgItemMessage(hwnd, 4201, CB_GETCURSEL, 0, 0L);

                            if (i != CB_ERR) {  /* Update the point size */
                                SendDlgItemMessage(hwnd, 4201, CB_GETLBTEXT, i, (LONG) ((char far *) buffer));
                                FillPointSizes(hwnd, 4202, buffer, DefAsciiFont.size);
                            }
                            return (TRUE);

                        case CBN_DBLCLK:
                            SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                            return (TRUE);
                    }
                    return (TRUE);

                case 4221:      /* Print font */
                    switch (NOTIFYCODE) {
                        case CBN_SELCHANGE:
                            FillAuxStuff(hwnd, 4232, 4231, 4221, 1.0);
                            return (TRUE);

                        case CBN_DBLCLK:
                            SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                            return (TRUE);
                    }
                    return (TRUE);

                case IDOK:
                    NeedReformat = FALSE;

                    /* Check the base font */

                    i = SendDlgItemMessage(hwnd, 4211, CB_GETCURSEL, 0, 0L);
                    if (i != CB_ERR) {
                        j = SendDlgItemMessage(hwnd, 4211, CB_GETITEMDATA, i, 0L);
                        if (j != BaseFontNum) {
                            BaseFontNum = j;
                            BASEFONT = &(InstalledFonts[j]);
                            NeedReformat = OptionsChanged = TRUE;
                        }
                    }

                    /* Check the printing font */

                    i = SendDlgItemMessage(hwnd, 4221, CB_GETCURSEL, 0, 0L);
                    if (i != CB_ERR) {
                        j = SendDlgItemMessage(hwnd, 4221, CB_GETITEMDATA, i, 0L);
                        if (j != PrintFontNum) {
                            PrintFontNum = j;
                            PRINTFONT = &(InstalledFonts[j]);
                            NeedReformat = OptionsChanged = TRUE;
                        }
                    }

                    /* Is the aspect ratio OK? */

                    ratio = (double) InstalledFonts[BaseFontNum].width /
                                (double) InstalledFonts[BaseFontNum].height;
                    ratio /= ((double) InstalledFonts[PrintFontNum].width /
                                 (double) InstalledFonts[PrintFontNum].height);

                    if (fabs(ratio - 1.0) > 0.01) {
                        ErrorMessage(hwnd, "The kanji font for display and printing must "
                                              "have the same aspect ratio!");
                        return (TRUE);
                    }

                    /* Check the printing scale */

                    GetDlgItemText(hwnd, 4231, buffer, 5);
                    ps = atof(buffer);       /* Printing point size */

                    GetDlgItemText(hwnd, 4202, buffer, 5);

                    if (atof(buffer) - ps > 0.01) {
                        ErrorMessage(hwnd, "WARNING!  You have chosen a text font that is larger "
                                              "than the kanji print font!");
                    }

                    if (fabs(global.printscale - ps) > 0.01) {
                        global.printscale = ps;
                        NeedReformat = OptionsChanged = TRUE;
                    }

                    /* ASCII font */

                    GetDlgItemText(hwnd, 4201, buffer, MAXLINELEN);
                    if (strcmp(DefAsciiFont.facename, buffer)) {
                        FreeMem(DefAsciiFont.facename);
                        if (buffer[0] == '*') {
                            DefAsciiFont.facename = DupString(buffer + 1);
                        } else {
                            DefAsciiFont.facename = DupString(buffer);
                        }
                        NeedReformat = OptionsChanged = TRUE;
                    }
                    GetDlgItemText(hwnd, 4202, buffer, 5);
                    ps = atof(buffer);
                    if (ps >= 0.01 && fabs(ps - DefAsciiFont.size) > 0.01) {
                        DefAsciiFont.size = ps;
                        NeedReformat = OptionsChanged = TRUE;
                    }

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, NeedReformat);
                    return (TRUE);

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            return (FALSE);
        }

        case WM_COMPAREITEM:
        case WM_DELETEITEM:
        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            return (JlistProc(hwnd, message, wParam, lParam, TRUE, ConvertFontName));
    }

    return (FALSE);
}
