/* Copyright (C) Kevin Ortman 1995.  All rights reserved. */

/*
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that all of the following conditions are satisfied:
 *
 * 1) The above copyright notices appear in all copies
 * 2) Both those copyright notices and this permission notice appear
 *    in supporting documentation
 *
 * This software is provided "as is" without express or implied warranty.
 *
 */

#include "jwp.h"
#include "idm.h"

#define BOXGAP          (AVGHEIGHT / 4)
#define CARETWIDTH      2

#define ICON_SIZE       24
#define ICON_GAP        8

#define IME_SETFOCUS    19100
#define IME_KILLFOCUS   19101

#define LINEWIN_HEIGHT  (SYSFONT->height + 2 * LINEGAP(f) + 1)

int IMEWinHeight = 0;

extern LONG FAR PASCAL FileWinProc (HWND, UINT, WPARAM, LPARAM);

static WPARAM LastKeyCode;



void HilightDlgBox (HWND hwnd, HWND itemhwnd, HDC hdc, BOOL hilight)
{
    RECT rect;
    HBRUSH hbrush;
    HPEN hpen;

    GetClientRect(hwnd, &rect);
    rect.top = -1;
    rect.bottom++;
    Create3DEffect(hdc, &rect, 1, 0);

    // Fill in the Edit Window

    /*
    if(hilight) {
        hbrush = SelectObject(hdc, CreateSolidBrush(GetSysColor(COLOR_ACTIVECAPTION)));
        hpen = SelectObject(hdc, CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOWFRAME)));
    } else {
        hbrush = SelectObject(hdc, CreateSolidBrush(GetSysColor(COLOR_BTNFACE)));
        hpen = SelectObject(hdc, CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOWFRAME)));
    }
    Rectangle(hdc, rect.left, rect.top, rect.left + rect.right, rect.top + rect.bottom);

    DeleteObject(SelectObject(hdc, hbrush));
    DeleteObject(SelectObject(hdc, hpen));
    */

    // Draw a bounding box around the Edit Line
    GetClientRect(itemhwnd, &rect);

    rect.left = (int) GetProp(itemhwnd, PROP_SIZEX);
    rect.top = (int) GetProp(itemhwnd, PROP_SIZEY);

    hbrush = SelectObject(hdc, CreateSolidBrush(GetSysColor(COLOR_WINDOW)));
    hpen = SelectObject(hdc, CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOWFRAME)));

    Rectangle(hdc, rect.left, rect.top, rect.left + rect.right + 2 * BOXGAP,
               rect.top + rect.bottom + 2 * BOXGAP);

    DeleteObject(SelectObject(hdc, hbrush));
    DeleteObject(SelectObject(hdc, hpen));
}

LONG FAR PASCAL IMELineProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    FILEOPTIONS *f;
    CREATESTRUCT *cp;
    int xpos, ypos;

    f = (FILEOPTIONS *) GetProp(hwnd, PROP_FILEOPTIONS);

    switch (message) {
        case WM_CREATE:
            f = NewFile (FN_CONTROL | FN_IME, TRUE);
            if (f == NULL) return (0);

            SetProp(hwnd, PROP_FILEOPTIONS, f);

            /* Now, we make sure that the window is wide enough */
            cp = (CREATESTRUCT *) lParam;
            xpos = cp->x + BOXGAP;
            ypos = cp->y + BOXGAP;

            f->hwnd = hwnd;
            f->parent = GetParent(hwnd);
            f->width = cp->cx;
            f->height = cp->cy;
            f->linelen = 30000;

            SetProp(hwnd, PROP_SIZEX, (HANDLE) (xpos - BOXGAP));
            SetProp(hwnd, PROP_SIZEY, (HANDLE) (ypos - BOXGAP));
            SetProp(hwnd, PROP_CURSOR, NULL);

            //MoveWindow(hwnd, xpos, ypos, f->width, f->height, FALSE);

            return(0);

        case WM_SIZE:
            if (f == NULL) return (0);
            f->width = LOWORD(lParam);
            f->height = HIWORD(lParam);
            return (0);

        case EM_GETHANDLE:
            return ((LONG) f);

        case EM_GETLINE:
            if (f == NULL) return (0);
            return ((LONG) f->paragraph->text);

        case EM_LINELENGTH:
            if (f == NULL) return (0);
            return ((LONG) unitlen(f->paragraph->text));

        case EM_SETMODIFY:
            if (f == NULL) return (0);
            f->type = wParam;
            return ((LONG) f);

        case EM_SETSEL:
            if (f == NULL) return (0);

            TurnOffSelection(f);

            SELPARA1(f) = SELPARA2(f) = f->paragraph;
            SELPOS1(f) = LOWORD(lParam);
            SELPOS2(f) = HIWORD(lParam);
            SELTYPE(f) = SEL_SELECTION;
            FlipHighlight(f);

            CURCHAR(f) = wParam;
            if (!FindCaret(f, TRUE)) {
                MoveIntoWindow(f);
                InvalidateRect(f->hwnd, NULL, TRUE);
                UpdateWindow(f->hwnd);
            }
            DoCaret(f, CURX(f), CURY(f) - CURLINE(f)->height, TRUE);
            return (0);

        case EM_GETSEL:
            if (f == NULL) return (0);

            if (SELPARA1(f) == NULL) return (0);
            return (MAKELONG(SELPOS1(f), SELPOS2(f)));

        case WM_KEYDOWN:
            if (f == NULL) return (0);

            LastKeyCode = wParam;

            switch (wParam) {
                case VK_HOME:
                case VK_END:
                case VK_PRIOR:
                case VK_NEXT:
                    if(SELECTING(f)) {
                        SendMessage(global.convhwnd, WM_KEYDOWN, wParam, 0L);
                        SendMessage(global.convhwnd, WM_KEYDOWN, VK_RETURN, 0L);
                        return (0);
                    }
                    break;

                case VK_UP:
                case VK_LEFT:
                    if(SELECTING(f)) {
                        SendMessage(global.convhwnd, WM_KEYDOWN, VK_LEFT, 0L);
                        SendMessage(global.convhwnd, WM_KEYDOWN, VK_RETURN, 0L);
                        return (0);
                    }
                    break;

                case VK_DOWN:
                case VK_RIGHT:
                    if(SELECTING(f)) {
                        SendMessage(global.convhwnd, WM_KEYDOWN, VK_RIGHT, 0L);
                        SendMessage(global.convhwnd, WM_KEYDOWN, VK_RETURN, 0L);
                        return (0);
                    }
                    break;

                case VK_ESCAPE:
                    SendMessage(f->parent, WM_COMMAND, IDCANCEL, 0L);
                    return (0);
            }
            break;

        case WM_CHAR:
            if (f == NULL) return (0);

            switch (wParam) {
                case '\r':
                    if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
                        MessageBeep(MB_ICONHAND);
                        return (0);
                    }

                    if (CURPARA(global.active)->pagebreak) {
                        MessageBeep(MB_ICONHAND);
                        return (0);
                    }

                    if (GetKeyState(VK_CONTROL) < 0) {
                        if (SendMessage(hwnd, EM_LINELENGTH, 0, 0L) > 0) {
                            MessageBeep(MB_ICONHAND);
                        } else {
                            SendMessage(global.hwnd, WM_COMMAND, IDM_EDITPAGEBREAK, 0L);
                            SetFocus(global.imehwnd);
                        }
                        return (0);
                    }

                    if (GetKeyState(VK_SHIFT) < 0) {
                        if (SendMessage(hwnd, EM_LINELENGTH, 0, 0L) > 0) {
                            MessageBeep(MB_ICONHAND);
                        } else {
                            CharInput(global.active, '\n');
                        }
                        return(0);
                    }

                    SendMessage(f->parent, WM_COMMAND, IDOK, 0L);
                    return (0);

                case '\b':
                    if(unitlen(f->paragraph->text) != 0) {
                        CharInput(f, '\b');
                    } else if (global.active != NULL &&
                              (global.active->type & FN_NORMAL)) {

                        CharInput(global.active, '\b');
                    }
                    return(0);

                case '\t':
                    if (global.active != NULL && (global.active->type & FN_NORMAL)) {
                        CharInput(global.active, '\t');
                    }
                    return(0);

                case '\n':
                    /* Ctrl-Enter produces ^J???!!! */
                    if (LastKeyCode == VK_RETURN) {
                        SendMessage(hwnd, WM_CHAR, '\r', 0L);
                        return (0);
                    }
                    break;

                case ' ':
                    if (global.mode == M_KANA && (unitlen(f->paragraph->text) || SELECTING(f)))
                    {
                        if(SELECTING(f)) {
                            SendMessage(global.convhwnd, WM_KEYDOWN,
                                        (GetKeyState(VK_SHIFT) < 0) ? VK_LEFT : VK_RIGHT, 0L);
                            SendMessage(global.convhwnd, WM_KEYDOWN, VK_RETURN, 0L);
                        } else {
                            TakeCareOfThings(f, TRUE);
                            CharInput(f, '\0');
                            FlipHighlight(f);

                            SELPARA1(f) = SELPARA2(f) = f->paragraph;
                            SELPOS1(f) = 0;
                            SELPOS2(f) = unitlen(f->paragraph->text) - 1;
                            SELNOW(f) = FALSE;
                            SELTYPE(f) = SEL_SELECTION;
                            FlipHighlight(f);

                            ConvertNow(f, TRUE);
                        }
                    } else {
                        CharInput(f, ' ');
                    }
                    return(0);
            }

            break;

        case EM_REPLACESEL: {
            KANJI far *cp;
            POSITION p;
            KANJI buf[1] = { 0 };

            if (f == NULL) return (0);

            cp = (KANJI far *) lParam;
            if (cp == NULL) cp = (KANJI far *) buf;

            PARAOF(p) = f->paragraph;
            LINEOF(p) = f->paragraph->lines;
            POSOF(p) = 0;

            if (!CheckInputTextType(f, PARAOF(p)->text, cp)) return (0);

            TakeCareOfThings(f, TRUE);      /* Get rid of selections */

            ReplaceString(f, p, unitlen(f->paragraph->text), cp,
                            OP_REFORMAT | OP_MOVETOEND | OP_CHOOSEKANJI);

            SendMessage(hwnd, WM_KEYDOWN, VK_HOME, 0L);     /* Move to beginning */
            InvalidateRect(hwnd, NULL, TRUE);
            return (TRUE);
        }

        case WM_CLEAR: {
            POSITION p;

            if (f == NULL) return (0);

            PARAOF(p) = f->paragraph;
            LINEOF(p) = f->paragraph->lines;
            POSOF(p) = 0;

            TakeCareOfThings(f, TRUE);      /* Get rid of selections */

            DeleteString (f, p, unitlen(f->paragraph->text),
                            OP_REFORMAT | OP_UPDATE);

            InvalidateRect(hwnd, NULL, TRUE);
            return (TRUE);
        }

        case WM_DESTROY:
            CloseFile(f);
            FreeMem(f);
            ClearAllProperties(hwnd);
            return (0);

        default:
            break;
    }

    if (f == NULL) return (DefWindowProc(hwnd, message, wParam, lParam));

    curfile = f;

    return (CallWindowProc((FARPROC) FileWinProc, hwnd, message, wParam, lParam));
}

LONG FAR PASCAL IMEWinProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    KANJI far *line;
    FILEOPTIONS *f, *gf;
    HDC hdc;
    PAINTSTRUCT ps;
    CREATESTRUCT *cp;
    static HWND linehwnd, iconhwnd;

    switch (message) {
        case WM_CHAR:
            SendMessage(linehwnd, WM_CHAR, wParam, 0L);
            return (0);

        case WM_CREATE: {
        // create the edit line window
            linehwnd= CreateWindow(PROGNAME " Input Line Editor", "Input Line Editor", WS_CHILD | WS_VISIBLE,
                                    ICON_SIZE + 2*ICON_GAP, BOXGAP, 0, 0,
                                    hwnd, (HMENU) 1, hInstance, NULL);
            SendMessage(linehwnd, EM_SETMODIFY, FN_CONTROL | FN_IME | FN_EITHER, 0L);

        // get the fileoptions from the edit line window
            f = (FILEOPTIONS *) SendMessage(linehwnd, EM_GETHANDLE, 0, 0L);
            SetProp(hwnd, PROP_FILEOPTIONS, f);

        // create the global variable with the correct window height
            IMEWinHeight = LINEWIN_HEIGHT + 2*BOXGAP + 2*LINEGAP(f) + 1;

            cp = (CREATESTRUCT *) lParam;
            MoveWindow(hwnd, cp->x, cp->y, GetSystemMetrics(SM_CXFULLSCREEN),
                         IMEWinHeight, FALSE);

            iconhwnd = CreateWindow(PROGNAME " Input Mode Icon", "Mode Icon", WS_CHILD | WS_VISIBLE,
                                    ICON_GAP, (IMEWinHeight - ICON_SIZE) /2, ICON_SIZE, ICON_SIZE,
                                    hwnd, (HMENU) 2, hInstance, NULL);

        // bind the mode icon and edit line windows
            SendMessage(linehwnd, EM_SETRECT, 0, (LPARAM) iconhwnd);
            SendMessage(iconhwnd, EM_SETHANDLE, 0, (LPARAM) linehwnd);

            return (0);
        }

        case WM_SIZE: {
            FILEOPTIONS *f;
            int x, y, width, height;

            f = (FILEOPTIONS *) GetProp(linehwnd, PROP_FILEOPTIONS);
            if (f == NULL) return (0);

            x = (int) GetProp(linehwnd, PROP_SIZEX) + BOXGAP;
            y = (int) GetProp(linehwnd, PROP_SIZEY) + BOXGAP;
            width = LOWORD(lParam) - x - 2 * BOXGAP;
            height = SYSFONT->height + 2 * LINEGAP(f) + 1;

            MoveWindow(linehwnd, x, y, width, height, TRUE);
            return (0);
        }

        case WM_DESTROY: {
            DestroyWindow(iconhwnd);
            DestroyWindow(linehwnd);

            linehwnd = iconhwnd = NULL;
            ClearAllProperties(hwnd);

            return (0);
        }

        case WM_PAINT: {
            hdc = BeginPaint(hwnd, &ps);
            HilightDlgBox (hwnd, linehwnd, hdc, (GetFocus() == linehwnd));
            EndPaint(hwnd, &ps);
            return (0);
        }

        case WM_SETFOCUS:
            SetFocus(linehwnd);
            return(0);

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK:
                // Insert into the active window
                    line = (KANJI far *) SendMessage(linehwnd, EM_GETLINE, 0, 0L);

                    gf = global.active;

                    if (gf == NULL || !(gf->type & FN_NORMAL)) {
                        MessageBeep(MB_ICONHAND);
                        return (0);
                    }

                    if (CURPARA(gf)->pagebreak) {
                        MessageBeep(MB_ICONHAND);
                        return (0);
                    }

                    if (!CheckInputTextType(gf, PARAOF(gf->current)->text, line)) {
                        MessageBeep(MB_ICONHAND);
                        return (0);
                    }

                    if (kanjilen(line) > 0) {
                      /* erase Edit Box text */
                        if(SELPARA1(gf) != NULL) /* Delete block */
                            DeleteSelection(gf);

                        UndoAddTyping(gf, gf->current, kanjilen(line), FALSE);
                        InsertString(gf, gf->current, line, OP_REFORMAT | OP_UPDATE);

                        TakeCareOfThings(gf, TRUE);      /* Get rid of selections */

                        SendMessage(linehwnd, EM_SETSEL, 0, MAKELONG(0,-1));
                        SendMessage(linehwnd, WM_CLEAR, 0, 0L);
                    } else {
                        TakeCareOfThings(gf, TRUE);
                        CharInput(gf, '\r');
                    }

                    return (0);

                case IDCANCEL:
                // Erase the Line Editor's buffer
                    line = (KANJI far *) SendMessage(linehwnd, EM_GETLINE, 0, 0L);

                    if (kanjilen(line) > 0) {
                        SendMessage(linehwnd, EM_SETSEL, 0, MAKELONG(0,-1));
                        SendMessage(linehwnd, WM_CLEAR, 0, 0L);
                    } else {
                        MessageBeep(MB_ICONEXCLAMATION);
                        return (0);
                    }

                    return (0);

                case IDM_TOGGLEMODE:
                    InvalidateRect(iconhwnd, NULL, TRUE);
                    UpdateWindow(iconhwnd);
                    return (0);
            }

    }
    return (DefWindowProc(hwnd, message, wParam, lParam));
}


