/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"

#include "idm.h"
#include <stdarg.h>
#include <shellapi.h>
#ifdef CTL3D
    #include <ctl3d.h>
#endif CTL3D


GLOBALOPTIONS   global;
FILEOPTIONS     *fileoptions = NULL;
FILEOPTIONS     *curfile = NULL;
KANJIFONT       *InstalledFonts = NULL;
ASCIIFONT       DefAsciiFont = { "Times New Roman", 12 };
HANDLE          hInstance = NULL;
HMENU           hmenu = NULL;
HMENU           FullMenu, MinimalMenu;
BOOL            Dialogs3D = FALSE;
HWND            hKanjiLookupDlg;               // DCY

static int      Untitled = 0;
static BOOL     AllInitialized = FALSE;

static TOOLBARICON toolbar[] = {
    { ID_KANA, 180, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 1, 0, NULL, "Kana_Undep", "Kana_Dep", "Kana_Grayed", "Kana_Pressed" },
    { ID_ASCII, 210, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 0, 0, NULL, "Ascii_Undep", "Ascii_Dep", "Ascii_Grayed", "Ascii_Pressed" },
    { ID_JASCII, 250, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 0, 0, NULL, "JAscii_Undep", "JAscii_Dep", "JAscii_Grayed", "JAscii_Pressed" },
    { ID_CONVERT, 290, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 0, 1, "CVT_Disabled", "CVT_Undep", NULL, NULL, "CVT_Pressed" },
    { ID_BOLD, 330, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, -1, 2, "Bold_Disabled", "Bold_Undep", "Bold_Dep", "Bold_Grayed", "Bold_Pressed" },
    { ID_ITALICS, 360, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, -1, 2, "Italics_Disabled", "Italics_Undep", "Italics_Dep", "Italics_Grayed", "Italics_Pressed" },
    { ID_UNDERLINE, 390, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, -1, 2, "Und_Disabled", "Und_Undep", "Und_Dep", "Und_Grayed", "Und_Pressed" },
    { ID_REVERSE, 420, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, -1, 2, "Rev_Disabled", "Rev_Undep", "Rev_Dep", "Rev_Grayed", "Rev_Pressed" },
//  { ID_HOLLOW, 450, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, -1, 2, "Hollow_Disabled", "Hollow_Undep", "Hollow_Dep", "Hollow_Grayed", "Hollow_Pressed" },
//  { ID_SHADOW, 490, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, -1, 2, "Shadow_Disabled", "Shadow_Undep", "Shadow_Dep", "Shadow_Grayed", "Shadow_Pressed" },
    { ID_GLOSSARY, 460, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 0, 0, NULL, "Glossary_Undep", "Glossary_Dep", "Glossary_Grayed", "Glossary_Pressed" },
    { ID_SPECIAL, 490, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 0, 0, NULL, "Special_Undep", "Special_Dep", "Special_Grayed", "Special_Pressed" },
    { ID_LEFTRIGHT, 530, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 1, 0, "LR_Disabled", "LR_Undep", "LR_Dep", "LR_Grayed", "LR_Pressed" },
    { ID_UPDOWN, 560, TBBORDER+2, TBICONWIDTH, TBICONHEIGHT, 0, 0, "UD_Disabled", "UD_Undep", "UD_Dep", "UD_Grayed", "UD_Pressed" }
};


LONG FAR PASCAL MainWinProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL FileParentWinProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL FileWinProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL JeditProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL JtextProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL RulerProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL ConvProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL StatusProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL StatisticsProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL FileOpenConfirmDlgProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL FileSaveFormatProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL FileSaveAsDlgProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL DictionaryProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL DictionaryOptionsProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL BigKanjiProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL KanjiInfoProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL MoreKanjiInfoProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL JISInputProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL LookupProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL LookupListProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL JISTableDlgProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL JISTableProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL HotKeysProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL ChangeHotKeyProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL OptionsProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL GlossaryProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL GlossaryListProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL AboutProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL SimpleProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL InputModeIconProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL InstalledFontsProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL UserConversionProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL EditUserConversionProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL SearchProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL ReplaceProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL ReplaceDlgProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL PageSetupProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL SummaryProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL HeaderProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL EditHeaderProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL FileFindProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL RunProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL TemplateProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL FormatParagraphProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL RadLookupProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL RadLookupListProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL IMELineProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL IMEWinProc (HWND, UINT, WPARAM, LPARAM);
extern LONG FAR PASCAL GotoProc (HWND, UINT, WPARAM, LPARAM);
extern BOOL FAR PASCAL SignOnProc (HWND, UINT, WPARAM, LPARAM);


static BOOL InitializeEverything(HWND hwnd)
{
    fileoptions = NULL;

    InitPrinting();

    if (!ReadOptionsFile()) {
        ErrorMessage(hwnd, "Error in program initialization!");
        return (FALSE);
    }

    SetupPrinter(TRUE, global.pagesetup.landscape);

    if (InitFonts() <= 0) {
        ErrorMessage(hwnd, "Cannot find any kanji fonts!");
        return (FALSE);
    }
    if (!InitConversion()) {
        ErrorMessage(hwnd, "Cannot initialize kana-kanji conversion!");
        return (FALSE);
    }
    ReadGlossary(global.glossary);
    ReadConversionCache(global.convcache);
    InitDDE(FALSE);
    return (TRUE);
}



void InitFileWindow(LPSTR lpszCmdParam)
{
    char buffer[MAXLINELEN];
    char word[MAXLINELEN];
    char remains[MAXLINELEN];

    word[0] = remains[0] = '\0';
    _fstrcpy(buffer, lpszCmdParam);
    sscanf(buffer, "%s %[^\n]", word, remains);

    if (!word[0]) {
        global.active = NewFile (FN_NORMAL, TRUE);

        if (global.active != NULL) {
            char buffer[50];

            global.active->filename[0] = '\0';
            sprintf(buffer, "Untitled #%d", ++Untitled);
            ShowFileWindow(buffer, global.active);
        }
    } else {
        do {
#ifdef USEWIN32
            if (word[0] == '\"') {
                word[0] = remains[0] = '\0';
                sscanf(buffer, "%*[\"]%[^\n\"]%*[\"] %[^\n]", word, remains);
            }
#endif USEWIN32

            if (!DoFileOpen(word))
                ErrorMessage(global.hwnd, "Cannot open file '%s'.", word);

            strcpy(buffer, remains);
            word[0] = remains[0] = '\0';
            sscanf(buffer, "%s %[^\n]", word, remains);
        } while (word[0]);
    }
}


#define LOGFILENAME "logfile.log"

void LogMessage(char *fmt, ...)
{
    va_list argptr;
    FILE *fp;
    char buffer[MAXLINELEN];

    if (fmt == NULL) return;

    sprintf(buffer, "%s" LOGFILENAME, global.jwppath);

    fp = fopen(buffer, "a");
    if (fp == NULL) return;

    va_start(argptr, fmt);
    vsprintf(buffer, fmt, argptr);
    va_end(argptr);

    //_InitEasyWin();
    //printf("%s\n", buffer);
    fprintf(fp, "%s\n", buffer);
    fclose(fp);
}


void ErrorMessage(HWND hwnd, char *fmt, ...)
{
    va_list argptr;
    char buffer[MAXLINELEN];

    va_start(argptr, fmt);
    vsprintf(buffer, fmt, argptr);
    va_end(argptr);

    MessageBeep(MB_ICONEXCLAMATION);
    MessageBox (hwnd, buffer, "Something is wrong...", MB_ICONEXCLAMATION | MB_OK);
}



int Confirmation (HWND hwnd, char *fmt, ...)
{
    va_list argptr;
    char buffer[MAXLINELEN];

    va_start(argptr, fmt);
    vsprintf(buffer, fmt, argptr);
    va_end(argptr);

    MessageBeep(MB_ICONQUESTION);
    return (MessageBox (hwnd, buffer, "Please Confirm...", MB_ICONQUESTION | MB_YESNOCANCEL));
}



int YesNo(HWND hwnd, char *fmt, ...)
{
    va_list argptr;
    char buffer[MAXLINELEN];

    va_start(argptr, fmt);
    vsprintf(buffer, fmt, argptr);
    va_end(argptr);

    MessageBeep(MB_ICONQUESTION);
    return (MessageBox (hwnd, buffer, "Please Confirm...", MB_ICONQUESTION | MB_YESNO));
}



BOOL RetryMessage(char *fmt, ...)
{
    va_list argptr;
    char buffer[MAXLINELEN];

    va_start(argptr, fmt);
    vsprintf(buffer, fmt, argptr);
    va_end(argptr);

    MessageBeep(MB_ICONEXCLAMATION);
    if (MessageBox (global.hwnd, buffer, "Something is wrong...",
        MB_ICONEXCLAMATION | MB_RETRYCANCEL) == IDCANCEL) return (FALSE);
    else
        return (TRUE);
}



void PlayExternalSoundFile (char *filename)
{
#ifdef USEWIN32
    char buffer[MAXLINELEN];

    sprintf(buffer, "%s%s%s.WAV", global.jwppath, PROGNAME, filename);
    sndPlaySound(buffer, //hInstance,
               SND_ASYNC | SND_NODEFAULT | SND_NOWAIT | SND_RESOURCE);
#endif USEWIN32
}



static void AdjustAllChildWindows (int width, int height)
{
    int i;
    extern int IMEWinHeight;
    int ToolBarHeight = 0, StatusBarHeight = 0, IMEHeight = 0;

    if (global.tbhwnd != NULL) {
        MoveWindow(global.tbhwnd, -1, -1, width + 2, TBHEIGHT + 2, TRUE);
        ToolBarHeight = TBHEIGHT + 1;
    }

    if (global.statushwnd != NULL) {
        MoveWindow(global.statushwnd, -1, height - STATUSHEIGHT,
                    width + 2, STATUSHEIGHT + 2, TRUE);
        StatusBarHeight = STATUSHEIGHT;
    }

    if (global.imehwnd != NULL) {
        IMEHeight = IMEWinHeight + 1;
    }

    if (global.clienthwnd == NULL || global.convhwnd == NULL) return;


    switch (global.listposition) {
        case 0: MoveWindow(global.clienthwnd, 0, ToolBarHeight, width,
                            height - StatusBarHeight - ToolBarHeight - IMEHeight, TRUE);

                if (global.imehwnd) {
                    MoveWindow(global.imehwnd, -1, height - StatusBarHeight - IMEHeight,
                                width + 2, IMEHeight, TRUE);
                }

                if (!(GetWindowLong(global.convhwnd, GWL_STYLE) & WS_POPUP)) {
                    DestroyWindow(global.convhwnd);
                    global.convhwnd = CreateWindow(PROGNAME " Conversion Bar", NULL,
                                                    WS_THICKFRAME | WS_VISIBLE | WS_POPUP,
                                                    global.convbar.left, global.convbar.top,
                                                    global.convbar.right, 0,
                                                    global.hwnd, NULL, hInstance, NULL);
                }
                break;

        case 1: i = SYSFONT->height + 2 * BORDERSPACE +
                        GetSystemMetrics(SM_CXVSCROLL) +
                        2 * GetSystemMetrics(SM_CYBORDER);

                if (GetWindowLong(global.convhwnd, GWL_STYLE) & WS_POPUP) {
                    DestroyWindow(global.convhwnd);
                    global.convhwnd = CreateWindow(PROGNAME " Conversion Bar", NULL,
                                                    WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE | WS_BORDER,
                                                    -1, ToolBarHeight - 1, width + 2, i,
                                                    global.hwnd, (HMENU) 3, hInstance, NULL);
                } else {
                    MoveWindow(global.convhwnd, -1, ToolBarHeight - 1, width + 2, i, TRUE);

                }

                if (global.imehwnd) {
                    MoveWindow(global.imehwnd, -1, ToolBarHeight + i - 2,
                                width + 2, IMEHeight, TRUE);
                }

                MoveWindow(global.clienthwnd, 0, ToolBarHeight + i + IMEHeight - 1, width,
                            height - StatusBarHeight - ToolBarHeight - IMEHeight - i, TRUE);

                break;

        case 2: i = SYSFONT->height + 2 * BORDERSPACE +
                        GetSystemMetrics(SM_CXVSCROLL) +
                        2 * GetSystemMetrics(SM_CYBORDER);

                if (GetWindowLong(global.convhwnd, GWL_STYLE) & WS_POPUP) {
                    DestroyWindow(global.convhwnd);
                    global.convhwnd = CreateWindow(PROGNAME " Conversion Bar", NULL,
                                                    WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE | WS_BORDER,
                                                    -1, height - StatusBarHeight - i, width + 2, i,
                                                    global.hwnd, (HMENU) 3, hInstance, NULL);
                } else {
                    MoveWindow(global.convhwnd, -1, height - StatusBarHeight - i,
                                width + 2, i, TRUE);
                }

                if (global.imehwnd) {
                    MoveWindow(global.imehwnd, -1, height - StatusBarHeight - IMEHeight - i + 1,
                                width + 2, IMEHeight, TRUE);
                }

                MoveWindow(global.clienthwnd, 0, ToolBarHeight , width,
                            height - StatusBarHeight - ToolBarHeight - IMEHeight  - i + 1, TRUE);
                break;

    }

    InvalidateRect(global.hwnd, NULL, TRUE);
    InvalidateRect(global.clienthwnd, NULL, TRUE);
    InvalidateRect(global.imehwnd, NULL, TRUE);
}



int PASCAL WinMain (HANDLE hThisInstance, HANDLE hPrevInstance,
                    LPSTR lpszCmdParam, int nCmdShow)
{
    MSG                 msg;
    HDC                 hdc;
    WNDCLASS            wndclass;
    RECT                rect;
    HCURSOR             hcursor;
    HBITMAP             hbitmap;
    CLIENTCREATESTRUCT  clientcreate;
    HWND                signonhwnd;
    HANDLE              haccel;
    extern BOOL         Read3DDialogOption(void);


    hInstance = hThisInstance;

    memset(&global, 0, sizeof(GLOBALOPTIONS));

    hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
    ShowCursor(TRUE);

    FullMenu = LoadMenu(hInstance, "FileMenu");
    MinimalMenu = LoadMenu(hInstance, "NoFileMenu");


#ifdef CTL3D
    /* 3D dialogs */

    Dialogs3D = Read3DDialogOption();

    if (Dialogs3D) {
        Dialogs3D = Ctl3dRegister(hInstance);
        if (Dialogs3D) Ctl3dAutoSubclass(hInstance);
    }
#endif CTL3D


    /* Register window classes */

    if (!hPrevInstance) {
        wndclass.style          = CS_HREDRAW | CS_VREDRAW;
        wndclass.lpfnWndProc    = MainWinProc;
        wndclass.cbClsExtra     = 0;
        wndclass.cbWndExtra     = 0;
        wndclass.hInstance      = hThisInstance;
        wndclass.hIcon          = LoadIcon(hThisInstance, "Icon1");
        wndclass.hCursor        = LoadCursor(NULL, IDC_ARROW);
        wndclass.hbrBackground  = COLOR_APPWORKSPACE + 1;
        wndclass.lpszMenuName   = NULL;
        wndclass.lpszClassName  = PROGNAME " Frame";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = FileParentWinProc;
        wndclass.hIcon          = LoadIcon(hThisInstance, "Icon3");
        wndclass.hbrBackground  = COLOR_WINDOW + 1;
        wndclass.lpszClassName  = PROGNAME " File Parent";

        RegisterClass(&wndclass);

        wndclass.style          = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        wndclass.lpfnWndProc    = FileWinProc;
        wndclass.hCursor        = LoadCursor(NULL, IDC_IBEAM);
        wndclass.lpszClassName  = PROGNAME " File";

        RegisterClass(&wndclass);

        wndclass.style          = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        wndclass.lpfnWndProc    = JeditProc;
        wndclass.hIcon          = NULL;
        wndclass.lpszClassName  = PROGNAME " Japanese Edit Control";

        RegisterClass(&wndclass);

        wndclass.style          = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        wndclass.lpfnWndProc    = IMELineProc;
        wndclass.hIcon          = NULL;
        wndclass.hCursor        = LoadCursor(NULL, IDC_IBEAM);
        wndclass.lpszClassName  = PROGNAME " Input Line Editor";

        RegisterClass(&wndclass);

        wndclass.style          = WS_CHILD | CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
        wndclass.lpfnWndProc    = IMEWinProc;
        wndclass.hIcon          = NULL;
        wndclass.lpszClassName  = PROGNAME " Input Line Editor Window";

        RegisterClass(&wndclass);

        wndclass.hCursor        = LoadCursor(NULL, IDC_ARROW);
        wndclass.lpfnWndProc    = RulerProc;
        wndclass.lpszClassName  = PROGNAME " Ruler";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = LookupListProc;
        wndclass.lpszClassName  = PROGNAME " Kanji Lookup List";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = JtextProc;
        wndclass.hbrBackground  = NULL;
        wndclass.lpszClassName  = PROGNAME " Text Box";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = BigKanjiProc;
        wndclass.lpszClassName  = PROGNAME " Big Kanji";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = JISTableProc;
        wndclass.lpszClassName  = PROGNAME " JIS Table";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = InputModeIconProc;
        wndclass.lpszClassName  = PROGNAME " Input Mode Icon";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = ConvProc;
        wndclass.hbrBackground  = NULL;
        wndclass.lpszClassName  = PROGNAME " Conversion Bar";

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc    = StatusProc;
        wndclass.lpszClassName  = PROGNAME " Status";

        RegisterClass(&wndclass);

        wndclass.style         = CS_HREDRAW | CS_VREDRAW ;      // DCY
        wndclass.hIcon         = LoadIcon(hThisInstance, "RadIcon");
        wndclass.hbrBackground = COLOR_BTNFACE + 1;
        wndclass.lpfnWndProc   = RadLookupProc ;
        wndclass.lpszMenuName  = NULL ;
        wndclass.lpszClassName = PROGNAME " Radical Lookup" ;

        RegisterClass(&wndclass);

        wndclass.lpfnWndProc   = RadLookupListProc;
        wndclass.hbrBackground = COLOR_WINDOW + 1;
        wndclass.lpszClassName = PROGNAME " Radical Kanji Lookup List";

        RegisterClass(&wndclass);
    }


    global.clienthwnd = NULL;

    {
        BOOL maximized;
        int x, y, width, length;

        maximized = ReadProgramSizeOptions (&x, &y, &width, &length);

        if (maximized) {
            global.hwnd = CreateWindowEx(WS_EX_ACCEPTFILES,
                                         PROGNAME " Frame", PROGNAME " " JWPVERSION,
                                         WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_MAXIMIZE,
                                         CW_USEDEFAULT, CW_USEDEFAULT, width, length,
                                         NULL, NULL, hThisInstance, NULL);
            nCmdShow = SW_SHOWMAXIMIZED;
        } else {
            global.hwnd = CreateWindowEx(WS_EX_ACCEPTFILES,
                                         PROGNAME " Frame", PROGNAME " " JWPVERSION,
                                         WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
                                         x, y, width, length,
                                        NULL, NULL, hThisInstance, NULL);
            nCmdShow = SW_SHOW;
        }
        DragAcceptFiles(global.hwnd, TRUE);
    }

    ShowWindow(global.hwnd, nCmdShow);
    UpdateWindow(global.hwnd);


    /* Play the sign-on sound and create the sign-on message */

    hbitmap = LoadBitmap(hInstance, "TITLE16");

    if (hbitmap == NULL) {
        signonhwnd = CreateDialog(hInstance, "SignOn", global.hwnd, SimpleProc);
    } else {
        DeleteObject(hbitmap);
        signonhwnd = CreateDialog(hInstance, "PictureFrame", global.hwnd, SignOnProc);
    }


    /* Load accelerators */

    haccel = LoadAccelerators(hInstance, "Accelerator");


    /* Initialize variables */

    global.mode = M_KANA;
    global.jascii = FALSE;
    global.active = NULL;
    global.convsel = -1;
    global.glisthwnd = NULL;
    global.imeactive = FALSE;
    global.vertical = FALSE;

    global.cpos = (POINT far *) BlockAlloc(CPOSPAGESIZE);
    global.cpossize = CPOSPAGESIZE;
    global.cpospara = NULL;
    global.cposcount = 0;

    if (!InitializeEverything(global.hwnd)) {
        PlayExternalSoundFile("ERROR");
        DestroyWindow(signonhwnd);
        return (-1);
    }

    PlayExternalSoundFile("START");

    global.leading = (int) ((double) PRINTFONT->leading * global.dispscale);
    if (global.leading % 2 != 0) global.leading++;

    global.spacing = (int) ((double) PRINTFONT->spacing * global.dispscale);
    if (global.spacing % 2 != 0) global.spacing++;

    SwitchMenu(0);      /* Full menu */


    /* Take away the sign-on message */

    DestroyWindow(signonhwnd);


    /* Others */

    global.gloslistdlg = MakeProcInstance(GlossaryListProc, hInstance);

    hdc = GetDC(global.hwnd);
    GetTextMetrics(hdc, &(global.textmetric));
    global.borderspace = AVGWIDTH;
    ReleaseDC(global.hwnd, hdc);


    if (global.draftview)   CheckMenuItem(hmenu, IDM_VIEWDRAFT, MF_CHECKED | MF_BYCOMMAND);
    if (global.showime)     SendMessage(global.hwnd, WM_COMMAND, IDM_VIEWIME, 0L);
    if (global.showstatus)  SendMessage(global.hwnd, WM_COMMAND, IDM_VIEWSTATUS, 0L);
    if (global.showribbon)  SendMessage(global.hwnd, WM_COMMAND, IDM_VIEWRIBBON, 0L);
    if (global.showruler)   CheckMenuItem(hmenu, IDM_VIEWRULER, MF_CHECKED | MF_BYCOMMAND);
    if (global.showspecial) SendMessage(global.tbhwnd, BM_SETSTATE, 1, ID_SPECIAL);


    /* Create the conversion bar */

    switch (global.listposition) {
        case 0: global.convhwnd = CreateWindow(PROGNAME " Conversion Bar", "Kanji List",
                                                WS_VISIBLE | WS_THICKFRAME | WS_POPUP,
                                                0, 0, 0, 0, global.hwnd, NULL, hInstance, NULL);
                ShowWindow(global.convhwnd, SW_SHOW);
                break;
        case 1:
        case 2: global.convhwnd = CreateWindow(PROGNAME " Conversion Bar", NULL,
                                                WS_CHILD | WS_VISIBLE | WS_BORDER,
                                                0, 0, 0, 0, global.hwnd, (HMENU) 3, hInstance, NULL);
                break;
    }


    /* Create the Client window */

    clientcreate.hWindowMenu = GetSubMenu(FullMenu, 6);
    clientcreate.idFirstChild = ID_FIRSTFILE;

    global.clienthwnd = CreateWindowEx (WS_EX_ACCEPTFILES,
                                        "MDICLIENT", NULL,
                                        WS_CHILD | WS_VISIBLE,
                                        0, 0, 0, 0, global.hwnd, (HMENU) 1, hInstance,
                                        (LPSTR) &clientcreate);
    DragAcceptFiles(global.clienthwnd, TRUE);


    /* Create the glossary list */

    if (global.glistvisible) SendMessage(global.hwnd, WM_COMMAND, IDM_VIEWGLOSSARY, 0L);

    SetActiveWindow(global.hwnd);


    GetClientRect(global.hwnd, &rect);
    AllInitialized = TRUE;
    AdjustAllChildWindows(rect.right, rect.bottom);
    UpdateWindow(global.hwnd);

    ShowCursor(FALSE);
    SetCursor(hcursor);


    /* Now create an empty file window? */

    InitFileWindow(lpszCmdParam);

    if (Dialogs3D != global.ctl3d) {
        ErrorMessage(global.hwnd, PROGNAME " " JWPVERSION " cannot create 3D dialog boxes!");
    }


    /* Message loop */

    while (GetMessage(&msg, NULL, 0, 0)) {
        if (!(IsWindow(hKanjiLookupDlg) &&     // DCY
             IsDialogMessage(hKanjiLookupDlg, &msg))) {
                 if (!TranslateAccelerator (global.hwnd, haccel, &msg)) {
                     TranslateMessage(&msg);
                     DispatchMessage(&msg);
                }
        }
     }

#ifdef CTL3D
    if (Dialogs3D) Ctl3dUnregister(hInstance);
#endif CTL3D

    return (msg.wParam);
}



void CenterDialogBox (HWND hwnd)
{
    RECT rect;

    GetWindowRect(hwnd, &rect);
    OffsetRect(&rect, -rect.left, -rect.top);
    MoveWindow(hwnd, ((GetSystemMetrics(SM_CXSCREEN) - rect.right ) / 2 + 4) & ~7,
                     (GetSystemMetrics(SM_CYSCREEN) - rect.bottom) / 2,
                     rect.right, rect.bottom, FALSE);
}



BOOL FAR PASCAL CloseAllEnumProc (HWND hwnd, LONG lParam)
{
    FILEOPTIONS *f;

    if (GetWindow(hwnd, GW_OWNER)) return (TRUE);      /* Check for icon title */

    for (f = fileoptions; f != NULL; f = f->next) if (f->parent == hwnd) break;

    if (f == NULL) return (TRUE);       /* Not a valid window */


    SendMessage(global.clienthwnd, WM_MDIMAXIMIZE, (WPARAM) hwnd, 0L);

    if (!SendMessage(hwnd, WM_QUERYENDSESSION, 0, 0L)) return (FALSE);

    SendMessage(global.clienthwnd, WM_MDIDESTROY, (WPARAM) hwnd, 0L);

    return (TRUE);
}



BOOL RedisplayFile (FILEOPTIONS *f)
{
    HDC hdc, hdcmem;
    HBITMAP hbitmap;
    HBRUSH hbrush;
    RECT rect;

    if (IsIconic(f->hwnd) || IsIconic(f->parent)) return (FALSE);

    hdc = GetDC(f->hwnd);
    hdcmem = CreateCompatibleDC(hdc);
    SetTextColor(hdcmem, GetSysColor(COLOR_WINDOWTEXT));
    SetBkColor(hdcmem, GetSysColor(COLOR_WINDOW));

    GetClientRect(f->hwnd, &rect);

    hbitmap = CreateCompatibleBitmap (hdc, rect.right, rect.bottom);
    if (hbitmap == NULL) {
        InvalidateRect(f->hwnd, NULL, TRUE);
        UpdateWindow(f->hwnd);
        return (FALSE);
    }

    SelectObject(hdcmem, hbitmap);
    hbrush = CreateSolidBrush(GetSysColor(COLOR_WINDOW));
    FillRect(hdcmem, &rect, hbrush);
    DeleteObject(hbrush);

    HideCaret(f->hwnd);
    RedrawFile(f, hdcmem, 0, 0, &rect, global.showspecial, TRUE);

    BitBlt(hdc, 0, 0, rect.right, rect.bottom, hdcmem, 0, 0, SRCCOPY);

    if (f == global.active) {
        if (FindCaret(f, TRUE))
            DoCaret(f, CURX(f), CURY(f) - CURLINE(f)->height, TRUE);
        else
            DoCaret(f, 0, 0, FALSE);
    }

    ShowCaret(f->hwnd);

    ReleaseDC(f->hwnd, hdc);
    DeleteDC(hdcmem);
    DeleteObject(hbitmap);

    return (TRUE);
}



LONG FAR PASCAL MainWinProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int         i;
    RECT        rect;
    char        buffer[15];
    FILEOPTIONS *f;
    static      LastSearchIsReplace = FALSE;


    /* Takes care of menu items */

    switch (message) {
    case WM_MOVE:
        if (AllInitialized) {
            GetWindowRect(hwnd, &rect);

            if (global.glisthwnd != NULL) {
                SetWindowPos(global.glisthwnd, NULL,
                             rect.left + global.glistsize.left,
                             rect.top + global.glistsize.top,
                             0, 0, SWP_NOSIZE | SWP_NOZORDER);
            }
            if (global.convhwnd != NULL && global.listposition == 0) {
                SetWindowPos(global.convhwnd, NULL,
                             rect.left + global.convbar.left,
                             rect.top + global.convbar.top,
                             0, 0, SWP_NOSIZE | SWP_NOZORDER);
            }

            SetActiveWindow(global.hwnd);
        }
        break;

    case WM_SIZE:
        if (AllInitialized) {
            AdjustAllChildWindows (LOWORD(lParam), HIWORD(lParam));

            for (f = fileoptions; f != NULL; f = f->next) {
                SELPARA1(f) = SELPARA2(f) = NULL;
                SELPOS1(f) = SELPOS2(f) = 0;
                SELTYPE(f) = SEL_SELECTION;
                SELNOW(f) = FALSE;
            }
        }
        return (0);

    case WM_KEYDOWN:
    case WM_CHAR:
        if (global.active != NULL)
            SendMessage(global.active->hwnd, message, wParam, lParam);
        return (0);

    case WM_INITMENUPOPUP: {
        if (hmenu == MinimalMenu) {
            /* Nothing */
        } else if (hmenu == FullMenu) {
            if (global.active == NULL) return (0);

            switch (LOWORD(lParam)) {
                case 1:     /* File */
                    if (global.hdevnames == NULL || global.hdevmode == NULL) {
                        EnableMenuItem(hmenu, IDM_FILEPRINT, MF_BYCOMMAND | MF_GRAYED);
                    }
                    UpdateQuickFiles();
                    break;

                case 2:     /* Edit */
                    if (SELPARA1(global.active) == NULL) {
                        EnableMenuItem(hmenu, IDM_EDITCUT, MF_BYCOMMAND | MF_GRAYED);
                        EnableMenuItem(hmenu, IDM_EDITCOPY, MF_BYCOMMAND | MF_GRAYED);
                    } else {
                        EnableMenuItem(hmenu, IDM_EDITCUT, MF_BYCOMMAND | MF_ENABLED);
                        EnableMenuItem(hmenu, IDM_EDITCOPY, MF_BYCOMMAND | MF_ENABLED);
                    }
                    if (ClipboardFull()) {
                        EnableMenuItem(hmenu, IDM_EDITPASTE, MF_BYCOMMAND | MF_ENABLED);
                    } else {
                        EnableMenuItem(hmenu, IDM_EDITPASTE, MF_BYCOMMAND | MF_GRAYED);
                    }
                    if (global.active->undo == NULL) {
                        EnableMenuItem(hmenu, IDM_EDITUNDO, MF_BYCOMMAND | MF_GRAYED);
                    } else {
                        EnableMenuItem(hmenu, IDM_EDITUNDO, MF_BYCOMMAND | MF_ENABLED);
                    }
                    if (global.active->redo == NULL) {
                        EnableMenuItem(hmenu, IDM_EDITREDO, MF_BYCOMMAND | MF_GRAYED);
                    } else {
                        EnableMenuItem(hmenu, IDM_EDITREDO, MF_BYCOMMAND | MF_ENABLED);
                    }
                    if (global.active->type & FN_NORMAL) {
                        EnableMenuItem(hmenu, IDM_EDITPAGEBREAK, MF_BYCOMMAND | MF_ENABLED);
                        EnableMenuItem(hmenu, IDM_EDITSOFTRETURN, MF_BYCOMMAND | MF_ENABLED);
                        EnableMenuItem(hmenu, IDM_EDITGOTO, MF_BYCOMMAND | MF_ENABLED);
                    } else {
                        EnableMenuItem(hmenu, IDM_EDITPAGEBREAK, MF_BYCOMMAND | MF_GRAYED);
                        EnableMenuItem(hmenu, IDM_EDITSOFTRETURN, MF_BYCOMMAND | MF_GRAYED);
                        EnableMenuItem(hmenu, IDM_EDITGOTO, MF_BYCOMMAND | MF_GRAYED);
                    }
                    break;

                case 4:     /* Kanji */
                    if (SELPARA1(global.active) == NULL) {
                        EnableMenuItem(hmenu, IDM_KANJIINFO, MF_BYCOMMAND | MF_GRAYED);
                    } else {
                        EnableMenuItem(hmenu, IDM_KANJIINFO, MF_BYCOMMAND | MF_ENABLED);
                    }
                    if (SELPARA1(global.active) == NULL) {
                        EnableMenuItem(hmenu, IDM_KANJICONVERT, MF_BYCOMMAND | MF_GRAYED);
                    } else {
                        EnableMenuItem(hmenu, IDM_KANJICONVERT, MF_BYCOMMAND | MF_ENABLED);
                    }
                    break;

                case 5:     /* Format */
                    if (!global.active->repage) {
                        EnableMenuItem(hmenu, IDM_FORMATREPAGE, MF_BYCOMMAND | MF_GRAYED);
                    } else {
                        EnableMenuItem(hmenu, IDM_FORMATREPAGE, MF_BYCOMMAND | MF_ENABLED);
                    }
                    if (!(global.active->type & FN_NORMAL)) {
                        CheckMenuItem(hmenu, IDM_FORMATVERTICAL, (global.vertical ? MF_CHECKED : MF_UNCHECKED) | MF_BYCOMMAND);
                    } else {
                        CheckMenuItem(hmenu, IDM_FORMATVERTICAL, (global.active->vertical ? MF_CHECKED : MF_UNCHECKED) | MF_BYCOMMAND);
                    }
                    break;
            }
        }
        return (0);
    }

    case WM_COMMAND:

        if (HIBYTE(WINCMDID) == ID_TOOLBAR) {
            switch (LOBYTE(WINCMDID)) {
                case ID_GLOSSARY:
                    SendMessage(hwnd, WM_COMMAND, IDM_VIEWGLOSSARY, 0L);
                    break;

                case ID_SPECIAL:
                    SendMessage(hwnd, WM_COMMAND, IDM_VIEWSPECIAL, 0L);
                    break;

                case ID_KANA:
                    if (global.mode != M_KANA) ToggleInputMode();
                    if (global.active != NULL)
                        TakeCareOfThings(global.active, FALSE);

                    if(global.imehwnd)
                        SendMessage(global.imehwnd, WM_COMMAND, IDM_TOGGLEMODE, 0L);
                    break;

                case ID_ASCII:
                    if (global.mode != M_ASCII) ToggleInputMode();
                    if (global.active != NULL)
                        TakeCareOfThings(global.active, FALSE);

                    if(global.imehwnd)
                        SendMessage(global.imehwnd, WM_COMMAND, IDM_TOGGLEMODE, 0L);
                    break;

                case ID_JASCII:
                    SendMessage(global.hwnd, WM_COMMAND, IDM_TOGGLEJASCII, 0L);
                    break;

                case ID_LEFTRIGHT:
                    if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
                        global.vertical = FALSE;
                    } else {
                        global.active->vertical = FALSE;
                        global.active->changed = TRUE;
                    }
                    SendMessage(global.tbhwnd, BM_SETSTATE, 1, ID_LEFTRIGHT);
                    SendMessage(global.tbhwnd, BM_SETSTATE, 0, ID_UPDOWN);
                    break;

                case ID_UPDOWN:
                    if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
                        global.vertical = TRUE;
                    } else {
                        global.active->vertical = TRUE;
                        global.active->changed = TRUE;
                    }
                    SendMessage(global.tbhwnd, BM_SETSTATE, 0, ID_LEFTRIGHT);
                    SendMessage(global.tbhwnd, BM_SETSTATE, 1, ID_UPDOWN);
                    break;

                case ID_CONVERT:
                    if (global.active != NULL) {
                        SendMessage(global.active->hwnd, WM_COMMAND, IDM_KANJICONVERT, 0L);
                    }
                    break;
            }

            return (0);
        }


        switch (WINCMDID) {
            case IDM_EDITUNDO:
            case IDM_EDITREDO:
            case IDM_EDITCUT:
            case IDM_EDITCOPY:
            case IDM_EDITPASTE:
            case IDM_EDITCLEAR:
            case IDM_EDITSELECTALL:
            case IDM_KANJICONVERT:
            case IDM_KANJILCONVERT:
            case IDM_KANJIINFO:
                if (global.active != NULL)
                    SendMessage(global.active->hwnd, WM_COMMAND, wParam, 0L);
                return (0);

            case IDM_EDITPAGEBREAK:
            case IDM_EDITSOFTRETURN:
            case IDM_EDITGOTO:
                if (global.active != NULL && (global.active->type & FN_NORMAL))
                    SendMessage(global.active->hwnd, WM_COMMAND, wParam, 0L);
                return (0);

            case IDM_TOGGLEMODE:
                ToggleInputMode();
                if (global.active != NULL)
                    TakeCareOfThings(global.active, TRUE);
                return (0);

            case IDM_TOGGLEJASCII:
                global.jascii = !global.jascii;
                SendMessage(global.tbhwnd, BM_SETSTATE, global.jascii ? 1 : 0, ID_JASCII);
                if (global.jascii) {
                    StatusMessage("Switched to JASCII mode");
                } else {
                    StatusMessage("Switched to normal ASCII mode");
                }
                return (0);

            case IDM_TOGGLEEDITWIN:  {
                if (!global.imeactive && global.imehwnd == NULL)
                    return (0);

                f = global.active;

                if (!global.imeactive && global.imehwnd != NULL) {
                    global.imeactive = TRUE;
                    if (f != NULL) f->markpos = TRUE;
                    SetFocus(global.imehwnd);
                } else {
                    global.imeactive = FALSE;
                    if (f != NULL) f->markpos = FALSE;
                    SetFocus(global.hwnd);
                }

                if (f != NULL) {
                    ReformatParagraph(f, f->current, PARAOF(f->current)->next,
                                      OP_UPDATE | OP_MINIMAL);
                }
                //InvalidateRect(global.imehwnd, NULL, FALSE);
                //UpdateWindow(global.imehwnd);
                return(0);
            }

            case IDM_TOGGLEGLOSSARY:
                global.dynamicglossary = !global.dynamicglossary;
                return (0);

            case IDM_FILENEW: {
                extern char far *TplChosen;

                i = DialogBox (hInstance, "FileNew", hwnd, TemplateProc);

                switch (i) {
                    case -2: return (0);    /* Cancel */
                    case -1: break;         /* Use None */
                    default: if (TplChosen != NULL && TplChosen[0]) {
                                CreateUsingTemplate(TplChosen, ++Untitled);
                                return (0);
                             }
                             break;
                }

                global.active = NewFile (FN_NORMAL, TRUE);
                if (global.active != NULL) {
                    global.active->filename[0] = '\0';
                    sprintf(buffer, "Untitled #%d", ++Untitled);
                    ShowFileWindow(buffer, global.active);
                } else {
                    ErrorMessage(global.hwnd, "Cannot open any more file windows!");
                }
                if (global.imeactive) global.active->markpos = TRUE;
                return (0);
            }

            case IDM_FILEOPEN:
                DoFileOpen(NULL);
                return (0);

            case IDM_FILECLOSE:
                if (global.active != NULL) {
                    if (!SendMessage(global.active->parent, WM_QUERYENDSESSION, 0, 0L))
                        return (0);

                    global.active->changed = FALSE;

                    SendMessage(global.clienthwnd, WM_MDIDESTROY, (WPARAM) global.active->parent, 0L);
                }
                return (0);

            case IDM_FILESAVE:
                DoFileSaveAs(FALSE);
                return (0);

            case IDM_FILESAVEAS:
                DoFileSaveAs(TRUE);
                return (0);

			case IDM_FILEPRINT:
                if (global.active != NULL) {
                    if (global.hdevnames == NULL || global.hdevmode == NULL) {
                        MessageBeep(MB_ICONHAND);
                    } else {
                        PrintFile (global.active);
                    }
				}
				return (0);

			case IDM_FILEPRINTER:
                DoPrinterSetup();
                return (0);

            case IDM_FILESUMMARY:
                if (global.active != NULL) {
                    DialogBox (hInstance, "Summary", hwnd, SummaryProc);
                }
                return (0);

            case IDM_FILEHEADERS:
                if (global.active != NULL) {
                    DialogBox (hInstance, "Header", hwnd, HeaderProc);
                }
                return (0);

            case IDM_FILEFIND:
                DialogBox (hInstance, "FileFind", hwnd, FileFindProc);
                return (0);

            case IDM_FILEEXIT:
                if (SendMessage(hwnd, WM_COMMAND, IDM_WINDOWCLOSE, 0L)) {
                    SendMessage(hwnd, WM_DESTROY, 0, 0L);
                }
                return (0);

            case IDM_FILEABOUT:
                DialogBox (hInstance, "About", hwnd, AboutProc);
                return (0);

            case IDM_VIEWDRAFT: {
                FILEOPTIONS *f;
                HCURSOR hcursor;

                if (global.draftview) {
                    CheckMenuItem(hmenu, IDM_VIEWDRAFT, MF_UNCHECKED | MF_BYCOMMAND);
                    global.draftview = FALSE;
                } else {
                    CheckMenuItem(hmenu, IDM_VIEWDRAFT, MF_CHECKED | MF_BYCOMMAND);
                    global.draftview = TRUE;
                }

                /* Try to redraw the entire window */

                hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
                ShowCursor(TRUE);

                for (f = fileoptions; f != NULL; f = f->next) {
                    ReformatFile(f);
                    RedisplayFile(f);
                }

                ShowCursor(FALSE);
                SetCursor(hcursor);

                return (0);
            }

            case IDM_VIEWIME: {
                RECT rect;

                GetClientRect(global.hwnd, &rect);

                if (global.imehwnd != NULL) {
                    CheckMenuItem(hmenu, IDM_VIEWIME, MF_UNCHECKED | MF_BYCOMMAND);
                    DestroyWindow(global.imehwnd);
                    global.imehwnd = NULL;
                    global.imeactive = FALSE;

                    if (global.active != NULL && (global.active->type & FN_NORMAL)) {
                        global.active->markpos = FALSE;
                        ReformatParagraph(global.active, global.active->current,
                                          PARAOF(global.active->current)->next,
                                          OP_UPDATE | OP_MINIMAL);
                    }
                    SetFocus(global.hwnd);
                } else {
                    global.imehwnd = CreateWindow(PROGNAME " Input Line Editor Window", "Input Line Editor", WS_CHILD | WS_VISIBLE,
                                                    0, 0, 0, 0, global.hwnd, (HMENU) 15, hInstance, NULL);
                    if (global.imehwnd == NULL) {
                        ErrorMessage(global.hwnd, "Cannot create the Input Line Editor!");
                        return (0);
                    }

                    CheckMenuItem(hmenu, IDM_VIEWIME, MF_CHECKED | MF_BYCOMMAND);
                    global.imeactive = TRUE;
                    ShowWindow(global.imehwnd, SW_SHOW);
                    SetFocus(global.imehwnd);

                    if (global.active != NULL && (global.active->type & FN_NORMAL)) {
                        global.active->markpos = TRUE;
                        ReformatParagraph(global.active, global.active->current,
                                          PARAOF(global.active->current)->next,
                                          OP_UPDATE | OP_MINIMAL);
                    }
                }

                AdjustAllChildWindows (rect.right, rect.bottom);
                return (0);
            }

            case IDM_VIEWRIBBON: {
                RECT rect;

                GetClientRect(global.hwnd, &rect);

                if (global.tbhwnd != NULL) {
                    CheckMenuItem(hmenu, IDM_VIEWRIBBON, MF_UNCHECKED | MF_BYCOMMAND);
                    DestroyWindow(global.fonthwnd);
                    DestroyWindow(global.tbhwnd);
                    global.tbhwnd = global.fonthwnd = NULL;
                } else {
#ifdef USEWIN32
                    DWORD WinVersion;

                    WinVersion = GetVersion();
                    if (LOBYTE(LOWORD(WinVersion)) > 3 ||   /* Windows 95 */
                        !(HIWORD(WinVersion) & 0x8000)) {   /* Windows NT */

                        global.tbhwnd = CreateToolbar (global.hwnd, 0, 0, 0, 0,
                                                        1, ID_TOOLBAR, sizeof(toolbar) / sizeof(TOOLBARICON),
                                                        hInstance, toolbar, "StopSign");
                    } else {    /* Windows 3.x */
#endif USEWIN32
                        global.tbhwnd = CreateToolbar (global.hwnd, 0, 0, 0, 0,
                                                        1, ID_TOOLBAR, sizeof(toolbar) / sizeof(TOOLBARICON),
                                                        hInstance, toolbar, "BWStopSign");
#ifdef USEWIN32
                    }
#endif USEWIN32

                    if (global.tbhwnd == NULL) {
                        ErrorMessage(global.hwnd, "Cannot create the Iconbar!");
                        return (0);
                    }

                    CheckMenuItem(hmenu, IDM_VIEWRIBBON, MF_CHECKED | MF_BYCOMMAND);

                    /* Create the font list in the toolbar */

                    global.fonthwnd = CreateWindow("combobox", "Font List", WS_CHILD | WS_VISIBLE | CBS_DROPDOWNLIST,
                                                    2 * TBBORDER, TBBORDER+2, 160 - 2 * TBBORDER, 200, global.tbhwnd, (HMENU) ID_FONTBOX, hInstance, NULL);
                    EnableWindow(global.fonthwnd, FALSE);
                }

                AdjustAllChildWindows (rect.right, rect.bottom);
                return (0);
            }

            case IDM_VIEWSTATUS: {
                RECT rect;

                GetClientRect(global.hwnd, &rect);

                if (global.statushwnd != NULL) {
                    CheckMenuItem(hmenu, IDM_VIEWSTATUS, MF_UNCHECKED | MF_BYCOMMAND);
                    DestroyWindow(global.statushwnd);
                    global.statushwnd = NULL;
                } else {
                    CheckMenuItem(hmenu, IDM_VIEWSTATUS, MF_CHECKED | MF_BYCOMMAND);
                    global.statushwnd = CreateWindow (PROGNAME " Status", "Status Bar",
                                                      WS_CHILD | WS_VISIBLE | WS_BORDER,
                                                      0, 0, 0, 0, global.hwnd, (HMENU) 2, hInstance, NULL);
                }

                AdjustAllChildWindows (rect.right, rect.bottom);
                return (0);
            }

            case IDM_VIEWRULER: {
                BOOL On;
                FILEOPTIONS *f;

                On = !global.showruler;
                global.showruler = On;

                if (On) {
                    CheckMenuItem(hmenu, IDM_VIEWRULER, MF_CHECKED | MF_BYCOMMAND);
                } else {
                    CheckMenuItem(hmenu, IDM_VIEWRULER, MF_UNCHECKED | MF_BYCOMMAND);
                }

                for (f = fileoptions; f != NULL; f = f->next) {
                    DisplayRuler(f, On);
                    //InvalidateRect(f->parent, NULL, TRUE);
                }
                return (0);
            }

            case IDM_VIEWGLOSSARY:
                if (global.glisthwnd != NULL) {
                    SendMessage(global.tbhwnd, BM_SETSTATE, 0, ID_GLOSSARY);
                    CheckMenuItem(hmenu, IDM_VIEWGLOSSARY, MF_UNCHECKED | MF_BYCOMMAND);
                    DestroyWindow(global.glisthwnd);
                    global.glisthwnd = NULL;
                } else {
                    SendMessage(global.tbhwnd, BM_SETSTATE, 1, ID_GLOSSARY);
                    CheckMenuItem(hmenu, IDM_VIEWGLOSSARY, MF_CHECKED | MF_BYCOMMAND);
                    global.glisthwnd = CreateDialog(hInstance, "GlossaryList", hwnd, global.gloslistdlg);
                    SetActiveWindow(global.hwnd);
                }
                return (0);

            case IDM_VIEWSPECIAL: {
                FILEOPTIONS *f;
                HCURSOR hcursor;

                if (global.showspecial) {
                    SendMessage(global.tbhwnd, BM_SETSTATE, 0, ID_SPECIAL);
                    CheckMenuItem(hmenu, IDM_VIEWSPECIAL, MF_UNCHECKED | MF_BYCOMMAND);
                    global.showspecial = FALSE;
                } else {
                    SendMessage(global.tbhwnd, BM_SETSTATE, 1, ID_SPECIAL);
                    CheckMenuItem(hmenu, IDM_VIEWSPECIAL, MF_CHECKED | MF_BYCOMMAND);
                    global.showspecial = TRUE;
                }

                /* Try to redraw the entire window */

                hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
                ShowCursor(TRUE);
                for (f = fileoptions; f != NULL; f = f->next) RedisplayFile(f);
                ShowCursor(FALSE);
                SetCursor(hcursor);
                return (0);
            }

            case IDM_EDITSEARCH:
                if (global.active == NULL) return (0);
                i = DialogBox (hInstance, "Search", hwnd, SearchProc);
                if (i) {
                    DoSearch();
                    LastSearchIsReplace = FALSE;
                }
                return (0);

            case IDM_CONTINUESEARCH:
                if (global.active == NULL) return (0);
                if (LastSearchIsReplace) DoReplace(); else DoSearch();
                return (0);

            case IDM_EDITREPLACE:
                if (global.active == NULL) return (0);
                i = DialogBox (hInstance, "Replace", hwnd, ReplaceProc);
                if (i) {
                    DoReplace();
                    LastSearchIsReplace = TRUE;
                }
                return (0);

            case IDM_WINDOWTILE:
                SendMessage(global.clienthwnd, WM_MDITILE, 0, 0L);
                return (0);

            case IDM_WINDOWCASCADE:
                SendMessage(global.clienthwnd, WM_MDICASCADE, 0, 0L);
                return (0);

            case IDM_WINDOWARRANGE:
                SendMessage(global.clienthwnd, WM_MDIICONARRANGE, 0, 0L);
                return (0);

            case IDM_WINDOWCLOSE: {
                BOOL DoIt;

                /* Any file opened? */
                if (fileoptions == NULL) return (TRUE);

                DoIt = EnumChildWindows(global.clienthwnd, CloseAllEnumProc, 0L);

                if (DoIt) SwitchMenu (1);     /* Minimal menu */
                return (DoIt);
            }

            case IDM_UTILITIESSTAT:
                DialogBox (hInstance, "Statistics", hwnd, StatisticsProc);
                return (0);

            case IDM_KANJIINPUT:
                if (global.active != NULL) curfile = global.active;
                else curfile = NULL;

                DialogBox (hInstance, "KanjiInput", global.hwnd, JISInputProc);
                return (0);

            case IDM_KANJILOOKUP:
                if (global.active != NULL) curfile = global.active;
                else curfile = NULL;

                DialogBox (hInstance, "KanjiLookup", global.hwnd, LookupProc);
                return (0);

            case IDM_RLOOKUP:   // DCY
                if (!IsWindow(hKanjiLookupDlg)) {
                         hKanjiLookupDlg = CreateWindow(
                             PROGNAME " Radical Lookup", "Radical Lookup",
                             WS_DLGFRAME | DS_MODALFRAME | WS_CAPTION | WS_VISIBLE | WS_SYSMENU | WS_MINIMIZEBOX,
                             30, 30,
                             CW_USEDEFAULT, CW_USEDEFAULT,
                             NULL, NULL, hInstance, NULL);
                } else {
                    SetActiveWindow(hKanjiLookupDlg);
                }
                return (0);

            case IDM_KANJITABLE:
                if (global.active != NULL) curfile = global.active;
                else curfile = NULL;

                DialogBox (hInstance, "JISTable", global.hwnd, JISTableDlgProc);
                return (0);

            case IDM_UTILITIESOPTIONS:
                DialogBox (hInstance, "Options", hwnd, OptionsProc);

                /* 3D dialog boxes */

#ifdef CTL3D
                if (Dialogs3D != global.ctl3d) {
                    if (Dialogs3D) {
                        Ctl3dUnregister(hInstance);
                        Dialogs3D = FALSE;
                    } else {
                        Dialogs3D = Ctl3dRegister(hInstance);
                        if (Dialogs3D) {
                            Ctl3dAutoSubclass(hInstance);
                        } else {
                            ErrorMessage(global.hwnd, PROGNAME " " JWPVERSION " cannot create 3D dialog boxes!");
                            global.ctl3d = FALSE;
                        }
                    }
                }
#endif CTL3D

                return (0);

            case IDM_UTILITIESDICT: {
                INPUTMODE oldmode;

                oldmode = global.mode;
                DialogBox (hInstance, "Dictionary", hwnd, DictionaryProc);

                if (oldmode != global.mode) ToggleInputMode();

                return (0);
            }

            case IDM_UTILITIESHOTKEYS:
                DialogBox (hInstance, "HotKeys", hwnd, HotKeysProc);
                return (0);

            case IDM_UTILITIESGLOSSARY: {
                INPUTMODE oldmode;
                BOOL dynamic;

                oldmode = global.mode;
                dynamic = global.dynamicglossary;
                global.dynamicglossary = FALSE;

                DialogBox (hInstance, "Glossary", hwnd, GlossaryProc);

                if (oldmode != global.mode) ToggleInputMode();
                global.dynamicglossary = dynamic;

                return (0);
            }

            case IDM_UTILITIESCONVERT:
                DialogBox (hInstance, "UserDict", hwnd, UserConversionProc);
                return (0);

            case IDM_UTILITIESFONTS: {
                FILEOPTIONS *f;

                i = DialogBox (hInstance, "KanjiFonts", hwnd, InstalledFontsProc);

                if (i) {
                    for (f = fileoptions; f != NULL; f = f->next) {
                        ChangePrinterOrFonts(f, f == fileoptions, TRUE);
                    }
                }
                return (0);
            }

            case IDM_FILERUN:
                DialogBox (hInstance, "RunProg", hwnd, RunProc);
                return (0);

            case IDM_FILE1:
            case IDM_FILE2:
            case IDM_FILE3:
            case IDM_FILE4: {
                char *cp;
                extern char *GetQuickFile(int);

                cp = GetQuickFile(WINCMDID - IDM_FILE1);
                if (cp != NULL) {
                    if (!DoFileOpen(cp)) {
                        ErrorMessage(global.hwnd, "Cannot open file '%s'.", cp);
                    }
                }
                return (0);
            }

            case IDM_FORMATPAGE:
                DialogBox (hInstance, "PageSetup", hwnd, PageSetupProc);
                return (0);

            case IDM_FORMATREPAGE:
                if (global.active != NULL) {
                    HCURSOR hcursor;
                    char buffer[MAXLINELEN];

                    hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
                    ShowCursor(TRUE);

                    StatusMessage("Repagenating file...");
                    RepagenateFile(global.active);
                    sprintf(buffer, "Total %d page%s.", global.active->nr_pages,
                                global.active->nr_pages > 1 ? "s" : "");
                    StatusMessage(buffer);
                    RedisplayFile(global.active);

                    ShowCursor(FALSE);
                    SetCursor(hcursor);
                }
                return (0);

            case IDM_FORMATPARAGRAPH: {
                POSITION p;

                if (global.active == NULL) return (0);
                i = DialogBox (hInstance, "FormatParagraph", hwnd, FormatParagraphProc);

                if (i) {
                    if (SELPARA1(global.active) != NULL) {
                        PARAOF(p) = SELPARA1(global.active);
                        LINEOF(p) = PARAOF(p)->lines;
                        POSOF(p) = 0;

                        ReformatParagraph(global.active, p, SELPARA2(global.active)->next, OP_UPDATE | OP_MOVETOEND);
                    } else {
                        PARAOF(p) = CURPARA(global.active);
                        LINEOF(p) = PARAOF(p)->lines;
                        POSOF(p) = 0;

                        ReformatParagraph(global.active, p, PARAOF(p)->next, OP_UPDATE | OP_MOVETOEND);
                    }
                    SetReformatProc(NULL);
                }
                return (0);
            }

            case IDM_FORMATVERTICAL: {
                BOOL i;

                if (global.active == NULL || !(global.active->type & FN_NORMAL)) {
                    i = global.vertical = !global.vertical;
                } else {
                    i = global.active->vertical = !global.active->vertical;
                    global.active->changed = TRUE;
                }
                SendMessage(global.tbhwnd, BM_SETSTATE, i ? 0 : 1, ID_LEFTRIGHT);
                SendMessage(global.tbhwnd, BM_SETSTATE, i ? 1 : 0, ID_UPDOWN);
                return (0);
            }

            case IDM_HELPINDEX:
                WinHelp(global.hwnd, global.help, HELP_INDEX, NULL);
                return (0);
            case IDM_HELPONHELP:
                WinHelp(global.hwnd, global.help, HELP_HELPONHELP, NULL);
                return (0);
            case IDM_HELPSEARCH:
                WinHelp(global.hwnd, global.help, HELP_KEY, NULL);
                return (0);
            case IDM_HELPINTRODUCTION:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_Introduction);
                return (0);
            case IDM_HELPENTERTEXT:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_EnterText);
                return (0);
            case IDM_HELPMOVING:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_MovingAround);
                return (0);
            case IDM_HELPKANJIS:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_ManagingKanjis);
                return (0);
            case IDM_HELPSPECIAL:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_SpecialFeatures);
                return (0);
            case IDM_HELPFILES:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_WorkingWithFiles);
                return (0);
            case IDM_HELPPRINTING:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_Printing);
                return (0);
            case IDM_HELPTECHSUPPORT:
                WinHelp(global.hwnd, global.help, HELP_CONTEXT, Topic_TechnicalSupport);
                return (0);
        }
        break;

#ifdef CTL3D
    case WM_SYSCOLORCHANGE:
        Ctl3dColorChange();
        break;
#endif CTL3D

    case WM_SYSCOMMAND:
        switch (wParam & 0xfff0) {
            case SC_CLOSE:
                if (!SendMessage(hwnd, WM_COMMAND, IDM_WINDOWCLOSE, 0L)) return (0);
                break;
        }
        break;

    case WM_DROPFILES: {
        int howmany, len;
        char buffer[MAXLINELEN];

#ifdef USEWIN32
        howmany = DragQueryFile((HDROP) wParam, 0xffffffff, buffer, MAXLINELEN - 1);
#else
        howmany = DragQueryFile((HDROP) wParam, -1, buffer, MAXLINELEN - 1);
#endif USEWIN32
        if (howmany <= 0) {
            DragFinish((HDROP) wParam);
            return (0);
        }

        for (i = 0; i < howmany; i++) {
            len = DragQueryFile((HDROP) wParam, i, buffer, MAXLINELEN - 1);
            buffer[len] = '\0';

            if (!buffer[0] || !DoFileOpen(buffer))
                ErrorMessage(global.hwnd, "Cannot open file '%s'.", buffer);
        }
        DragFinish((HDROP) wParam);
        return (0);
    }

    case WM_DESTROY:
        PlayExternalSoundFile("END");
        if (IsWindow(hKanjiLookupDlg)) DestroyWindow(hKanjiLookupDlg);
        if (!SendMessage(hwnd, WM_COMMAND, IDM_WINDOWCLOSE, 0L)) return (0);
        InitDDE(TRUE);
        if(global.imehwnd) DestroyWindow(global.imehwnd);
        if(global.hdevmode != NULL) GlobalFree(global.hdevmode);
        if(global.hdevnames != NULL) GlobalFree(global.hdevnames);

        FreeProcInstance (global.gloslistdlg);

        WriteOptionsFile(FALSE);
        WriteGlossary(global.glossary);
        WriteConversionCache(global.convcache);
        WriteUserConversions(global.userdict);

        /* Free some ad hoc stuff */

        DisplayKanjiBitmap(NULL, 0, 0, 0, 0, -1, 0, NULL, FALSE);
        DeleteObject(DefAsciiFont.hfont);       /* ASCII font */
        DeleteObject(FullMenu);                 /* The menus */
        DeleteObject(MinimalMenu);

        FreeAllMemory();

        WinHelp(global.hwnd, global.help, HELP_QUIT, NULL);
        DragAcceptFiles(global.clienthwnd, FALSE);
        DragAcceptFiles(global.hwnd, FALSE);
        PostQuitMessage(0);
        return (0);

    case WM_RENDERFORMAT:
    case WM_DESTROYCLIPBOARD:
    case WM_RENDERALLFORMATS:
    case WM_ASKCBFORMATNAME:
    case WM_SIZECLIPBOARD:
    case WM_PAINTCLIPBOARD:
    case WM_HSCROLLCLIPBOARD:
    case WM_VSCROLLCLIPBOARD:
        return (ProcessClipboardMessage(hwnd, message, wParam, lParam));

   }

    return (DefFrameProc(hwnd, global.clienthwnd, message, wParam, lParam));
}
