/* Copyright (C) Stephen Chung, 1995.  All rights reserved. */

#include "jwp.h"

#include "idm.h"
#include <shellapi.h>
#include <stdarg.h>
#include <sys/stat.h>

static INT32 Percentage, OverheadPercentage;
static int HeaderNum;
static BOOL LeftRight;

static char *FileSelectionSpec = NULL;
static KANJI far *criteria[NRSUMMARIES] = { NULL, NULL, NULL, NULL, NULL };
static char *FileInfoName;

extern BOOL Dialogs3D;

#ifdef USEWIN32
  #define FILENAMELEN       MAXFILENAMELEN  /* To cater for long filenames */
#else
  #define FILENAMELEN       15              /* More than enough for 8.3 */
#endif USEWIN32

#define PARENTDIR           "<Go Up>"
#define DRIVECHAR           '['
#define NRQUICKFILE         4

#define NRPROGS             5

static UINT TimerHandle;
static HWND AnimateIconHwnd;
static char *AnimateIconName;
static int AnimateIconMax;
static int AnimateIconFrame;

static struct {
    char *fullname;
    char *filename;
} QuickFiles[NRQUICKFILE];

#define TPLBLOCKSIZE    5

typedef struct {
    char far *filename;
    KANJI far *desc;
} TPLFILE;

char far *TplChosen = NULL;
static TPLFILE far *Templates = NULL;
static int nr_templates = 0;



int bytecmp (BYTE far *p1, BYTE far *p2)
{
    for (; *p1 == *p2; p1++, p2++) {
        if (*p1 == 0) return (0);
    }

    return (*p1 - *p2);
}


int bytencmp (BYTE far *p1, BYTE far *p2, int n)
{
    int i;

    for (i = 0; i < n; i++) {
        if (p1[i] != p2[i]) return (p1[i] - p2[i]);
        if (p1[i] == 0) return (0);
    }

    return (0);
}


KANJI far *kanjicpy (KANJI far *p1, KANJI far *p2)
{
    while (*p1++ = *p2++);

    return (p1);
}



int kanjicmp (KANJI far *p1, KANJI far *p2)
{
    for (; (*p1 == *p2) && *p1; p1++, p2++);

    return (*p1 - *p2);
}


int kanjincmp (KANJI far *p1, KANJI far *p2, int n)
{
    int i;

    for (i = 0; i < n; i++) {
        if (p1[i] != p2[i]) return (p1[i] - p2[i]);
        if (p1[i] == 0) return (0);
    }

    return (0);
}


int kanjilen (KANJI far *s)
{
    int i;

    for (i = 0; *s; s++, i++);
    return (i);
}


int unitlen (KUNIT far *s)
{
    int i;

    for (i = 0; s->kanji; s++, i++);
    return (i);
}


KANJI far *kanjicat (KANJI far *d, KANJI far *s)
{
    kanjicpy(d + kanjilen(d), s);
    return (d);
}


int ReadShortInt(HFILE fd, int *n)
{
    INT16 sht;
    int r;

    r = _lread(fd, &sht, sizeof(INT16));
    *n = sht;
    return (r);
}



int ReadChar(HFILE fd)
{
    unsigned char buf[4];

    if (_lread(fd, buf, sizeof(char)) != sizeof(char)) return (EOF);
    else return (buf[0]);
}



int WriteShortInt(HFILE fd, int n)
{
    INT16 sht = n;

    return (_lwrite(fd, &sht, sizeof(INT16)));
}



int WriteChar(HFILE fd, char n)
{
    return (_lwrite(fd, (char far *) &n, sizeof(char)));
}



#ifndef USEWIN32

int GetCurrentDirectory (int size, char *buf) { return (strlen(getcwd(buf, size))); }
BOOL SetCurrentDirectory (char *path) { return (chdir(path) < 0 ? FALSE : TRUE); }
BOOL MakeDirectory (char *path) { return (mkdir(path) < 0 ? FALSE : TRUE); }
BOOL SetFileAttributes (char *path, int mode) { return (chmod(path, mode) == 0); }

#else

BOOL MakeDirectory (char *path)
{
    SECURITY_ATTRIBUTES sa;

    sa.nLength = sizeof(sa);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = TRUE;
    return (CreateDirectory(path, &sa));
}

static char *DayOfWeek[] = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };
static char *Month[] = { "Error", "January", "February", "March", "April", "May", "June",
                         "July", "August", "September", "October", "November", "December" };

static char *TimeString (FILETIME *tp)
{
    SYSTEMTIME stime;
    char timediv[10], amstr[10], pmstr[10];
    static char buffer[MAXLINELEN];

    buffer[0] = '\0';
    if (tp->dwLowDateTime == 0 && tp->dwHighDateTime == 0) return (buffer);

    if (!FileTimeToSystemTime(tp, &stime)) return (buffer);

    GetProfileString("intl", "sTime", ":", timediv, 9);
    GetProfileString("intl", "s1159", "AM", amstr, 9);
    GetProfileString("intl", "s2359", "PM", pmstr, 9);

    sprintf(buffer, "%s, %s %d, %d  %d%s%02d %s\n",
                DayOfWeek[stime.wDayOfWeek], Month[stime.wMonth],
                stime.wDay, stime.wYear,
                stime.wHour == 0 ? 12 : stime.wHour > 12 ? stime.wHour - 12 : stime.wHour,
                timediv, stime.wMinute,
                stime.wHour >= 12 ? pmstr : amstr);

    return (buffer);
}

#endif USEWIN32



void ClearAllProperties (HWND hwnd)
{
    RemoveProp(hwnd, PROP_FILEOPTIONS);
    RemoveProp(hwnd, PROP_HWND);
    RemoveProp(hwnd, PROP_CURSOR);
    RemoveProp(hwnd, PROP_SIZEX);
    RemoveProp(hwnd, PROP_SIZEY);
    RemoveProp(hwnd, PROP_POINTER);
    RemoveProp(hwnd, PROP_NUMBER);
    RemoveProp(hwnd, PROP_HELPWINDOW);
    RemoveProp(hwnd, PROP_HELPCONTEXT);

#ifdef USEWIN32
    RemoveProp(hwnd, PROP_HELPORIGADDR);
#else
    RemoveProp(hwnd, PROP_HELPORIGADDRLOW);
    RemoveProp(hwnd, PROP_HELPORIGADDRHIGH);
#endif USEWIN32
}



/* Window: -1 = popup, 0 = Main, 1 = Note, 2 = Dialog, 3 = Dialog2 */

LONG CALLBACK TrapHelpProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    FARPROC addp;

#ifdef USEWIN32
    addp = (FARPROC) GetProp(hwnd, PROP_HELPORIGADDR);
#else
    WORD low, high;

    low = GetProp(hwnd, PROP_HELPORIGADDRLOW);
    high = GetProp(hwnd, PROP_HELPORIGADDRHIGH);
    addp = (FARPROC) MAKELONG(low, high);
#endif USEWIN32

    switch (message) {
        case WM_CHAR:
        case WM_KEYDOWN: {
            int i, j, window;
            char *context;
            BOOL shift, ctrl;
            HCURSOR hcursor;
            char buffer[MAXLINELEN];

            shift = (GetKeyState(VK_SHIFT) < 0);
            ctrl = (GetKeyState(VK_CONTROL) < 0);

            if (message == WM_CHAR) {
                if (wParam < ' ') {
                    j = wParam + 0x40;
                } else if (ctrl && 'a' <= wParam && wParam <= 'z') {
                    j = wParam - 0x20;
                } else {
                    j = wParam;
                }
                i = DecodeKeyStroke(WM_CHAR, j, shift, TRUE);
            } else {
                if (wParam < VK_F1 || wParam > VK_F12) break;
                i = DecodeKeyStroke(WM_KEYDOWN, wParam, shift, ctrl);
            }

            if (i != IDM_HELPINDEX) break;

            /* Context-sensitive help! */

            window = (int) GetProp(hwnd, PROP_HELPWINDOW);
            context = (char *) GetProp(hwnd, PROP_HELPCONTEXT);

            if (window < 0) {
                strcpy(buffer, "PI(\"");
            } else {
                strcpy(buffer, "JI(\"");
            }

            j = strlen(buffer);
            for (i = 0; global.help[i]; i++) {
                if (global.help[i] == '\\') buffer[j++] = '\\';
                buffer[j++] = global.help[i];
            }
            buffer[j] = '\0';

            strcat(buffer, "\",\"");
            strcat(buffer, context);

/*
            switch (p->window) {
                case -1: break;
                case 0:  break;
                case 1:  strcat(buffer, ">Note"); break;
                case 2:  strcat(buffer, ">Dialog"); break;
                case 3:  strcat(buffer, ">Dialog2"); break;
            }
*/

            strcat(buffer, "\")");

            hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
            ShowCursor(TRUE);

            WinHelp(hwnd, global.help, HELP_COMMAND, (DWORD) buffer);

            ShowCursor(FALSE);
            SetCursor(hcursor);

            return (TRUE);
        }
    }

    if (addp == NULL) {
        ErrorMessage(global.hwnd, "WARNING (TrapHelpProc): Cannot get help-subclass original address!");
        return (TRUE);
    }

    return (CallWindowProc(addp, hwnd, message, wParam, lParam));
}




static BOOL TrapHelpOnce (HWND hwnd, char *Context, int window)
{
    LONG wndproc;

    wndproc = GetWindowLong(hwnd, GWL_WNDPROC);
    if (wndproc == NULL) return (FALSE);

#ifdef USEWIN32
    SetProp(hwnd, PROP_HELPORIGADDR, (HANDLE) wndproc);
#else
    SetProp(hwnd, PROP_HELPORIGADDRLOW, LOWORD(wndproc));
    SetProp(hwnd, PROP_HELPORIGADDRHIGH, HIWORD(wndproc));
#endif USEWIN32
    SetProp(hwnd, PROP_HELPWINDOW, (HANDLE) window);
    SetProp(hwnd, PROP_HELPCONTEXT, (HANDLE) Context);

    SetWindowLong(hwnd, GWL_WNDPROC, (LONG) TrapHelpProc);
    return (TRUE);
}



BOOL CALLBACK TrapHelpEnumProc (HWND hwnd, LPARAM lParam)
{
    char buffer[MAXLINELEN];

    //if (GetParent(hwnd) != HIWORD(lParam)) return (TRUE);

    buffer[0] = '\0';
    GetClassName(hwnd, buffer, MAXLINELEN);
    if (!stricmp(buffer, "static")) return (TRUE);

#ifdef USEWIN32
    if (GetWindowLong(hwnd, GWL_ID) < 0) return (TRUE);
#else
    if (GetWindowWord(hwnd, GWW_ID) < 0) return (TRUE);
#endif

    TrapHelpOnce (hwnd, (char *) lParam, 1);
    return (TRUE);
}



BOOL TrapHelpInDialogBoxes (HWND hwnd, char *Context, int window, BOOL children)
{
    if (!TrapHelpOnce(hwnd, Context, window)) return (FALSE);
    if (!children) return (TRUE);

    EnumChildWindows(hwnd, TrapHelpEnumProc, (LPARAM) Context);
    return (TRUE);
}



BOOL CALLBACK UnhookHelpTrapEnumProc (HWND hwnd, LPARAM lParam)
{
    WORD low, high;
    FARPROC addp;
    char buffer[MAXLINELEN];

    //if (GetParent(hwnd) != HIWORD(lParam)) return (TRUE);

    buffer[0] = '\0';
    GetClassName(hwnd, buffer, MAXLINELEN);
    if (!stricmp(buffer, "static")) return (TRUE);

#ifdef USEWIN32
    if (GetWindowLong(hwnd, GWL_ID) < 0) return (TRUE);
    addp = (FARPROC) GetProp(hwnd, PROP_HELPORIGADDR);
#else
    if (GetWindowWord(hwnd, GWW_ID) < 0) return (TRUE);
    low = GetProp(hwnd, PROP_HELPORIGADDRLOW);
    high = GetProp(hwnd, PROP_HELPORIGADDRHIGH);
    addp = (FARPROC) MAKELONG(low, high);
#endif USEWIN32

    if (addp == NULL) {
        ErrorMessage(global.hwnd, "WARNING (UnhookHelpTrap): Cannot get help-subclass original address!");
        return (NULL);
    }

    SetWindowLong(hwnd, GWL_WNDPROC, (LONG) addp);

#ifdef USEWIN32
    RemoveProp(hwnd, PROP_HELPORIGADDR);
#else
    RemoveProp(hwnd, PROP_HELPORIGADDRLOW);
    RemoveProp(hwnd, PROP_HELPORIGADDRHIGH);
#endif USEWIN32
    RemoveProp(hwnd, PROP_HELPWINDOW);
    RemoveProp(hwnd, PROP_HELPCONTEXT);

    WinHelp(hwnd, global.help, HELP_QUIT, 0L);
    return (TRUE);
}



void UnhookHelpTraps (HWND hwnd, BOOL children)
{
    UnhookHelpTrapEnumProc(hwnd, NULL);
    if (!children) return;

    EnumChildWindows(hwnd, UnhookHelpTrapEnumProc, NULL);
}



void CALLBACK AnimateIconProc (HWND hwnd, UINT message, UINT id, DWORD time)
{
    HDC hdc;
    HICON hicon;
    char buffer[MAXLINELEN];

    if (AnimateIconFrame >= AnimateIconMax) AnimateIconFrame = 0;
    sprintf(buffer, "%s%d", AnimateIconName, ++AnimateIconFrame);

    hicon = LoadIcon(hInstance, buffer);
    if (hicon == NULL) return;

    hdc = GetDC(AnimateIconHwnd);
    if (hdc == NULL) return;

    InvalidateRect(AnimateIconHwnd, NULL, TRUE);
    UpdateWindow(AnimateIconHwnd);

    DrawIcon(hdc, 0, 0, hicon);

    ReleaseDC(AnimateIconHwnd, hdc);
}




void AnimateIcon (HWND hwnd, char *name, int max, double freq)
{
    if (name != NULL) {
        AnimateIconHwnd = hwnd;
        AnimateIconName = name;
        AnimateIconMax = max;
        AnimateIconFrame = -1;      /* First icon is 0 */

        TimerHandle = SetTimer(NULL, NULL, (int) (1000.0 / freq), (TIMERPROC) AnimateIconProc);
        if (TimerHandle == 0) return;
    } else {
        KillTimer(NULL, TimerHandle);
        AnimateIconFrame = AnimateIconMax;      /* Do the last icon */
        AnimateIconMax++;
    }
    AnimateIconProc (NULL, WM_TIMER, TimerHandle, 0L);
}



BOOL CALLBACK StatisticsProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {

        case WM_INITDIALOG: {
            INT32       allocated, requests, hits, usage, overhead;
            INT32       TotalBytes;
            int         num;
            PARAGRAPH   far *pp;
            char        buffer[BUFSIZE];

            TrapHelpInDialogBoxes(hwnd, "StatisticsDlg", 1, TRUE);

            FontCacheStatistics (&num, &usage, &requests, &hits);
            sprintf(buffer, "%d Font Cache%s:", num, (num > 1) ? "s" : "");
            SetDlgItemText(hwnd, 4201, buffer);
            if (requests > 0L) {
                sprintf(buffer, "%ld / %ld  (%d%%)  %ld cached",
                            hits, requests, (int) (100.0 * ((double) hits) / ((double) requests)), usage);
            } else {
                sprintf(buffer, "%ld cached", usage);
            }
            SetDlgItemText(hwnd, 4202, buffer);

            ConvCacheStatistics (&usage, &requests, &hits);
            if (requests > 0L) {
                sprintf(buffer, "%ld / %ld  (%d%%)  %ld cached",
                            hits, requests, (int) (100.0 * ((double) hits) / ((double) requests)), usage);
            } else {
                sprintf(buffer, "%ld cached", usage);
            }
            SetDlgItemText(hwnd, 4212, buffer);

            if (global.active != NULL) {
                for (pp = global.active->paragraph, TotalBytes = 0;
                     pp != NULL; pp = pp->next) {

                        TotalBytes += unitlen(pp->text);
                }

                global.active->nr_bytes = TotalBytes;

                sprintf(buffer, "%ld character%s", TotalBytes, (TotalBytes > 1L) ? "s" : "");
                SetDlgItemText(hwnd, 4221, buffer);

                if (global.active->undolevels > 0) {
                    sprintf(buffer, "%d change%s retained", global.active->undolevels,
                            (global.active->undolevels > 1) ? "s" : "");
                } else {
                    strcpy(buffer, "No changes retained");
                }
                SetDlgItemText(hwnd, 4222, buffer);
            } else {
                SetDlgItemText(hwnd, 4221, "No active file");
            }

            CountMemoryUsage(&allocated, &usage, &overhead);

#ifdef USEWIN32
            sprintf(buffer,"Memory Usage:\n%ldK Allocated", (allocated / 1024L) + 1);
#else
            sprintf(buffer,"Memory Usage Efficiency (%ldK Allocated):", (allocated / 1024L) + 1);
#endif USEWIN32

            SetDlgItemText(hwnd, 4231, buffer);

#ifndef USEWIN32
            if (allocated > 0L) {
                Percentage = (((usage + overhead) * 100L) / allocated) + 1L;
                if (Percentage > 100L) Percentage = 100L;

                OverheadPercentage = ((usage * 100L) / allocated) + 1L;
                if (OverheadPercentage > 100L) OverheadPercentage = 100L;
            } else {
                Percentage = 100L;
                OverheadPercentage = 100L;
            }

            sprintf(buffer,"%ld%%", Percentage);
            SetDlgItemText(hwnd, 4233, buffer);
#endif USEWIN32

            CenterDialogBox(hwnd);

            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_PAINT: {
            INT32    len;
            HDC      hdc;
            HWND     BarHwnd;
            RECT     rect;

            BarHwnd = GetDlgItem(hwnd, 4232);
            GetClientRect(BarHwnd, &rect);

            len = (rect.right - 4) * Percentage / 100L;

            hdc = GetDC(BarHwnd);
            SelectObject(hdc, GetStockObject(GRAY_BRUSH));
            Rectangle(hdc, 2, 2, len + 2, rect.bottom - 2);

            len = (rect.right - 4) * OverheadPercentage / 100L;
            if (Dialogs3D) {
                SelectObject(hdc, GetStockObject(BLACK_BRUSH));
            } else {
                SelectObject(hdc, GetStockObject(LTGRAY_BRUSH));
            }
            Rectangle(hdc, 2, 2, len + 2, rect.bottom - 2);
            ReleaseDC(BarHwnd, hdc);
            break;
        }

        case WM_KEYDOWN:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK:
                case IDCANCEL:  UnhookHelpTraps(hwnd, TRUE);
                                EndDialog(hwnd, 0);
                                return (TRUE);
            }
            break;
    }

    return (FALSE);
}



BOOL CALLBACK SummaryProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i, len;
            FILEOPTIONS *f, *f1;
            KANJI buffer[MAXLINELEN];

            /* Set the type and mode-change icon of the Jedit controls */

            f1 = global.active;

            for (i = 0; i < NRSUMMARIES; i++) {
                SendDlgItemMessage(hwnd, 4201 + i, EM_SETRECT, 0, (LONG) GetDlgItem(hwnd, 4211));
                f = (FILEOPTIONS *) SendDlgItemMessage(hwnd, 4201 + i, EM_GETHANDLE, 0, 0L);

                if (f1 != NULL && f1->summary[i] != NULL) {
                    kanjicpy(buffer, f1->summary[i]);
                    SendDlgItemMessage(hwnd, 4201 + i, EM_REPLACESEL, 0, (LONG) buffer);
                } else {
                    buffer[0] = 0;
                }
                if (i == 0) {
                    len = kanjilen(buffer);
                    SendDlgItemMessage(hwnd, 4211, EM_SETHANDLE, 0, (LONG) f->hwnd);  /* icon-change */
                }
            }

            if (len > 0) SendDlgItemMessage(hwnd, 4201, EM_SETSEL, len, MAKELONG(0, len - 1));
            SetFocus(GetDlgItem(hwnd, 4201));

            TrapHelpInDialogBoxes(hwnd, "SummaryDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

        case WM_PAINT: {
            int i;
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);

            for (i = 0; i < NRSUMMARIES; i++) DrawBoundingBox(hwnd, hdc, 4201 + i);

            EndPaint(hwnd, &ps);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            int i, len;
            KANJI far *kp, far *up;

            switch (WINCMDID) {
                case IDOK:
                    if (global.active != NULL) {
                        for (i = 0; i < NRSUMMARIES; i++) {
                            up = (KANJI far *) SendDlgItemMessage(hwnd, 4201 + i, EM_GETLINE, 0, 0L);
                            len = kanjilen (up);

                            if (len <= 0) {
                                if (global.active->summary[i] != NULL) {
                                    FreeBlock(global.active->summary[i]);
                                    global.active->summary[i] = NULL;
                                }
                            } else {
                                kp = global.active->summary[i];
                                if (kp == NULL) {
                                    kp = (KANJI far *) BlockAlloc((len + 5) * sizeof(KANJI));
                                } else if ((SegHeapGetSize(kp) / sizeof(KANJI)) < len + 5) {
                                    FreeBlock(kp);
                                    kp = (KANJI far *) BlockAlloc((len + 5) * sizeof(KANJI));
                                }
                                kanjicpy(kp, up);
                                global.active->summary[i] = kp;
                            }
                        }
                    }
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            break;
        }
    }
    return (FALSE);
}



BOOL CALLBACK EditHeaderProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i, len;
            FILEOPTIONS *f;
            KANJI far *kp;

            if (LeftRight) {
                switch (HeaderNum) {
                    case 0: SetWindowText(hwnd, "Edit Left Page Header"); break;
                    case 1: SetWindowText(hwnd, "Edit Right Page Header"); break;
                    case 2: SetWindowText(hwnd, "Edit Left Page Footer"); break;
                    case 3: SetWindowText(hwnd, "Edit Right Page Footer"); break;
                }
            } else {
                switch (HeaderNum) {
                    case 1:
                    case 0: SetWindowText(hwnd, "Edit Header"); break;
                    case 3:
                    case 2: SetWindowText(hwnd, "Edit Footer"); break;
                }
            }

            /* Set the type and mode-change icon */
            for (i = 0; i < NRHEADERS; i++) {
                SendDlgItemMessage(hwnd, 4201 + i, EM_SETRECT, 0, (LONG) GetDlgItem(hwnd, 4211));
                kp = global.active->header[HeaderNum][i];
                if (kp != NULL) {
                    SendDlgItemMessage(hwnd, 4201 + i, EM_REPLACESEL, 0, (LONG) kp);
                    len = kanjilen(kp);
                    if (i == 0 && len > 0) {
                        SendDlgItemMessage(hwnd, 4201, EM_SETSEL, len, MAKELONG(0, len - 1));
                    }
                }
            }

            f = (FILEOPTIONS *) SendDlgItemMessage(hwnd, 4201, EM_GETHANDLE, 0, 0L);
            SendDlgItemMessage(hwnd, 4211, EM_SETHANDLE, 0, (LONG) f->hwnd);  /* Mode-change icon */
            SetFocus(GetDlgItem(hwnd, 4201));

            TrapHelpInDialogBoxes(hwnd, "EditHeaderDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

        case WM_PAINT: {
            int i;
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);

            for (i = 0; i < NRHEADERS; i++) DrawBoundingBox(hwnd, hdc, 4201 + i);

            EndPaint(hwnd, &ps);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK: {
                    int i, len;
                    KANJI far *up, far *kp;

                    for (i = 0; i < NRHEADERS; i++) {
                        up = (KANJI far *) SendDlgItemMessage(hwnd, 4201 + i, EM_GETLINE, 0, 0L);
                        len = kanjilen(up);

                        if (len <= 0) {
                            if (global.active->header[HeaderNum][i] != NULL) {
                                FreeBlock(global.active->header[HeaderNum][i]);
                                global.active->header[HeaderNum][i] = NULL;
                            }
                        } else {
                            kp = global.active->header[HeaderNum][i];
                            if (kp == NULL) {
                                kp = (KANJI far *) BlockAlloc((len + 5) * sizeof(KANJI));
                            } else if ((SegHeapGetSize(kp) / sizeof(KANJI)) < len + 5) {
                                FreeBlock(kp);
                                kp = (KANJI far *) BlockAlloc((len + 5) * sizeof(KANJI));
                            }
                            kanjicpy(kp, up);
                            global.active->header[HeaderNum][i] = kp;
                        }
                    }

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
                }

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            break;
    }
    return (FALSE);
}



BOOL CALLBACK HeaderProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            if (global.active->lrheader) {
                LeftRight = TRUE;

                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Left Header");
                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Right Header");
                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Left Footer");
                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Right Footer");
            } else {
                LeftRight = FALSE;

                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Header");
                SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Footer");
            }

            SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, 0, 0L);

            CheckDlgButton(hwnd, 4221, global.active->lrheader);
            CheckDlgButton(hwnd, 4222, global.active->nofirstpage);

            TrapHelpInDialogBoxes(hwnd, "HeaderDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4211:
                    if (NOTIFYCODE == LBN_DBLCLK) {
                        SENDNOTIFY(hwnd, WM_COMMAND, 4201, BN_CLICKED, GetDlgItem(hwnd, 4201));     /* Edit button */
                    }
                    return (TRUE);

                case 4221:      /* Left/Right headers */
                    if (LeftRight != IsDlgButtonChecked(hwnd, 4221)) {
                        LeftRight = IsDlgButtonChecked(hwnd, 4221);

                        SendDlgItemMessage(hwnd, 4211, WM_SETREDRAW, FALSE, 0L);
                        SendDlgItemMessage(hwnd, 4211, LB_RESETCONTENT, 0, 0L);
                        if (LeftRight) {
                            SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Left Header");
                            SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Right Header");
                            SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Left Footer");
                            SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Right Footer");
                        } else {
                            SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Header");
                            SendDlgItemMessage(hwnd, 4211, LB_ADDSTRING, 0, (LONG) "Footer");
                        }
                        SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, 0, 0L);
                        SendDlgItemMessage(hwnd, 4211, WM_SETREDRAW, TRUE, 0L);
                        InvalidateRect(hwnd, NULL, TRUE);
                    }
                    return (TRUE);

                case 4201: {    /* Edit */
                    int i;

                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) {
                        MessageBeep(MB_ICONEXCLAMATION);
                        return (TRUE);
                    }

                    if (LeftRight) {
                        HeaderNum = i;
                    } else {
                        HeaderNum = i * 2;
                    }

                    i = DialogBox (hInstance, "EditHeader", hwnd, EditHeaderProc);

                    if (i) ShowWindow(GetDlgItem(hwnd, IDCANCEL), SW_HIDE);

                    return (TRUE);
                }

                case IDOK:
                    global.active->lrheader = IsDlgButtonChecked(hwnd, 4221);
                    global.active->nofirstpage = IsDlgButtonChecked(hwnd, 4222);
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
            }
            break;
    }

    return (FALSE);
}



static int SearchForFiles (HWND hwnd, KANJI *buffer, BOOL recurrsive)
{
    int i, k, n, len;
    int num = 0, dirs = 0;
    char *cp;
    HFILE fd;
    BOOL hascriteria = FALSE;
    OFSTRUCT of;
    FILEHEADER header;

#ifdef USEWIN32
    WIN32_FIND_DATA FindFileData;
    HANDLE hFindFile;
#else
    struct ffblk ffblk;
#endif USEWIN32


    /* Puts out a message */
    {
        char curdir[MAXFILENAMELEN];

        GetCurrentDirectory(MAXFILENAMELEN, curdir);
        sprintf(buffer, "Checking directory: %s", curdir);
        StatusMessage(buffer);
    }

    for (i = 0; i < NRSUMMARIES; i++) {
        if (criteria[i] != NULL) hascriteria = TRUE;
    }

    /* Get sub-directories */

#ifdef USEWIN32

    if (recurrsive &&
        (hFindFile = FindFirstFile("*.*", &FindFileData)) != INVALID_HANDLE_VALUE) {

        do {
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_TEMPORARY) continue;
            if (!(FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) continue;

            if (!strcmp(FindFileData.cFileName,".") || !strcmp(FindFileData.cFileName,"..")) continue;

            dirs++;
        } while (FindNextFile(hFindFile, &FindFileData));

        FindClose(hFindFile);

        if (dirs > 0) {
            i = 0;
            cp = (char *) MemAlloc(dirs * FILENAMELEN + 5);

            if ((hFindFile = FindFirstFile("*.*", &FindFileData)) != INVALID_HANDLE_VALUE) {
                do {
                    if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) continue;
                    if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) continue;
                    if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_TEMPORARY) continue;
                    if (!(FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) continue;

                    if (!strcmp(FindFileData.cFileName,".") || !strcmp(FindFileData.cFileName,"..")) continue;

                    strcpy(cp + i * FILENAMELEN, FindFileData.cFileName);
                    i++;
                } while (FindNextFile(hFindFile, &FindFileData));

                FindClose(hFindFile);
            } else {
                FreeMem(cp);
            }

            dirs = i;
        }
    }

#else

    if (recurrsive && findfirst("*.*", &ffblk, FA_RDONLY | FA_DIREC) == 0) {
        do {
            if (ffblk.ff_name[0] == '.') continue;
            if (!(ffblk.ff_attrib & FA_DIREC)) continue;
            dirs++;
        } while (findnext(&ffblk) == 0);

        if (dirs > 0) {
            i = 0;
            cp = (char *) MemAlloc(dirs * FILENAMELEN + 5);

            findfirst("*.*", &ffblk, FA_RDONLY | FA_DIREC);

            do {
                if (ffblk.ff_name[0] == '.') continue;
                if (!(ffblk.ff_attrib & FA_DIREC)) continue;
                strcpy(cp + i * FILENAMELEN, ffblk.ff_name);
                i++;
            } while (findnext(&ffblk) == 0);
        }
    }

#endif USEWIN32

    /* Get files */

#ifdef USEWIN32

    if ((hFindFile = FindFirstFile(FileSelectionSpec == NULL ? "*.*" : FileSelectionSpec, &FindFileData)) != INVALID_HANDLE_VALUE) {
        do {
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_TEMPORARY) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;

            fd = OpenFile(FindFileData.cFileName, &of, OF_READ);
            if (fd == HFILE_ERROR) continue;

            if (hascriteria) {
                _llseek(fd, 0L, 0);
                _lread(fd, &header, sizeof(FILEHEADER));
                if (header.magic == MAGIC && header.summary) {
                    for (i = 0; i < NRSUMMARIES; i++) {
                        ReadShortInt(fd, &n);

                        if (n > 0) _lread(fd, (char *) buffer, n * sizeof(KANJI));
                        if (criteria[i] == NULL) continue;

                        len = kanjilen(criteria[i]);
                        if (n < len) break;
                        for (k = 0; k < n - len; k++) {
                            if (!kanjincmp(criteria[i], buffer + k, len)) break;
                        }
                        if (k >= n - len) break;
                    }

                    if (i >= NRSUMMARIES) {
                        SendMessage(hwnd, LB_ADDSTRING, 0, (LONG) ((char far *) &of.szPathName));
                        num++;
                    }
                }
            } else {
                SendMessage(hwnd, LB_ADDSTRING, 0, (LONG) ((char far *) &of.szPathName));
                num++;
            }
            _lclose(fd);
        } while (FindNextFile(hFindFile, &FindFileData));

        FindClose(hFindFile);
    }

#else

    if (findfirst(FileSelectionSpec == NULL ? "*.*" : FileSelectionSpec, &ffblk, FA_RDONLY) == 0) {
        do {
            if (ffblk.ff_name[0] == '.') continue;
            if (ffblk.ff_attrib & FA_DIREC) continue;

            fd = OpenFile(ffblk.ff_name, &of, OF_READ);
            if (fd == HFILE_ERROR) continue;

            if (hascriteria) {
                _llseek(fd, 0L, 0);
                _lread(fd, &header, sizeof(FILEHEADER));
                if (header.magic == MAGIC && header.summary) {
                    for (i = 0; i < NRSUMMARIES; i++) {
                        ReadShortInt(fd, &n);

                        if (n > 0) _lread(fd, (char *) buffer, n * sizeof(KANJI));
                        if (criteria[i] == NULL) continue;

                        len = kanjilen(criteria[i]);
                        if (n < len) break;
                        for (k = 0; k < n - len; k++) {
                            if (!kanjincmp(criteria[i], buffer + k, len)) break;
                        }
                        if (k >= n - len) break;
                    }

                    if (i >= NRSUMMARIES) {
                        SendMessage(hwnd, LB_ADDSTRING, 0, (LONG) ((char far *) &of.szPathName));
                        num++;
                    }
                }
            } else {
                SendMessage(hwnd, LB_ADDSTRING, 0, (LONG) ((char far *) &of.szPathName));
                num++;
            }
            _lclose(fd);
        } while (findnext(&ffblk) == 0);
    }

#endif USEWIN32


    /* Recurrse into sub-directories */

    if (dirs > 0) {
        for (i = 0; i < dirs; i++) {
            if (!SetCurrentDirectory(cp + i * FILENAMELEN)) continue;
            num += SearchForFiles(hwnd, buffer, TRUE);
            SetCurrentDirectory("..");
        }

        FreeMem(cp);
    }

    return (num);
}



BOOL CALLBACK CriteriaProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i;
            FILEOPTIONS *f;

            /* Set the type and mode-change icon of the Jedit controls */

            SendDlgItemMessage(hwnd, 4211, EM_LIMITTEXT, 120, 0L);
            if (FileSelectionSpec != NULL) SetDlgItemText(hwnd, 4211, FileSelectionSpec);

            for (i = 0; i < NRSUMMARIES; i++) {
                SendDlgItemMessage(hwnd, 4201 + i, EM_SETRECT, 0, (LONG) GetDlgItem(hwnd, 4221));
                f = (FILEOPTIONS *) SendDlgItemMessage(hwnd, 4201 + i, EM_GETHANDLE, 0, 0L);

                if (criteria[i] != NULL) {
                    SendDlgItemMessage(hwnd, 4201 + i, EM_REPLACESEL, 0, (LONG) criteria[i]);
                }
                if (i == 0) {
                    SendDlgItemMessage(hwnd, 4221, EM_SETHANDLE, 0, (LONG) f->hwnd);  /* icon-change */
                }
            }

            SetFocus(GetDlgItem(hwnd, 4211));

            TrapHelpInDialogBoxes(hwnd, "FileCriteriaDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

        case WM_PAINT: {
            int i;
            HDC hdc;
            PAINTSTRUCT ps;

            hdc = BeginPaint(hwnd, &ps);

            for (i = 0; i < NRSUMMARIES; i++) DrawBoundingBox(hwnd, hdc, 4201 + i);

            EndPaint(hwnd, &ps);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK: {
                    int i, len;
                    KANJI far *up, far *kp;
                    char far *cbufp;
                    KANJI buffer[MAXLINELEN];


                    cbufp = (char far *) buffer;
                    GetDlgItemText(hwnd, 4211, cbufp, 120);

                    /* Trims spaces */

                    for (i = _fstrlen(cbufp) - 1; i >= 0 && cbufp[i] <= ' '; i--);
                    cbufp[i+1] = '\0';
                    for (i = 0; cbufp[i] && cbufp[i] <= ' '; i++);
                    if (!cbufp[i]) {
                        if (FileSelectionSpec != NULL) FreeMem(FileSelectionSpec);
                        FileSelectionSpec = NULL;
                    } else {
                        if (FileSelectionSpec != NULL) FreeMem(FileSelectionSpec);
                        FileSelectionSpec = (char *) MemAlloc(_fstrlen(cbufp + i) + 5);
                        _fstrcpy(FileSelectionSpec, cbufp + i);
                    }

                    /* Get the summary criteria */

                    for (i = 0; i < NRSUMMARIES; i++) {
                        up = (KANJI far *) SendDlgItemMessage(hwnd, 4201 + i, EM_GETLINE, 0, 0L);
                        len = kanjilen(up);

                        if (len <= 0) {
                            if (criteria[i] != NULL) {
                                FreeBlock(criteria[i]);
                                criteria[i] = NULL;
                            }
                        } else {
                            kp = criteria[i];
                            if (kp == NULL) {
                                kp = (KANJI far *) BlockAlloc((len + 5) * sizeof(KANJI));
                            } else if ((SegHeapGetSize(kp) / sizeof(KANJI)) < len + 5) {
                                FreeBlock(kp);
                                kp = (KANJI far *) BlockAlloc((len + 5) * sizeof(KANJI));
                            }
                            kanjicpy(kp, up);
                            criteria[i] = kp;
                        }
                    }
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
                }

                case IDCANCEL:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            break;
    }

    return (FALSE);
}



BOOL CALLBACK FileInfoProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i, n;
            HFILE fd;
            OFSTRUCT of;
            FILEHEADER header;
            char buffer[MAXLINELEN];

#ifdef USEWIN32
            WIN32_FIND_DATA FindFileData;
            HANDLE hFindFile;
#else
            struct stat statbuf;
            struct ffblk ffblk;
#endif USEWIN32

            SetDlgItemText(hwnd, 4211, FileInfoName);

#ifdef USEWIN32
            if ((hFindFile = FindFirstFile(FileInfoName, &FindFileData)) == INVALID_HANDLE_VALUE)
                return (TRUE);

            FindClose(hFindFile);
#else
            if (findfirst(FileInfoName, &ffblk, FA_RDONLY) != 0) return (TRUE);

#endif USEWIN32

            /* File date and time */

#ifdef USEWIN32
            SetDlgItemText(hwnd, 4212, TimeString(&FindFileData.ftCreationTime));
            SetDlgItemText(hwnd, 4213, TimeString(&FindFileData.ftLastWriteTime));
            sprintf(buffer, "%ld bytes, %4.2lf K", FindFileData.nFileSizeLow, ((double) FindFileData.nFileSizeLow / 1024.0));
            SetDlgItemText(hwnd, 4214, buffer);
#else
            if (stat(FileInfoName, &statbuf) == 0) {
                SetDlgItemText(hwnd, 4213, ctime(&statbuf.st_mtime));
                sprintf(buffer, "%ld bytes, %4.2lf K", statbuf.st_size, ((double) statbuf.st_size / 1024.0));
                SetDlgItemText(hwnd, 4214, buffer);
            }
#endif USEWIN32

            /* File type */

            fd = OpenFile(FileInfoName, &of, OF_READ);
            if (fd == HFILE_ERROR) return (TRUE);

            _llseek(fd, 0L, 0);
            _lread(fd, &header, sizeof(FILEHEADER));

            if (header.magic == MAGIC) {
                strcpy(buffer, PROGNAME " File");
            } else {
                strcpy(buffer, "DOS file");
            }

#ifdef USEWIN32
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                if (buffer[0]) strcat(buffer, ", ");
                strcat(buffer, "Directory / Folder");
            }
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_READONLY) {
                if (buffer[0]) strcat(buffer, ", ");
                strcat(buffer, "Read-Only");
            }
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE) {
                if (buffer[0]) strcat(buffer, ", ");
                strcat(buffer, "Archive bit set");
            }
#else
            if (ffblk.ff_attrib & FA_DIREC) {
                if (buffer[0]) strcat(buffer, ", ");
                strcat(buffer, "Directory");
            }
            if (ffblk.ff_attrib & FA_RDONLY) {
                if (buffer[0]) strcat(buffer, ", ");
                strcat(buffer, "Read-Only");
            }
            if (ffblk.ff_attrib & FA_ARCH) {
                if (buffer[0]) strcat(buffer, ", ");
                strcat(buffer, "Archive bit set");
            }
#endif USEWIN32

            SetDlgItemText(hwnd, 4215, buffer);

            /* Summary? */

            if (header.magic == MAGIC && header.summary) {
                for (i = 0; i < NRSUMMARIES; i++) {
                    ReadShortInt(fd, &n);

                    if (n <= 0) continue;
                    if (n > 0) _lread(fd, buffer, n * sizeof(KANJI));
                    SendDlgItemMessage(hwnd, 4201 + i, EM_REPLACESEL, 0, (LONG) buffer);
                }
            }

            _lclose(fd);

            TrapHelpInDialogBoxes(hwnd, "FileInfoDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK:
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
            }
            break;

        case WM_KEYDOWN:
            switch (wParam) {
                case VK_ESCAPE: SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                                return (TRUE);
            }
            break;
    }

    return (FALSE);
}



static int FillDirectories (HWND ListBoxHwnd, HWND TextHwnd)
{
    int i, n = 0;
    char buffer[MAXFILENAMELEN];
    char drive[5];

#ifdef USEWIN32
    WIN32_FIND_DATA FindFileData;
    HANDLE hFindFile;
#else
    struct ffblk ffblk;
#endif

    if (GetCurrentDirectory(MAXFILENAMELEN, buffer) <= 0) {
        if (!SetCurrentDirectory("\\")) return (-1);
        strcpy(buffer, "\\");
    }

    /* Fill the directory combo box */

    SendMessage(ListBoxHwnd, WM_SETREDRAW, FALSE, 0L);
    SendMessage(ListBoxHwnd, LB_RESETCONTENT, 0, 0L);

    if (strcmp(buffer + 2, "\\")) {
        SendMessage(ListBoxHwnd, LB_ADDSTRING, 0, (LONG) ((char far *) PARENTDIR));
        n++;
        strcat(buffer, "\\*.*");
    } else {
        strcat(buffer, "*.*");
    }

#ifdef USEWIN32
    if ((hFindFile = FindFirstFile(buffer, &FindFileData)) != INVALID_HANDLE_VALUE) {
        do {
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) continue;
            if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_TEMPORARY) continue;
            if (!(FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) continue;

            if (!strcmp(FindFileData.cFileName,".") || !strcmp(FindFileData.cFileName,"..")) continue;

            n++;
            SendMessage(ListBoxHwnd, LB_ADDSTRING, 0, (LONG) ((char far *) FindFileData.cFileName));
        } while (FindNextFile(hFindFile, &FindFileData));

        FindClose(hFindFile);
    }
#else
    if (findfirst(buffer, &ffblk, FA_RDONLY | FA_DIREC) == 0) {
        do {
            if (!(ffblk.ff_attrib & FA_DIREC)) continue;
            if (ffblk.ff_name[0] == '.') continue;

            n++;
            SendMessage(ListBoxHwnd, LB_ADDSTRING, 0, (LONG) ((char far *) ffblk.ff_name));
        } while (findnext(&ffblk) == 0);
    }
#endif USEWIN32

    /* Fill the drives */

    for (i = 0; i < 26; i++) {
        if (i == buffer[0] - 'A') continue;
#ifdef USEWIN32
        sprintf(drive, "%c:\\", 'A' + i);
        switch (GetDriveType(drive)) {
#else
        switch (GetDriveType(i)) {
#endif USEWIN32
            case DRIVE_REMOVABLE:
                sprintf(buffer, "%c %c: floppy drive ]", DRIVECHAR, 'A' + i);
                break;

            case DRIVE_REMOTE:
                sprintf(buffer, "%c %c: network drive ]", DRIVECHAR, 'A' + i);
                break;

            case DRIVE_FIXED:
                sprintf(buffer, "%c %c: hard drive ]", DRIVECHAR, 'A' + i);
                break;

            default: continue;
        }
        SendMessage(ListBoxHwnd, LB_ADDSTRING, 0, (LONG) ((char far *) buffer));
    }

    if (SendMessage(ListBoxHwnd, LB_GETCOUNT, 0, 0L) > 0) {
        SendMessage(ListBoxHwnd, LB_SETCURSEL, 0, 0L);
    }

    GetCurrentDirectory(MAXFILENAMELEN, buffer);
    SetWindowText(TextHwnd, buffer);

    SendMessage(ListBoxHwnd, WM_SETREDRAW, TRUE, 0L);
    InvalidateRect(ListBoxHwnd, NULL, TRUE);

    return (n);
}


BOOL CALLBACK FileFindProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG:
            FillDirectories(GetDlgItem(hwnd, 4221), GetDlgItem(hwnd, 4222));

            /* Default to search only the current path */

            SendDlgItemMessage(hwnd, 4231, BM_SETCHECK, TRUE, 0L);
            EnableWindow(GetDlgItem(hwnd, 4201), FALSE);
            EnableWindow(GetDlgItem(hwnd, 4202), FALSE);

            TrapHelpInDialogBoxes(hwnd, "FindFileDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case 4211:      /* Files list */
                    switch (NOTIFYCODE) {
                        case LBN_DBLCLK:
                            SendMessage(hwnd, WM_COMMAND, 4201, 0L);    /* Open */
                            return (TRUE);
                    }
                    return (TRUE);

                case 4221: {    /* Directories */
                    int i;
                    char buffer[MAXFILENAMELEN];

                    switch (NOTIFYCODE) {
                        case LBN_DBLCLK:
                            i = SendDlgItemMessage(hwnd, 4221, LB_GETCURSEL, 0, 0L);
                            if (i == LB_ERR) {
                                MessageBeep(MB_ICONEXCLAMATION);
                                return (TRUE);
                            }
                            SendDlgItemMessage(hwnd, 4221, LB_GETTEXT, i, (LONG) ((char far *) buffer));
                            if (!strcmp(buffer, PARENTDIR)) {
                                SetCurrentDirectory("..");
                            } else if (buffer[0] == DRIVECHAR) {
#ifdef USEWIN32
                                char drivename[5];
                                sprintf(drivename, "%c:", buffer[2]);
                                SetCurrentDirectory(drivename);
#else USEWIN32
                                setdisk(buffer[2] - 'A');
#endif USEWIN32
                            } else {
                                SetCurrentDirectory(buffer);
                            }
                            FillDirectories(GetDlgItem(hwnd, 4221), GetDlgItem(hwnd, 4222));
                            return (TRUE);
                    }
                    return (TRUE);
                }

                case 4231:      /* Current path */
                    if (IsDlgButtonChecked(hwnd, 4231)) {
                        EnableWindow(GetDlgItem(hwnd, 4221), TRUE);
                        EnableWindow(GetDlgItem(hwnd, 4222), TRUE);
                    }
                    return (TRUE);

                case 4232:      /* All subdirectories */
                    if (IsDlgButtonChecked(hwnd, 4232)) {
                        EnableWindow(GetDlgItem(hwnd, 4221), TRUE);
                        EnableWindow(GetDlgItem(hwnd, 4222), TRUE);
                    }
                    return (TRUE);

                case 4233:      /* Entire disk */
                    if (IsDlgButtonChecked(hwnd, 4233)) {
                        EnableWindow(GetDlgItem(hwnd, 4221), FALSE);
                        EnableWindow(GetDlgItem(hwnd, 4222), FALSE);
                    }
                    return (TRUE);

                case 4234:      /* All drives */
                    if (IsDlgButtonChecked(hwnd, 4234)) {
                        EnableWindow(GetDlgItem(hwnd, 4221), FALSE);
                        EnableWindow(GetDlgItem(hwnd, 4222), FALSE);
                    }
                    return (TRUE);

                case 4201: {    /* Open */
                    int i;
                    char buffer[MAXFILENAMELEN];

                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) return (TRUE);
                    SendDlgItemMessage(hwnd, 4211, LB_GETTEXT, i, (LONG) ((char far *) buffer));
                    if (!DoFileOpen(buffer)) return (TRUE);

                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
                }

                case 4202: {    /* Info */
                    int i;
                    char buffer[MAXFILENAMELEN];

                    i = SendDlgItemMessage(hwnd, 4211, LB_GETCURSEL, 0, 0L);
                    if (i == LB_ERR) return (TRUE);
                    SendDlgItemMessage(hwnd, 4211, LB_GETTEXT, i, (LONG) ((char far *) buffer));
                    FileInfoName = buffer;

                    DialogBox(hInstance, "FileInfo", hwnd, FileInfoProc);
                    return (TRUE);
                }

                case IDOK: {    /* Search */
                    BOOL Doit;
                    HCURSOR hcursor;

                    Doit = DialogBox(hInstance, "SearchCriteria", hwnd, CriteriaProc);

                    if (Doit) {
                        int i, n;
                        BOOL Recurrsive = FALSE;
                        char DriveName[10];
                        char OldPath[MAXFILENAMELEN];
                        KANJI buffer[MAXLINELEN];

                        if (!GetCurrentDirectory(MAXFILENAMELEN, OldPath))
                            return (TRUE);

                        if (IsDlgButtonChecked(hwnd, 4232)) {   /* Sub-directories */
                            Recurrsive = TRUE;
                        }
                        if (IsDlgButtonChecked(hwnd, 4233)) {   /* Entire disk */
                            SetCurrentDirectory("\\");
                            Recurrsive = TRUE;
                        }
                        if (IsDlgButtonChecked(hwnd, 4234)) {   /* All disks */
                            Recurrsive = TRUE;
                        }

                        hcursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
                        ShowCursor(TRUE);

                        SendDlgItemMessage(hwnd, 4211, LB_RESETCONTENT, 0, 0L);

                        if (IsDlgButtonChecked(hwnd, 4234)) {
                            for (n = i = 0; i < 26; i++) {
                                sprintf(DriveName, "%c:\\", 'A' + i);
#ifdef USEWIN32
                                switch (GetDriveType(DriveName)) {
#else
                                switch (GetDriveType(i)) {
#endif USEWIN32
                                    default:
                                    case DRIVE_REMOVABLE: break;

                                    case DRIVE_REMOTE:
                                    case DRIVE_FIXED:
                                        if (!SetCurrentDirectory(DriveName)) break;
                                        n += SearchForFiles (GetDlgItem(hwnd, 4211), buffer, TRUE);
                                        break;
                                }
                            }
                        } else {
                            n = SearchForFiles (GetDlgItem(hwnd, 4211), buffer, Recurrsive);
                        }

                        SetCurrentDirectory(OldPath);
                        FillDirectories(GetDlgItem(hwnd, 4221), GetDlgItem(hwnd, 4222));

                        if (n <= 0) {
                            ShowCursor(FALSE);
                            SetCursor(hcursor);
                            StatusMessage("");
                            ErrorMessage (hwnd, "No files found!");
                            EnableWindow(GetDlgItem(hwnd, 4201), FALSE);
                            EnableWindow(GetDlgItem(hwnd, 4202), FALSE);
                        } else {
                            ShowCursor(FALSE);
                            SetCursor(hcursor);
                            sprintf((char *) buffer, "%d file%s found.", n, (n > 1) ? "s" : "");
                            StatusMessage((char *) buffer);
                            SendDlgItemMessage(hwnd, 4211, LB_SETCURSEL, 0, 0L);
                            EnableWindow(GetDlgItem(hwnd, 4201), TRUE);
                            EnableWindow(GetDlgItem(hwnd, 4202), TRUE);
                            SetFocus(GetDlgItem(hwnd, 4211));
                        }

                    }

                    return (TRUE);
                }

                case IDCANCEL:  /* Cancel */
                    UnhookHelpTraps(hwnd, TRUE);
                    EndDialog(hwnd, FALSE);
                    return (TRUE);
            }
            break;
    }

    return (FALSE);
}



void UpdateQuickFiles (void)
{
    int i, j, n;
    char buffer[MAXLINELEN];
    char curdir[MAXFILENAMELEN];

    for (i = j = 0; i < NRQUICKFILE; i++) {
        if (QuickFiles[i].fullname != NULL) j++;
        DeleteMenu(hmenu, IDM_FILE1 + i, MF_BYCOMMAND);
    }

    /* Find the last file menu item */

    for (n = GetMenuItemCount(hmenu) - 1; n >= 0; n--) {
        if (GetMenuItemID(GetSubMenu(hmenu, n), 0) == IDM_FILENEW) break;
    }

    if (n < 0) return;

    for (i = GetMenuItemCount(GetSubMenu(hmenu, n)) - 1; i >= 0; i--) {
        if (GetMenuItemID(GetSubMenu(hmenu, n), i) == IDM_FILEEXIT) break;
    }

    if (i >= 0) DeleteMenu(GetSubMenu(hmenu, n), i+1, MF_BYPOSITION);

    if (j > 0) {
        InsertMenu(GetSubMenu(hmenu, n), i+1, MF_BYPOSITION | MF_SEPARATOR, 0, "");

        curdir[0] = '\0';
        GetCurrentDirectory(MAXFILENAMELEN, curdir);

        for (i = 0; i < NRQUICKFILE && QuickFiles[i].fullname != NULL; i++) {
            if (!strnicmp(curdir, QuickFiles[i].fullname,
                          QuickFiles[i].filename - QuickFiles[i].fullname - 1)) {

                sprintf(buffer, "&%d %s", i + 1, QuickFiles[i].filename);
            } else {
                sprintf(buffer, "&%d %s", i + 1, QuickFiles[i].fullname);
            }
            InsertMenu(GetSubMenu(hmenu, n), -1, MF_BYPOSITION | MF_UNCHECKED | MF_ENABLED | MF_STRING, IDM_FILE1 + i, buffer);
        }
    }
}



void PutQuickFile (char *pathname)
{
    int i, j;

    if (pathname == NULL) {
        for (i = 0; i < NRQUICKFILE; i++)
            QuickFiles[i].fullname = QuickFiles[i].filename = NULL;
        return;
    }

    for (i = 0; i < NRQUICKFILE; i++) {
        if (QuickFiles[i].fullname == NULL) continue;
        if (!stricmp(QuickFiles[i].fullname, pathname)) {
            FreeMem(QuickFiles[i].fullname);
            for (j = i+1; j < NRQUICKFILE; j++) QuickFiles[j-1] = QuickFiles[j];
            QuickFiles[NRQUICKFILE-1].fullname = NULL;
        }
    }

    if (QuickFiles[NRQUICKFILE-1].fullname != NULL)
        FreeMem(QuickFiles[NRQUICKFILE-1].fullname);

    for (i = NRQUICKFILE - 1; i > 0; i--) QuickFiles[i] = QuickFiles[i-1];

    QuickFiles[0].fullname = MemAlloc(strlen(pathname) + 5);
    strcpy(QuickFiles[0].fullname, pathname);

#ifndef USEWIN32
    //AnsiUpper(QuickFiles[0].fullname);
#endif USEWIN32

    for (i = strlen(pathname)-1; i >= 0; i--) {
        if (pathname[i] == '\\') break;
    }

    if (i >= 0) {
        QuickFiles[0].filename = QuickFiles[0].fullname + i + 1;
    } else {
        QuickFiles[0].filename = QuickFiles[0].fullname;
    }

    UpdateQuickFiles();
    OptionsChanged = TRUE;
}


char *GetQuickFile (int n)
{
    return (n >= NRQUICKFILE ? NULL : QuickFiles[n].fullname);
}



BOOL CALLBACK RunProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i;

            for (i = 0; i < NRPROGS; i++) {
                SendDlgItemMessage(hwnd, 4201 + i, BM_SETCHECK, (i == 0), 0L);
            }

            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND:
            switch (WINCMDID) {
                case IDOK: {
                    int len;
                    char *cp;
                    char pathname[MAXLINELEN];

                    GetWindowsDirectory(pathname, MAXLINELEN);
                    len = strlen(pathname);
                    if (pathname[len-1] == '\\') pathname[len-1] = '\0';

                    if (IsDlgButtonChecked(hwnd, 4201)) {
                        strcat(pathname, "\\CONTROL.EXE");
                        cp = "Control Panel";
                    } else if (IsDlgButtonChecked(hwnd, 4202)) {
                        strcat(pathname, "\\WINFILE.EXE");
                        cp = "File Manager";
                    } else if (IsDlgButtonChecked(hwnd, 4203)) {
                        sprintf(pathname, "%sBTNMAD.EXE", global.jwppath);
                        cp = "Button Madness";
                    } else if (IsDlgButtonChecked(hwnd, 4204)) {
                        sprintf(pathname, "%sPUZZLE.EXE", global.jwppath);
                        cp = "Puzzle";
                    } else if (IsDlgButtonChecked(hwnd, 4205)) {
                        sprintf(pathname, "%sMB.EXE", global.jwppath);
                        cp = "Mile Bones";
                    } else {
                        pathname[0] = '\0';
                        cp = NULL;
                    }

                    if (cp != NULL) {
                        if (WinExec(pathname, SW_SHOW) < 32)
                            ErrorMessage(hwnd, "Cannot run %s!", cp);
                    }
                    EndDialog(hwnd, TRUE);
                    return (TRUE);
                }

                case IDCANCEL:
                    EndDialog(hwnd, FALSE);
                    return (TRUE);

                case 4201:
                case 4202:
                case 4203:
                case 4204:
                case 4205: {
                    int i;

                    for (i = 0; i < NRPROGS; i++)
                        SendDlgItemMessage(hwnd, 4201 + i, BM_SETCHECK, (4201 + i == WINCMDID), 0L);

                    if (NOTIFYCODE == BN_DOUBLECLICKED) SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    return (TRUE);
                }
            }
            break;
    }

    return (FALSE);
}



static KANJI far *ConvertOffset (int id, LONG lParam, KANJI *buf)
{
    if (Templates == NULL) {
        buf[0] = 0;
        return (buf);
    }
    return (Templates[lParam].desc);
}



BOOL CALLBACK TemplateProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) {
        case WM_INITDIALOG: {
            int i, j;
            HFILE fd;
            OFSTRUCT of;
            FILEHEADER header;
            KANJI tplname[MAXLINELEN];
            char buffer[MAXLINELEN];

#ifdef USEWIN32
            WIN32_FIND_DATA FindFileData;
            HANDLE hFindFile;
#else
            struct ffblk ffblk;
#endif USEWIN32


            sprintf(buffer, "%s*%s", global.templatepath, FileExtensions[FF_TEMPLATE]);
            StatusMessage("Loading templates...");

            j = 0;

#ifdef USEWIN32

            if ((hFindFile = FindFirstFile(buffer, &FindFileData)) != INVALID_HANDLE_VALUE) {
                do {
                    strcpy(buffer, global.templatepath);
                    strcat(buffer, FindFileData.cFileName);
                    fd = OpenFile(buffer, &of, OF_READ);
                    if (fd == HFILE_ERROR) continue;

                    _llseek(fd, 0L, 0);
                    _lread(fd, &header, sizeof(FILEHEADER));
                    if (header.magic != MAGIC) {
                        _lclose(fd);
                        continue;
                    }

                    if (!header.summary) {
                        for (i = 0; FindFileData.cFileName[i]; i++) tplname[i] = FindFileData.cFileName[i];
                        tplname[i] = 0;
                    } else {
                        ReadShortInt(fd, &i);

                        if (i > 0) _lread(fd, (char *) tplname, i * sizeof(KANJI));
                        else {
                            for (i = 0; FindFileData.cFileName[i]; i++) tplname[i] = FindFileData.cFileName[i];
                            tplname[i] = 0;
                        }
                    }

                    _lclose(fd);

                    if (Templates == NULL) {
                        Templates = (TPLFILE far *) BlockAlloc(TPLBLOCKSIZE * sizeof(TPLFILE));
                    } else if (j % TPLBLOCKSIZE == 0) {
                        Templates = (TPLFILE far *) BlockRealloc(Templates, (j + TPLBLOCKSIZE) * sizeof(TPLFILE));
                    }
                    Templates[j].filename = (char far *) BlockAlloc(strlen(of.szPathName) + 5);
                    _fstrcpy(Templates[j].filename, of.szPathName);
                    Templates[j].desc = (KANJI far *) BlockAlloc((kanjilen(tplname) + 5) * sizeof(KANJI));
                    kanjicpy(Templates[j].desc, tplname);

                    SendDlgItemMessage(hwnd, 4201, LB_ADDSTRING, 0, (LONG) j);
                    j++;
                } while (FindNextFile(hFindFile, &FindFileData));
            }

            FindClose(hFindFile);

#else

            if (findfirst(buffer, &ffblk, FA_RDONLY) == 0) {
                do {
                    strcpy(buffer, global.templatepath);
                    strcat(buffer, ffblk.ff_name);
                    fd = OpenFile(buffer, &of, OF_READ);
                    if (fd == HFILE_ERROR) continue;

                    _llseek(fd, 0L, 0);
                    _lread(fd, &header, sizeof(FILEHEADER));
                    if (header.magic != MAGIC) {
                        _lclose(fd);
                        continue;
                    }

                    if (!header.summary) {
                        for (i = 0; ffblk.ff_name[i]; i++) tplname[i] = ffblk.ff_name[i];
                        tplname[i] = 0;
                    } else {
                        ReadShortInt(fd, &i);

                        if (i > 0) _lread(fd, (char *) tplname, i * sizeof(KANJI));
                        else {
                            for (i = 0; ffblk.ff_name[i]; i++) tplname[i] = ffblk.ff_name[i];
                            tplname[i] = 0;
                        }
                    }

                    _lclose(fd);

                    if (Templates == NULL) {
                        Templates = (TPLFILE far *) BlockAlloc(TPLBLOCKSIZE * sizeof(TPLFILE));
                    } else if (j % TPLBLOCKSIZE == 0) {
                        Templates = (TPLFILE far *) BlockRealloc(Templates, (j + TPLBLOCKSIZE) * sizeof(TPLFILE));
                    }
                    Templates[j].filename = (char far *) BlockAlloc(strlen(of.szPathName) + 5);
                    _fstrcpy(Templates[j].filename, of.szPathName);
                    Templates[j].desc = (KANJI far *) BlockAlloc((kanjilen(tplname) + 5) * sizeof(KANJI));
                    kanjicpy(Templates[j].desc, tplname);

                    SendDlgItemMessage(hwnd, 4201, LB_ADDSTRING, 0, (LONG) j);
                    j++;
                } while (findnext(&ffblk) == 0);
            }

#endif USEWIN32

            StatusMessage("");

            if (j <= 0) {
                EndDialog(hwnd, -1);
                return (TRUE);
            }

            SendDlgItemMessage(hwnd, 4201, LB_SETCURSEL, 0, 0L);

            nr_templates = j;

            TrapHelpInDialogBoxes(hwnd, "FileNewDlg", 1, TRUE);
            CenterDialogBox(hwnd);
            return (TRUE);
        }

#ifdef USEWIN32
#ifdef CTL3D
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORDLG:
        case WM_CTLCOLOREDIT:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORMSGBOX:
        case WM_CTLCOLORSCROLLBAR:
        case WM_CTLCOLORSTATIC:
           return Ctl3dCtlColorEx(message, wParam, lParam);
#endif CTL3D
#endif USEWIN32

        case WM_COMMAND: {
            int i, j = -5;

            switch (WINCMDID) {
                case 4201:          /* Japanese list box */
                    if (NOTIFYCODE == LBN_DBLCLK) {
                        SendMessage(hwnd, WM_COMMAND, IDOK, 0L);
                    }
                    return (TRUE);

                case IDOK:
                    j = SendDlgItemMessage(hwnd, 4201, LB_GETCURSEL, 0, 0L);
                    if (j == LB_ERR) {
                        j = -1;
                        if (TplChosen != NULL) TplChosen[0] = '\0';
                    } else {
                        j = SendDlgItemMessage(hwnd, 4201, LB_GETITEMDATA, j, 0L);
                        if (TplChosen == NULL) TplChosen = BlockAlloc(MAXFILENAMELEN);
                        _fstrcpy(TplChosen, Templates[j].filename);
                    }
                    break;

                case IDCANCEL:
                    j = -2;
                    if (TplChosen != NULL) TplChosen[0] = '\0';
                    break;

                case 4211:
                    j = -1;
                    if (TplChosen != NULL) TplChosen[0] = '\0';
                    break;
            }

            if (j <= -5) return (FALSE);

            for (i = 0; i < nr_templates; i++) {
                FreeBlock(Templates[i].filename);
                FreeBlock(Templates[i].desc);
            }
            FreeBlock(Templates);
            Templates = NULL;

            UnhookHelpTraps(hwnd, TRUE);
            EndDialog(hwnd, j);
            return (TRUE);
        }

        case WM_COMPAREITEM:
        case WM_DELETEITEM:
        case WM_DRAWITEM:
        case WM_MEASUREITEM:
            return (JlistProc(hwnd, message, wParam, lParam, TRUE, ConvertOffset));
    }

    return (FALSE);
}
