/* Copyright (C) Stephen Chung, 1994.  All rights reserved. */

#include <stdio.h>

typedef struct {
    unsigned char strokes;
    unsigned char bushu;
    unsigned short int index;
    unsigned char grade;
    unsigned char onkun;
    unsigned char meanings:4;
    unsigned char padding;           /* To pad it to 32-bits */
    unsigned long int offset;
} JWPRECORD;

typedef struct {
    unsigned char strokes;
    unsigned char bushu;
    unsigned short int indexn;   /* Nelson index no */
    unsigned short int indexh;   /* Halpern index no */
    unsigned char grade;
    unsigned char onkun;         /* two 4-bit counts */
    unsigned char meanings;
    unsigned char version;       /* current value is 23 (2.3) */
    long int offset;
} JDICRECORD;


short int IndexToJIS(int index)
{
    int div, rem;
    short int ch;

    div = index / 94;
    rem = index % 94;

    ch = (div << 8) | rem;
    ch += 0x3021;

    return (ch);
}



main()
{
    long int i, j;
    FILE *ifp, *ofp;
    JDICRECORD record;
    JWPRECORD jwprecord;
    int num;
    long int offsetdiff;
    char line[1024];

    printf("MKKINFO - Converts Jim Breen's v2.3 KINFO.DAT file for use with JWP\n");
    printf("          All Nelson and Halpern information will be lost.\n");
    printf("          Please make sure the original file KINFO.DAT is in the\n");
    printf("          same directory.\n\n");

    ifp = fopen("kinfo.dat", "rb");
    if(ifp == NULL) {
        fprintf(stderr,"ERROR - cannot open KINFO.DAT!\n");
        exit(1);
    }
    ofp = fopen("jwpkinfo.dat", "wb");
    if(ifp == NULL) {
        fprintf(stderr,"ERROR - cannot open JWPKINFO.DAT!\n");
        exit(1);
    }


    /* Reads the number of records by calculating the first offset */

    fseek(ifp, 0L, 0);
    fread(&record, sizeof(JDICRECORD), 1, ifp);
    num = record.offset / sizeof(JDICRECORD);

    printf("%d records...\n", num);


    /* Calculates the difference in offsets */

    offsetdiff = sizeof(JDICRECORD) - sizeof(JWPRECORD);
    printf("Difference in record size = %ld bytes...\n", offsetdiff);
    offsetdiff *= num;
    printf("Difference in offset = %ld bytes...\n\n", offsetdiff);


    /* Converts the records */

    fseek(ifp, 0L, 0);
    fseek(ofp, 0L, 0);

    printf("Converting records...\n");

    for (i = j = 0; i < num; i++) {
        if (i % 10 == 9) printf("Record #%ld...\r", i);
        fread(&record, sizeof(JDICRECORD), 1, ifp);
        if (record.strokes <= 0 || record.bushu <= 0) continue;

        jwprecord.strokes = record.strokes;
        jwprecord.bushu = record.bushu;
        jwprecord.index = IndexToJIS(i);
        jwprecord.grade = record.grade;
        jwprecord.onkun = record.onkun;

        if (record.meanings > 15) {
            printf("   >>> Record #%ld, JIS %X: %d meanings truncated to 15\n",
                    i, jwprecord.index, record.meanings);
            jwprecord.meanings = 15;
        } else {
            jwprecord.meanings = record.meanings;
        }
        jwprecord.offset = record.offset - offsetdiff;

        j++;
    }

    printf("%ld records converted\n\n", j);

    printf("Copying strings...\n");
    while (fgets(line, 1024, ifp) != NULL) {
        fprintf(ofp, "%s", line);
    }
    printf("Strings copied.\n\n");

    fclose(ifp);
    fclose(ofp);

    printf(">>> A new file JWPKINFO.DAT has been created.  Store the original\n");
    printf(">>> KINFO.DAT in a safe place, then rename JWPKINFO.DAT to KINFO.DAT.\n");
    printf(">>> Run KINFOIDX.EXE to generate the index file KINFO.IDX.\n");

}
