/* Routines for the TOOLBAR library         */
/* Copyright (C) Stephen Chung, 1995.       */
/* All rights reserved.                     */


#include <windows.h>
#include "toolbar.h"

#define MAXSTATES       3


static BOOL ClassRegistered = FALSE;
static BOOL Capturing = FALSE;

long CALLBACK ToolbarProc (HWND, UINT, WPARAM, LPARAM);
long CALLBACK ToolbarButtonProc (HWND, UINT, WPARAM, LPARAM);


/* 3D Styles:

        0 = Normal 3D
        1 = Normal 3D hole
        2 = Normal 3D hole, white inside
*/

void Create3DEffect (HDC hdc, RECT *rect, int thickness, int style)
{
    int     i;
    int     x1 = rect->left;
    int     y1 = rect->top;
    int     x2 = rect->right;
    int     y2 = rect->bottom;
    HBRUSH  hbrush, oldbrush;
    HPEN    hpen, oldpen;

    oldpen = SelectObject(hdc, CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOWFRAME)));

    switch (style) {
        default:
        case 0:
        case 1: hbrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)); break;
        case 2: hbrush = GetStockObject(WHITE_BRUSH); break;
    }

    oldbrush = SelectObject(hdc, hbrush);
    Rectangle(hdc, x1, y1, x2, y2);

    switch (style) {
        default:
        case 0: hpen = CreatePen(PS_SOLID, 1, RGB(255, 255, 255));
                break;
        case 1: hpen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNSHADOW));
                break;
        case 2: hpen = CreatePen(PS_SOLID, 1, RGB(0, 0, 0));
                break;
    }
    DeleteObject(SelectObject(hdc, hpen));

    for (i = 1; i <= thickness; i++) {
        MoveToEx(hdc, x1 + i, y1 + i, NULL);    LineTo(hdc, x1 + i, y2 - 1);
        MoveToEx(hdc, x1 + i, y1 + i, NULL);    LineTo(hdc, x2 - 1, y1 + i);
    }

    switch (style) {
        default:
        case 0: hpen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNSHADOW));
                break;
        case 1: hpen = CreatePen(PS_SOLID, 1, RGB(255, 255, 255));
                break;
        case 2: hpen = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNSHADOW));
                break;
    }
    DeleteObject(SelectObject(hdc, hpen));

    for (i = 1; i <= thickness; i++) {
        MoveToEx(hdc, x1 + i, y2 - 1 - i, NULL);
        LineTo(hdc, x2 - 1, y2 - 1 - i);
        MoveToEx(hdc, x2 - 1 - i, y2 - 2, NULL);
        LineTo(hdc, x2 - 1 - i, y1 + i);
    }

    SelectObject(hdc, oldbrush);
    SelectObject(hdc, oldpen);
    DeleteObject(hpen);

    switch (style) {
        case 0:
        case 1: DeleteObject(hbrush); break;
        case 2: break;
    }
}



static void CursorShape (HWND hwnd, int state)
{
    HCURSOR hcursor;
    static HCURSOR arrow = NULL;

    if (hwnd == NULL) return;

    if (state < 0) {
        hcursor = (HCURSOR) GetProp(GetParent(hwnd), "TB CURSOR");
    } else {
        if (arrow == NULL) {
            hcursor = arrow = LoadCursor (NULL, IDC_ARROW);
        } else {
            hcursor = arrow;
        }
    }

    SetCursor(hcursor);
}


void EnableToolbarButton (HWND hwnd, int id, BOOL enabled)
{
    int             i, n;
    TOOLBARICON     *icons;

    if (hwnd == NULL) return;

    icons = (TOOLBARICON *) GetProp(hwnd, "TB ICONS");
    n = (int) GetProp(hwnd, "TB NUMICONS");

    for (i = 0; i < n && icons[i].id != id; i++);

    if (i >= n) return;

    if (enabled) {
        if (icons[i].state >= 0) return;
        icons[i].state = icons[i].oldstate;
    } else {
        if (icons[i].state < 0) return;
        icons[i].oldstate = icons[i].state;
        icons[i].state = -1;
    }
    InvalidateRect(icons[i].hwnd, NULL, FALSE);
    UpdateWindow(icons[i].hwnd);
}



HWND GetToolbarButton (HWND hwnd, int id, TOOLBARICON *ip)
{
    int             i, n;
    TOOLBARICON     *icons;

    if (hwnd == NULL) return (NULL);

    icons = (TOOLBARICON *) GetProp(hwnd, "TB ICONS");
    n = (int) GetProp(hwnd, "TB NUMICONS");

    for (i = 0; i < n && icons[i].id != id; i++);

    if (i >= n) return (NULL);

    if (ip != NULL) *ip = icons[i];

    return (icons[i].hwnd);
}



void ModifyToolbarButton (HWND hwnd, TOOLBARICON *icon)
{
    TOOLBARICON *old;

    if (hwnd == NULL) return;

    old = (TOOLBARICON *) GetProp(hwnd, "TB ICONS");

    old->id = icon->id;
    old->x = icon->x;
    old->y = icon->y;
    old->width = icon->width;
    old->height = icon->height;
    old->state = icon->state;
    old->cycle = icon->cycle;
    old->disabled = icon->disabled;
    old->undepressed = icon->undepressed;
    old->depressed = icon->depressed;
    old->grayed = icon->grayed;
    old->pressing = icon->pressing;

    InvalidateRect(hwnd, NULL, TRUE);
    UpdateWindow(hwnd);
}



HWND CreateToolbar (HWND parent, int x, int y, int width, int height,
                    int thickness, int id, int n, HANDLE hInstance,
                    TOOLBARICON *icons, char *xcursor)
{
    int         i;
    HWND        hwnd;
    WNDCLASS    wndclass;

    if (!ClassRegistered) {
        wndclass.style          = CS_HREDRAW | CS_VREDRAW;
        wndclass.lpfnWndProc    = ToolbarProc;
        wndclass.cbClsExtra     = 0;
        wndclass.cbWndExtra     = 0;
        wndclass.hInstance      = hInstance;
        wndclass.hIcon          = LoadIcon(NULL, IDI_APPLICATION);
        wndclass.hCursor        = LoadCursor(NULL, IDC_ARROW);
        wndclass.hbrBackground  = NULL;
        wndclass.lpszMenuName   = NULL;
        wndclass.lpszClassName  = TOOLBARCLASS;

        RegisterClass(&wndclass);

        wndclass.hCursor        = NULL;
        wndclass.lpfnWndProc    = ToolbarButtonProc;
        wndclass.lpszClassName  = TOOLBARBUTTONCLASS;

        RegisterClass(&wndclass);
    }

    hwnd = CreateWindow(TOOLBARCLASS, "Iconbar", WS_CHILD | WS_VISIBLE, x, y,
                        (width <= 0) ? CW_USEDEFAULT : width, (height <= 0) ? CW_USEDEFAULT : height,
                        parent, (HMENU) id, hInstance, NULL);

    if (hwnd == NULL) return (NULL);

    SetProp(hwnd, "TB ICONS", (HANDLE) icons);
    SetProp(hwnd, "TB NUMICONS", (HANDLE) n);
    SetProp(hwnd, "TB THICKNESS", (HANDLE) thickness);
    if (xcursor != NULL) {
        SetProp(hwnd, "TB CURSOR", LoadCursor(hInstance, xcursor));
    } else {
        SetProp(hwnd, "TB CURSOR", NULL);
    }

    /* Create the children */

    for (i = 0; i < n; i++) {
        icons[i].oldstate = 0;
        icons[i].hwnd = CreateWindow(TOOLBARBUTTONCLASS, "",
                                     WS_CHILD | WS_VISIBLE,
                                     icons[i].x, icons[i].y,
                                     icons[i].width, icons[i].height,
                                     hwnd, (HMENU) icons[i].id, hInstance, NULL);

        SetProp(icons[i].hwnd, "TB ICONS", &(icons[i]));
    }

    return (hwnd);
}



long CALLBACK ToolbarProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int             i;
    HDC             hdc;
    PAINTSTRUCT     ps;
    RECT            rect;
    TOOLBARICON     *icons;


    switch (message) {

        case WM_PAINT:
            GetClientRect(hwnd, &rect);

            hdc = BeginPaint(hwnd, &ps);
            Create3DEffect(hdc, &rect, (int) GetProp(hwnd, "TB THICKNESS"), 0);
            EndPaint(hwnd, &ps);
            return (0);

        case WM_COMMAND:
#ifdef USEWIN32
            SendMessage(GetParent(hwnd), WM_COMMAND,
                        wParam | (GetWindowLong(hwnd, GWL_ID) << 8), lParam);
#else
            SendMessage(GetParent(hwnd), WM_COMMAND,
                        wParam | (GetWindowWord(hwnd, GWW_ID) << 8), lParam);
#endif USEWIN32
            return (0);

        case BM_SETSTATE:
        case BM_GETSTATE:
            icons = (TOOLBARICON *) GetProp(hwnd, "TB ICONS");
            for (i = 0; icons[i].id != LOWORD(lParam); i++);
            return (SendMessage(icons[i].hwnd, message, wParam, 0L));

        case WM_DESTROY:
            RemoveProp(hwnd, "TB ICONS");
            RemoveProp(hwnd, "TB NUMICONS");
            RemoveProp(hwnd, "TB THICKNESS");
            RemoveProp(hwnd, "TB CURSOR");
    }

    return (DefWindowProc(hwnd, message, wParam, lParam));
}



long CALLBACK ToolbarButtonProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int             i, j;
    BOOL            crossed;
    HDC             hdc, hdcmem;
    PAINTSTRUCT     ps;
    TOOLBARICON     *icon;
    HBITMAP         hbitmap;
    HANDLE          hInstance;


    icon = (TOOLBARICON *) GetProp(hwnd, "TB ICONS");

    switch (message) {

    case BM_SETSTATE:
        if (icon->state == wParam) return (0);

        icon->oldstate = icon->state = wParam;
        InvalidateRect(hwnd, NULL, FALSE);
        UpdateWindow(hwnd);
        return (0);

    case BM_GETSTATE:
        return (icon->state);

    case WM_LBUTTONDOWN:
        if (icon->state < 0) return (0);

        icon->oldstate = icon->state;
        icon->state = MAXSTATES;
        InvalidateRect(hwnd, NULL, FALSE);
        UpdateWindow(hwnd);
        SetCapture(hwnd);
        Capturing = TRUE;
        return (0);

    case WM_LBUTTONUP:
        if (!Capturing || icon->state < 0) return (0);

        ReleaseCapture();
        Capturing = FALSE;

        i = LOWORD(lParam);
        j = HIWORD(lParam);

        if (i < 0 || i >= icon->width || j < 0 || j >= icon->height) {
            return (0);
        } else if (icon->cycle <= 0) {
            icon->state = icon->oldstate;
        } else {
            icon->oldstate++;
            if (icon->oldstate >= icon->cycle) icon->oldstate = 0;
            icon->state = icon->oldstate;
        }

        InvalidateRect(hwnd, NULL, FALSE);
        UpdateWindow(hwnd);
        CursorShape (hwnd, icon->state);
        SendMessage(GetParent(hwnd), WM_COMMAND, icon->id, MAKELONG(hwnd, BN_CLICKED));
        return (0);

    case WM_MOUSEMOVE:
        CursorShape (hwnd, icon->state);
        if (!Capturing || icon->state < 0) return (0);

        i = LOWORD(lParam);
        j = HIWORD(lParam);

        if (i < 0 || i >= icon->width || j < 0 || j >= icon->height) {
            crossed = (icon->state == MAXSTATES);
            icon->state = icon->oldstate;
        } else {
            crossed = (icon->state != MAXSTATES);
            icon->state = MAXSTATES;
        }

        if (crossed) {
            InvalidateRect(hwnd, NULL, FALSE);
            UpdateWindow(hwnd);
        }
        return (0);

    case WM_PAINT:
#ifdef USEWIN32
        hInstance = (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE);
#else
        hInstance = GetWindowWord(hwnd, GWW_HINSTANCE);
#endif USEWIN32

        hdc = BeginPaint(hwnd, &ps);

        switch (icon->state) {
            default: if (icon->disabled != NULL) {
                        hbitmap = LoadBitmap(hInstance, icon->disabled);
                     }
                     break;
            case 0:  hbitmap = LoadBitmap(hInstance, icon->undepressed); break;
            case 1:  hbitmap = LoadBitmap(hInstance, icon->depressed); break;
            case 2:  hbitmap = LoadBitmap(hInstance, icon->grayed); break;

            case MAXSTATES:
                    hbitmap = LoadBitmap(hInstance, icon->pressing); break;
        }

        hdcmem = CreateCompatibleDC(hdc);
        SelectObject(hdcmem, hbitmap);

        BitBlt(hdc, 0, 0, icon->width, icon->height, hdcmem, 0, 0, SRCCOPY);

        DeleteDC(hdcmem);
        DeleteObject(hbitmap);
        EndPaint(hwnd, &ps);
        return (0);

    }

    return (DefWindowProc(hwnd, message, wParam, lParam));
}
